
c * * * periodic 1d electrostatic particle simulation kernel code * * *
c for IBM SP1 ---> SM 

      program beps1k

      parameter( indx =   7, npx =    5120, npxb = 512)
cbenc1 parameter( indx =  11, npx =   409600, npxb = 40960)
cbenc2 parameter( indx =  14, npx =  4096000, npxb =   409600)
cbenc3 parameter( indx =  18, npx = 40960000, npxb =  4096000)

      parameter( tend =  45.000, dt = 0.2000000e+00)

      parameter( vtx =   1.000,  vdx= 5.000, vtdx =   0.500)
      parameter( avdx=   5.000)


      parameter( indnvp= 2, idps =    2, idimp =   2, mshare =   1)

      parameter(np=npx+npxb)
      parameter(nx=2**indx,nxh=nx/2)

      parameter(nloop=tend/dt+.0001)

      parameter(nvp=2**indnvp,nblok=1+mshare*(nvp-1))
      parameter(npmax=(np/nvp)*1.01+250,nxpmx=(nx-1)/nvp+4)

      parameter(kxp=(nxh-1)/nvp+1,kblok=1+mshare*(nxh/kxp-1))
      parameter(nbmax=1+(2*(npx*vtx+npxb*vtdx)+1.4*npxb*avdx)*dt/nx)
      parameter(ntmax=2*nbmax)

      parameter(npd=128)
      complex ffc, fc, qc, sct


      dimension part(idimp,npmax,nblok)

      dimension q(nxpmx,nblok), fx(nxpmx,nblok)
      dimension qc(kxp,kblok), fc(kxp,kblok)
      dimension ffc(kxp,kblok)
      dimension mixup(kxp,kblok), sct(kxp,kblok)
      dimension edges(idps,nblok)
      dimension nxp(nblok), noff(nblok)
      dimension npp(nblok), nps(nblok)

      dimension sbufl(idimp,nbmax,nblok), sbufr(idimp,nbmax,nblok)
      dimension rbufl(idimp,nbmax,nblok), rbufr(idimp,nbmax,nblok)

      dimension ihole(ntmax,nblok)
      dimension jsl(idps,nblok), jsr(idps,nblok)
      dimension jss(idps,nblok), scr(idps,nblok)

cext      common /large/ part
cext   dimension maskp(npmax,nblok)
cext   dimension nn(npd,nblok), nu(npd,nblok), nl(npd,nblok)
cext   dimension amx(npd,nblok), dxu(npd,nblok), dxl(npd,nblok)
cext   dimension tx(nxpmx,nblok), ty(nxpmx,nblok)

  991 format (5h t = ,i7)
  992 format (19h * * * q.e.d. * * *)
  993 format (34h field, kinetic, total energies = ,3e14.7)

      data qme,ax /-1.,.8666667/

      real PGI_t0, PGI_total 


!HPF$ PROCESSORS :: PROCS(4)
!HPF$ DISTRIBUTE (*, BLOCK) ONTO PROCS :: q, qc, scr
!HPF$ DISTRIBUTE (BLOCK) ONTO PROCS :: nps, npp, nxp, noff
!HPF$ DISTRIBUTE (*,*,BLOCK) ONTO PROCS :: part

!HPF$ DISTRIBUTE (*, BLOCK) ONTO PROCS :: jsl, jsr, jss
c ------------------------------------
c initialize for parallel processing
c ------------------------------------
      call PPINIT(idproc,nvp)
      kstrt = idproc + 1
      if (kstrt.eq.1) then
         open(unit=6,file='output1',form='formatted',status='unknown')
      endif
c initialize timer
      call TIMER(PGI_t0)

c initialize constants
      itime = 0
      anx = float(nx)
      qtme = qme*dt
      affp = anx/float(np)
      zero = 0.
c ------------------------------
c calculate partition variables
c ------------------------------
      call DCOMP1(edges,nxp,noff,nx,kstrt,nvp,idps,nblok)
c prepare fft tables
      isign = 0
      call PFFT1R(qc,fc,isign,mixup,sct,indx,kstrt,kxp,kblok)
c calculate form factors
      call PPOIS1(qc,fc,isign,ffc,ax,affp,we,nx,kstrt,kxp,kblok)
c -------------------------------------------------------------
c initial density profile and maxwellian velocity distribution
c background electrons
c -------------------------------------------------------------
      nps(1:nblok) = 1
      npp(1:nblok) = 0
    
      forall(k=1:nblok) q(1:nxp(k)+3, k) = 0.  !indep, new(npx3,j)

      if (npx.gt.0) call PISTR1 (part,edges,npp,nps,vtx,zero,npx,nx,idim
     1p,npmax,nblok,idps)
c -----------------
c beam electrons
c -----------------
      nps(1:nblok) = npp(1:nblok) + 1
      if (npxb.gt.0) call PISTR1 (part,edges,npp,nps,vtdx,vdx,npxb,nx,
     1idimp,npmax,nblok,idps)
c ----------------------------------------
c deposit charge for initial distribution /pxost1/pyost1
c ----------------------------------------
      call PDOST1 (part,q,npp,noff,qme,idimp,npmax,nblok,nxpmx)

c add background ion density
      qi0 = -qme/affp
      forall (k=1:nblok) q(2:nxp(k)+1, k) = q(2:nxp(k)+1, k) + qi0 !indep(nxp1, j)

c =======================================
c
c * * * start main iteration loop * * *
c
c =======================================
  500 if (nloop.le.itime) go to 2000
      if (kstrt.eq.1) write (6,991) itime
c ----------------------------------
c transform charge to fourier space
c ----------------------------------
      isign = -1
c -------------------------------------------------------------------
c copy data from particle to field partition, and add up guard cells
c -------------------------------------------------------------------
      call cppfp1 (q,qc,isign,scr,kstrt,nvp,nxpmx,nblok,kxp,kblok,idps)
      call pfft1r(qc,fc,isign,mixup,sct,indx,kstrt,kxp,kblok)
c ----------------------------------------
c calculate force/charge in fourier space
c ----------------------------------------
      call ppois1 (qc,fc,isign,ffc,ax,affp,we,nx,kstrt,kxp,kblok)
c -------------------------------------
c transform force/charge to real space
c -------------------------------------
      isign = 1
      call pfft1r(fc,qc,isign,mixup,sct,indx,kstrt,kxp,kblok)
c --------------------------------------------------------------------
c copy data from field to particle partition, and copy to guard cells
c --------------------------------------------------------------------
      call cppfp1 (fx,fc,isign,scr,kstrt,nvp,nxpmx,nblok,kxp,kblok,idps)
c ----------------------------------------
c particle push and charge density update
c ----------------------------------------
c     call timera(-1,'push    ')
c -----------------------------------
c initialize charge density to zero
c -----------------------------------
      forall(k=1:nblok) q(1:nxp(k)+3, k) = 0.   ! indep new(j)

      wke = 0.
c ---------------
c push particles
c ---------------
      call ppush1(part,fx,npp,noff,qtme,dt,wke,idimp,npmax,nblok,nxpmx)
c ------------------------------------------------
c move particles into appropriate spatial regions /pxmov1
c ------------------------------------------------
      call pmove1 (part,edges,npp,sbufr,sbufl,rbufr,rbufl,ihole,jsr,jsl,
     1jss,nx,kstrt,nvp,idimp,npmax,nblok,idps,nbmax,ntmax,ierr)
c ---------------
c deposit charge /pxost1 /pyost1
c ---------------
      call pdost1 (part,q,npp,noff,qme,idimp,npmax,nblok,nxpmx)
c ----------------------
c add background charge
c ----------------------
      forall (k=1:nblok) q(2:nxp(k)+1, k) = q(2:nxp(k)+1, k) + qi0 !indep new(nxp1,j) 
c ------------------
c energy diagnostic
c ------------------
      wt = we + wke
      if (kstrt.eq.1) write (6,993) we, wke, wt
      itime = itime + 1
      go to 500
 2000 continue
c ====================================
c
c * * * end main iteration loop * * *
c
c ====================================
      if (kstrt.eq.1) write (6,992)
      call TIMER(PGI_total)
      print *,'total time = ',(PGI_total - PGI_t0), 'seconds',
     1      PGI_total, PGI_t0
      stop
      end
C =========================================================================
c
c - - - this subroutine initializes parallel processing
c idproc = processor id (OUT)
c nvp = number of real or virtual processors requested (IN)
c nproc = number of real or virtual processors obtained
c nprec = (4,8) = number of bytes per word 
C =========================================================================
      subroutine ppinit(idproc,nvp)
      character*8 jbname
      common /pparms/ nproc, nprec, lgrp, jbname

      nproc = nvp
      idproc = 0

      if (nproc.ne.nvp) then
         write (6,*) ' processor number error: nvp,nproc=',nvp,nproc
         stop
      endif
      return
      end

C =========================================================================
c - - -  this subroutine determines spatial boundaries for particle
c - - -  decomposition, calculates number of grid points in each spatial
c - - -  region, and the offset of these grid points from the global address
C =========================================================================
      subroutine dcomp1(edges,nxp,noff,nx,kstrt,nvp,idps,nblok)
      dimension edges(idps,nblok)
      dimension nxp(nblok), noff(nblok)
      ks = kstrt - 2
      at1 = float(nx)/float(nvp)
!HPF$ independent, new(kb, kr)
      do 10 k = 1, nblok
         kb = k + ks
         edges(1,k) = at1*float(kb)
         noff(k) = edges(1,k) + .5
         edges(2,k) = at1*float(kb + 1)
         kr = edges(2,k) + .5
         nxp(k) = kr - noff(k)
   10 continue
      return
      end
C =========================================================================
c - - - for 1d code, this subroutine calculates initial particle co-ordinate
c - - - and velocity, with uniform density and maxwellian velocity with drift,
c - - - for distributed data.
c ranorm = gaussian random number with zero mean and unit variance
c with spatial decomposition  
C =========================================================================
      subroutine pistr1 (part,edges,npp,nps,vtx,vdx,npx,nx,idimp,npmax,
     1   nblok,idps)
      character*8 jbname
      double precision ranorm
      common /pparms/ nproc, nprec, lgrp, jbname
      external s_vadd
      dimension part(idimp,npmax,nblok)
      dimension edges(idps,nblok), npp(nblok), nps(nblok)
c particle distribution
      at1 = float(nx)/float(npx)
      do 20 j = 1, npx
c uniform density profile
         xt = at1*(float(j) - .5)
c maxwellian velocity distribution
         vxt = vtx*ranorm(d)
! HPF$    independent
         do 10 k = 1, nblok
          if ((xt.ge.edges(1,k)).and.(xt.lt.edges(2,k))) then
            npp(k) = npp(k) + 1
            part(1,npp(k),k) = xt
            part(2,npp(k),k) = vxt
          endif
   10    continue
   20 continue
c add correct drift
      sum1 = 0.
      do 40 k = 1, nblok
      sum0 = 0.
      do 30 j = nps(k), npp(k)
         sum0 = sum0 + part(2,j,k)
   30 continue
      sum1 = sum1 + sum0
   40 continue


      sum1 = sum1/float(npx) - vdx
!HPF$ independent , new(j)
      do 60 k = 1, nblok
         do 50 j = nps(k), npp(k)
            part(2,j,k) = part(2,j,k) - sum1
   50    continue
   60 continue
      return
      end
C =========================================================================
c for 1d code, this subroutine calculates particle charge density
c using second-order spline interpolation, with periodic boundaries
c and distributed data.
c charge density is approximated by values at the nearest grid points
c q(n)=qm*(.75-dx**2), q(n+1)=.5*qm*(.5+dx)**2, q(n-1)=.5*qm*(.5-dx)**2
c where n = nearest grid point and dx = x-n
c scalar version with spatial decomposition
C =========================================================================
      subroutine pdost1 (part,q,npp,noff,qm,idimp,npmax,nblok,nxpmx)
      dimension part(idimp,npmax,nblok), q(nxpmx,nblok)
      dimension npp(nblok), noff(nblok)
      do 20 k = 1, nblok
         nnoff = noff(k) - 2
         do 10 j = 1, npp(k)
            nn = part(1,j,k) + .5
            dx = part(1,j,k) - float(nn)
            nn = nn - nnoff
            q(nn,k) = q(nn,k) + qm*(.75 - dx*dx)
            nn = nn + 1
            q(nn,k) = q(nn,k) + .5*qm*(.5 + dx)**2
            nn = nn - 2
            q(nn,k) = q(nn,k) + .5*qm*(.5 - dx)**2
   10    continue
   20 continue
      return
      end


C =========================================================================
c thus subroutine copies data between particle and field partitions,
c where the field and particle partitions are assumed to be the same.
c in going from particle to field partitions, data from guard cells
c is added to the field partition.  in going from field to particle
c partitions, data is copied to the guard cells.
c * f(j,k) = real data for grid j in particle partition k.  
c     # grids per partition is uniform and includes 3 extra guard cells.
c * fc(j,k) = complex data for grid j in field partition k.  
c     # grids per partition is uniform and maximum # partitions is nxh
c * if isign = -1, data is copied from particle to field partition.
c   if isign = +1, data is copied from field to particle partition.
C =========================================================================
      subroutine cppfp1 (f,fc,isign,scr,kstrt,nvp,nxpmx,nblok,kxp,kblok,
     1  idps)
      character*8 jbname
      common /pparms/ nproc, nprec, lgrp, jbname
      complex fc
      dimension f(nxpmx,nblok), fc(kxp,kblok)
      dimension scr(idps,nblok)
      ks = kstrt - 2
      moff = nxpmx*nvp
      if (isign) 10, 60, 70
c ---------------------------------------
c copy from particle to field partition
c ---------------------------------------
   10 CONTINUE
      forall (k=1:kblok) fc(1:kxp,k) = cmplx(f(2:2*kxp:2,k), 
     2     f(3:2*kxp+1:2, k))

! HPF$ INDEPENDENT, NEW(J)
!      DO k = 1, kblok
!         do 20 j = 1, kxp
!            fc(j,k) = cmplx(f(2*j,k),f(2*j+1,k))
!   20    continue
!      ENDDO
c ---------------------------------------
c add guard cells
c ---------------------------------------
      scr(1,:) = CSHIFT(f(2*kxp+2, :), -1)  !left to right
      scr(2,:) = CSHIFT(f(2*kxp+3, :), -1)
      fc(1,:) = fc(1,:) + cmplx(scr(1,:),scr(2,:))

      scr(1,:) = CSHIFT(f(1, :), +1)        ! right to left
      fc(kxp,:) = fc(kxp,:) + cmplx(0.,scr(1,:))

   60  return
c --------------------------------------
c copy from field to particle partition
c --------------------------------------
   70 CONTINUE
!HPF$ INDEPENDENT , NEW(J)
      DO k = 1, kblok
         do 80 j = 1, kxp
            f(2*j,k) = real(fc(j,k))
            f(2*j+1,k) = aimag(fc(j,k))
   80    continue
      ENDDO
c --------------------
c copy to guard cells
c -------------------- 
      f(1,:) = aimag(CSHIFT(fc(kxp,:), -1))     ! from left

      f(2*kxp+2,:) = real(CSHIFT(fc(1,:), +1))  ! from right
      f(2*kxp+3,:) = aimag(CSHIFT(fc(1,:), +1))

      return
      end
C =========================================================================
c this subroutine solves 1d poisson's equation in fourier space for
c force/charge (or convolution of electric field over particle shape)
c with periodic boundary conditions, for distributed data.
c the equation used is:
c fx(k) = -sqrt(-1)*k*g(k)*s(k), where k = 2pi*j/nx, j = fourier mode,
c g(k) = (affp/k**2)*s(k), and s(k) = exp(-(k*ax)**2/2), except for
c fx(k=0) = fx(k=pi) = 0.
c kstrt = starting data block number
c if isign = 0, form factor array is prepared
c if isign is not equal to 0, force/charge is calculated.
c aimag(ffc(j,k)) = particle shape factor s for fourier mode jj - 1
c real(ffc(j,k)) = potential green's function g for fourier mode jj - 1
c ax = half-width of particle in x direction
c affp = normalization constant = nx/np, where np = number of particles
c electric field energy is also calculated, using
c wp = nx*sum((affp/k**2)*|q(k)*s(k)|**2)
c nx = system length in x direction
C =========================================================================
      subroutine ppois1 (q,fx,isign,ffc,ax,affp,wp,nx,kstrt,kxp,kblok)
      character*8 jbname
      common /pparms/ nproc, nprec, lgrp, jbname
      complex q, fx, ffc
      external s_vadd
      dimension q(kxp,kblok), fx(kxp,kblok)
      dimension ffc(kxp,kblok)
      nxh = nx/2
      ks = kstrt - 2
      dnx = 6.28318530717959/float(nx)
      if (isign.ne.0) go to 30
      if (kstrt.gt.nxh) return
c ----------------------------
c prepare form factor array
c ----------------------------
      do 20 k = 1, kblok
         koff = kxp*(k + ks) - 1
         do 10 j = 1, kxp
            dkx = dnx*float(j + koff)
            at1 = exp(-.5*(dkx*ax)**2)
            if (dkx.eq.0.) then
               ffc(j,k) = cmplx(0.,1.)
            else
               ffc(j,k) = cmplx(affp*at1/(dkx*dkx),at1)
            endif
   10    continue
   20 continue
      return
c ---------------------------------------------
c calculate force/charge and sum field energy
c ---------------------------------------------
   30 wp = 0.
      if (kstrt.gt.nxh) go to 60
      do 50 k = 1, kblok
         koff = kxp*(k + ks) - 1
         sum1 = 0.
         do 40 j = 1, kxp
            at1 = real(ffc(j,k))*aimag(ffc(j,k))
            at2 = dnx*float(j + koff)*at1
            fx(j,k) = at2*cmplx(aimag(q(j,k)),-real(q(j,k)))
            sum1 = sum1 + at1*real(q(j,k)*conjg(q(j,k)))
   40    continue
         wp = wp + sum1
   50 continue
   60 continue

      wp = float(nx)*wp
      return
      end
C =========================================================================
c For 1d code, this subroutine updates particle co-ordinate and velocity
c using leap-frog scheme in time and second-order spline interpolation
c in space, with periodic boundary conditions, for distributed data.
c  Equations used are:
c *  v(t+dt/2) = v(t-dt/2) + (q/m)*fx(x(t))*dt, where q/m is charge/mass,
c    and x(t+dt) = x(t) + v(t+dt/2)*dt
c *  fx(x(t)) is approximated by interpolation from the nearest grid points
c    fx(x) = (.75-dx**2)*fx(n)+.5*(fx(n+1)*(.5+dx)**2+fx(n-1)*(.5-dx)**2)
c    where n = nearest grid point and dx = x-n
c * qtm = particle charge/mass ratio times dt
c * kinetic energy/mass at time t is also calculated, using
c    ek = .125*sum((v(t+dt/2)+v(t-dt/2))**2)
c scalar version with spatial decomposition
C =========================================================================
      subroutine ppush1(part,fx,npp,noff,qtm,dt,ek,idimp,npmax,nblok,nxp
     1mx)
      character*8 jbname
      common /pparms/ nproc, nprec, lgrp, jbname
      external s_vadd
      dimension part(idimp,npmax,nblok), fx(nxpmx,nblok)
      dimension npp(nblok), noff(nblok)
      dimension sum_d(nblok)
!HPF$ processors :: PROCS(4)
!HPF$ distribute (block) onto PROCS :: sum_d
      sum_d = 0.
      sum1 = 0.
!HPF$ INDEPENDENT, NEW(nnoff, j, nn, dx)
      do 20 k = 1, nblok
         nnoff = noff(k) - 2
         do 10 j = 1, npp(k)
c ----------------------------
c find interpolation weights
c ----------------------------
            nn = part(1,j,k) + .5
            dx = part(1,j,k) - float(nn)
            nn = nn - nnoff
c ----------------------------
c find acceleration
c ----------------------------
            dx = (.75 - dx*dx)*fx(nn,k) + .5*(fx(nn+1,k)*(.5 + dx)**2 
     1        + fx(nn-1,k)*(.5 - dx)**2)
c ----------------------------
c new velocity
c ----------------------------
            dx = part(2,j,k) + qtm*dx
c ----------------------------
c average kinetic energy
c ----------------------------
            sum_d(k) = sum_d(k) +(part(2,j,k) + dx)**2  
!            sum1 = sum1 + (part(2,j,k) + dx)**2
            part(2,j,k) = dx
c ----------------------------
c new position
c ----------------------------
            part(1,j,k) = part(1,j,k) + dx*dt
   10    continue
   20 continue
      sum1 = SUM(sum_d)


c normalize kinetic energy
      ek = ek + .125*sum1
      return
      end

c vector version with spatial dec.::push routine,INCLUDE 'vector/pvpsh1.f'
c optimized version with spatial dec.:: push routine INCLUDE 'vector/pxpsh1.f'
C =========================================================================
c this subroutine moves particles into appropriate spatial regions
C =========================================================================
      subroutine pmove1 (part,edges,npp,sbufr,sbufl,rbufr,rbufl,ihole,
     1jsr,jsl,jss,nx,kstrt,nvp,idimp,npmax,nblok,idps,nbmax,ntmax,ierr)
      character*8 jbname
      common /pparms/ nproc, nprec, lgrp, jbname
      external i_vadd
      dimension part(idimp,npmax,nblok)
      dimension edges(idps,nblok), npp(nblok)
      dimension sbufl(idimp,nbmax,nblok), sbufr(idimp,nbmax,nblok)
      dimension rbufl(idimp,nbmax,nblok), rbufr(idimp,nbmax,nblok)
      dimension jsl(idps,nblok), jsr(idps,nblok), jss(idps,nblok)
      dimension ihole(ntmax,nblok)
      anx = float(nx)
      ks = kstrt - 2
      iter = 4
c ----------------------------------------------------------------
c debugging section: count total number of particles before move
c ----------------------------------------------------------------
      npr = SUM(npp)

c --------------------------
c buffer outgoing particles
c --------------------------
   10 forall (k=1:nblok) jsl(1,k) = 0
      forall (k=1:nblok) jsr(1,k) = 0
      forall (k=1:nblok) jss(2,k) = 0
      do 30 k = 1, nblok
         kb = k + ks
         do 20 j = 1, npp(k)
            xt = part(1,j,k)
c ------------------------------
c particles going to the right
c ------------------------------
            if (xt.ge.edges(2,k)) then
               if (jsr(1,k).lt.nbmax) then
                  jsr(1,k) = jsr(1,k) + 1
                  if ((kb+1).eq.nvp) xt = xt - anx
                  sbufr(1,jsr(1,k),k) = xt
                  sbufr(2,jsr(1,k),k) = part(2,j,k)
                  ihole(jsl(1,k)+jsr(1,k),k) = j
               else
                  jss(2,k) = 1
               endif
c ----------------------------
c particles going to the left
c ----------------------------
            elseif (xt.lt.edges(1,k)) then
               if (jsl(1,k).lt.nbmax) then
                  jsl(1,k) = jsl(1,k) + 1
                  if (kb.eq.0) xt = xt + anx
                  sbufl(1,jsl(1,k),k) = xt
                  sbufl(2,jsl(1,k),k) = part(2,j,k)
                  ihole(jsl(1,k)+jsr(1,k),k) = j
               else
                  jss(2,k) = 1
            endif
         endif
   20    continue
         jss(1,k) = jsl(1,k) + jsr(1,k)
   30 continue
c --------------------------------
c check for full buffer condition
c --------------------------------
      ibflg = SUM(jss(2,:))

c -------------------------
c copy particle buffers
c -------------------------
   50 iter = iter + 4
c -------------------------
c get particles from left
c -------------------------
      jsl(2,:) = CSHIFT(jsr(1,:), -1)


! HPF$ independent, NEW (kl, j)
!      DO k = 1, nblok
!         do 60 j = 1, jsl(2,k)
!            rbufl(1,j,:) = CSHIFT(sbufr(1,j,:), -1)
!  60     enddo
!      ENDDO

! HPF$ INDEPENDENT 
!      DO k = 1, nblok
!         rbufl(1,1:jsl(2,k),:) = CSHIFT(sbufr(1,1:jsl(2,k),:), -1, 2)
!      ENDDO

      do 80 k = 1, nblok
         kl = k + ks
      do 60 j = 1, jsl(2,k)
         if (kl.lt.1) then
            rbufl(1,j,k) = sbufr(1,j,kl+nvp)
            rbufl(2,j,k) = sbufr(2,j,kl+nvp)
         else
            rbufl(1,j,k) = sbufr(1,j,kl)
            rbufl(2,j,k) = sbufr(2,j,kl)
         endif
   60 continue
   80 enddo
c --------------------------
c get particles from right
c --------------------------
      jsr(2,:) = CSHIFT(jsl(1,:), +1)
      do 75 k = 1, nblok
         kr = k + ks + 2

         do 70 j = 1, jsr(2,k)
         if (kr.gt.nvp) then
           rbufr(1,j,k) = sbufl(1,j,kr-nvp)
           rbufr(2,j,k) = sbufl(2,j,kr-nvp)
         else
           rbufr(1,j,k) = sbufl(1,j,kr)
           rbufr(2,j,k) = sbufl(2,j,kr)
         endif
  70  continue


  75  enddo

c ------------------------------------------- 
c check if particles must be passed further
c ------------------------------------------- 
      nps = 0
      do 110 k = 1, nblok
c ------------------------------------------------------------------ 
c check if any particles coming from right must go further to left
c ------------------------------------------------------------------ 
         jsl(1,k) = 0
         do 90 j = 1, jsr(2,k)
            if (rbufr(1,j,k).lt.edges(1,k)) jsl(1,k) = 1
   90    continue
c ------------------------------------------------------------------ 
c check if any particles coming from left must go further to right
c ------------------------------------------------------------------ 
         jsr(1,k) = 0
         do 100 j = 1, jsl(2,k)
            if (rbufl(1,j,k).ge.edges(2,k)) jsr(1,k) = 1
  100    continue
         nps = nps + (jsl(1,k) + jsr(1,k))
  110 continue


      if (nps.eq.0) go to 150
c -------------------------------------------
c remove particles which do not belong here
c -------------------------------------------
      jsl(1,:) = 0
      jss(2,:) = 0
      jsr(1,:) = 0
      do 140 k = 1, nblok         
         kb = k + ks
c -------------------------------------------
c first check particles coming from right
c -------------------------------------------
         do 120 j = 1, jsr(2,k)
            xt = rbufr(1,j,k)
c -------------------------------------------
c particles going to the left
c -------------------------------------------
            if (xt.lt.edges(1,k)) then
               jsl(1,k) = jsl(1,k) + 1
               if (kb.eq.0) xt = xt + anx
               sbufl(1,jsl(1,k),k) = xt
               sbufl(2,jsl(1,k),k) = rbufr(2,j,k)
c -------------------------------------------
c particles staying here
c -------------------------------------------
            else
               jss(2,k) = jss(2,k) + 1
               rbufr(1,jss(2,k),k) = xt
               rbufr(2,jss(2,k),k) = rbufr(2,j,k)
            endif
  120    continue
         jsr(2,k) = jss(2,k)
c -------------------------------------------
c next check particles coming from left
c -------------------------------------------
!         jsr(1,k) = 0
!         jss(2,k) = 0
         do 130 j = 1, jsl(2,k)
            xt = rbufl(1,j,k)
c -------------------------------------------
c particles going to the right
c -------------------------------------------
            if (xt.ge.edges(2,k)) then
               jsr(1,k) = jsr(1,k) + 1
               if ((kb+1).eq.nvp) xt = xt - anx
               sbufr(1,jsr(1,k),k) = xt
               sbufr(2,jsr(1,k),k) = rbufl(2,j,k)
c -------------------------------------------
c particles staying here
c -------------------------------------------
            else
               jss(2,k) = jss(2,k) + 1
               rbufl(1,jss(2,k),k) = xt
               rbufl(2,jss(2,k),k) = rbufl(2,j,k)
            endif
  130    continue
         jsl(2,k) = jss(2,k)
  140 continue
c -------------------------------------------
c check if move would overflow particle array
c -------------------------------------------
  150 ierr = 0
      do 160 k = 1, nblok
         if ((npp(k)+jsl(2,k)+jsr(2,k)-jss(1,k)).gt.npmax) then
            jss(2,k) = 1
         else
            jss(2,k) = 0
         endif
         ierr = ierr + jss(2,k)
  160 continue
      if (ierr.gt.0) then
         write (6,*) 'particle overflow error, ierr =', ierr
         return
      endif
      do 200 k = 1, nblok
c ------------------------------------------------------
c distribute incoming particles from buffers
c distribute particles coming from left into holes
c ------------------------------------------------------
         jss(2,k) = min0(jss(1,k),jsl(2,k))
         do 170 j = 1, jss(2,k)
            part(1,ihole(j,k),k) = rbufl(1,j,k)
            part(2,ihole(j,k),k) = rbufl(2,j,k)
  170    continue
         if (jss(1,k).gt.jsl(2,k)) then
            jss(2,k) = min0(jss(1,k)-jsl(2,k),jsr(2,k))
         else
            jss(2,k) = jsl(2,k) - jss(1,k)
         endif
         do 180 j = 1, jss(2,k)
c ------------------------------------------------------
c no more particles coming from left
c distribute particles coming from right into holes
c ------------------------------------------------------
            if (jss(1,k).gt.jsl(2,k)) then
               part(1,ihole(j+jsl(2,k),k),k) = rbufr(1,j,k)
               part(2,ihole(j+jsl(2,k),k),k) = rbufr(2,j,k)
            else
c ------------------------------------------------------
c no more holes
c distribute remaining particles from left into bottom
c ------------------------------------------------------
               part(1,j+npp(k),k) = rbufl(1,j+jss(1,k),k)
               part(2,j+npp(k),k) = rbufl(2,j+jss(1,k),k)
            endif
  180    continue
         if (jss(1,k).le.jsl(2,k)) then
            npp(k) = npp(k) + (jsl(2,k) - jss(1,k))
            jss(1,k) = jsl(2,k)
         endif
         jss(2,k) = jss(1,k) - (jsl(2,k) + jsr(2,k))
         if (jss(2,k).gt.0) then
            jss(1,k) = (jsl(2,k) + jsr(2,k))
            jsr(2,k) = jss(2,k)
         else
            jss(1,k) = jss(1,k) - jsl(2,k)
            jsr(2,k) = -jss(2,k)
         endif
         do 190 j = 1, jsr(2,k)
c ------------------------------------------------------
c holes left over
c fill up remaining holes in particle array with particles from bottom
c ------------------------------------------------------
            if (jss(2,k).gt.0) then
               j1 = npp(k) - j + 1
               j2 = jss(1,k) + jss(2,k) - j + 1
               if (j1.gt.ihole(j2,k)) then
c ------------------------------------------------------
c move particle only if it is below current hole
c ------------------------------------------------------
                  part(1,ihole(j2,k),k) = part(1,j1,k)
                  part(2,ihole(j2,k),k) = part(2,j1,k)
               endif
            else
c ------------------------------------------------------
c no more holes
c distribute remaining particles from right into bottom
c ------------------------------------------------------
               part(1,j+npp(k),k) = rbufr(1,j+jss(1,k),k)
               part(2,j+npp(k),k) = rbufr(2,j+jss(1,k),k)
            endif
  190    continue
         if (jss(2,k).gt.0) then
            npp(k) = npp(k) - jsr(2,k)
         else
            npp(k) = npp(k) + jsr(2,k)
         endif
         jss(1,k) = 0
  200 continue
c ------------------------------------------------------
c check if any particles have to be passed further
c ------------------------------------------------------
      if (nps.gt.0) go to 50
c check if buffer overflowed and more particles remain to be checked
      if (ibflg.gt.0) go to 10
c debugging section: count total number of particles after move
      nps = 0
      do 205 k = 1, nblok
      nps = nps + npp(k)
  205 continue
      if (npr.ne.nps) write (6,*) 'number error, np-old,np-new=',npr,nps
      return
      end


C =========================================================================
c - - - this subroutine performs a one dimensional real to complex fast
c - - - fourier transform and its inverse, using complex arithmetic,
c - - - for data which is distributed in blocks
c f = input and output data
c g = scratch array
c indx = power of 2 which determines length of transform, nx = 2**indx
c if isign = 0, the fft tables are prepared
c if isign = -1, an inverse fourier transform is performed
c f(n,m) = (1/nx)*sum(f(j,k)*exp(-sqrt(-1)*2pi*nn*jj/nx))
c if isign = 1, a forward fourier transform is performed
c f(j,k) = sum(f(n,m)*exp(sqrt(-1)*2pi*nn*jj/nx))
c where nn = n + kxp*(m - 1) and jj = j + kxp*(k - 1)
c kxp = number of data values per block
c mixup = array of bit reversed addresses
c sct = sine/cosine table
c the real data is stored in a complex array of length nx/2,
c with the odd/even points stored in the real/imaginary parts.
c the real part of fourier mode nx/2 is stored in the imaginary part
c of mode zero.
c written by viktor k. decyk, ucla
c parallel version
C =========================================================================
      subroutine pfft1r(f,g,isign,mixup,sct,indx,kstrt,kxp,kblok)
      character*8 jbname
      common /pparms/ nproc, nprec, lgrp, jbname
      complex f, g, sct, t1
      dimension f(kxp,kblok), g(kxp,kblok)
      dimension mixup(kxp,kblok), sct(kxp,kblok)
      nx = 2**indx
      nxh = nx/2
      if (kstrt.gt.nxh) return
      indx1 = indx - 1
      kbs = nxh/kxp
      kxp2 = kxp + 2
      ks = kstrt - 2
      dnx = 6.28318530717959/float(nx)
      if (isign) 60, 10, 220
c ---------------------------
c prepare fft tables
c bit-reverse index table:
c mixup(j,k) = 1 + reversed bits of (j + kxp*(k + kstrt - 2) - 1)
c ---------------------------
   10 call bitrv1(f,g,mixup,isign,indx1,kstrt,kxp,kblok)
c ---------------------------
c local sine/cosine table for the angles 2*n*pi/kxp
c ---------------------------
      do 50 k = 1, kblok
      do 40 j = 1, kxp
         arg = dnx*float((nxh*(j - 1))/kxp)
         sct(j,k) = cmplx(cos(arg),-sin(arg))
   40 continue
   50 continue
      return
c ---------------------------
c inverse fourier transform
c bit-reverse array elements to temporary
c ---------------------------
   60 call bitrv1(f,g,mixup,isign,indx1,kstrt,kxp,kblok)
      do 160 l = 1, indx1
      nxs = 2**(l - 1)
      kxs = nxs/kxp
c ---------------------------
c local calculation
c ---------------------------
      if (kxs.eq.0) then
c ---------------------------
c copy local data
c ---------------------------
         do 110 k = 1, kblok
         do 90 j = 1, kxp
            lb = (j - 1)/nxs
            if (lb.eq.(2*(lb/2))) then
               f(j,k) = g(j+nxs,k)
            else
               f(j,k) = g(j-nxs,k)
            endif
   90    continue
c ---------------------------
c perform reduction
c ---------------------------
         km = kxp/nxs
         do 100 j = 1, kxp
            kb = j - 1
            lb = kb/nxs
            t1 = sct(1+km*(kb-nxs*lb),k)
            if (lb.eq.(2*(lb/2))) then
               g(j,k) = g(j,k) + t1*f(j,k)
            else
               g(j,k) = f(j,k) - t1*g(j,k)
            endif
  100    continue
  110    continue
c ---------------------------
c distributed calculation
c ---------------------------
      else
c ---------------------------
c copy distributed data
c ---------------------------
         do 130 k = 1, kblok
            kb = k + ks
            lb = kb/kxs
            kb = kb + 1
        do 120 j = 1, kxp
        if (lb.eq.(2*(lb/2))) then
           f(j,k) = g(j,kb+kxs)
        else
           f(j,k) = g(j,kb-kxs)
        endif
 120    continue


  130    continue
c ------------------
c perform reduction
c ------------------
         km = nxh/nxs
         dns = dnx*float(km)
         do 150 k = 1, kblok
            kb = k + ks
            lb = kb/kxs
            kb = kxp*(kb - kxs*lb) - 1
            lb = lb - 2*(lb/2)
            do 140 j = 1, kxp
               arg = dns*float(j + kb)
               t1 = cmplx(cos(arg),-sin(arg))
               if (lb.eq.0) then
                  g(j,k) = g(j,k) + t1*f(j,k)
               else
                  g(j,k) = f(j,k) - t1*g(j,k)
               endif
  140       continue
  150    continue
      endif
  160 continue
c this segment is used for distributed memory sp1 computers
c unscramble coefficients and normalize result
c copy distributed data
      do 180 k = 1, kblok
         kb = k + ks
         kk = kbs - kb
      do 170 j = 2, kxp
         f(j,k) = g(j,kk)
 170  continue
      if (kb.gt.0) f(1,k) = g(1,kk+1)

  180  continue
c ------------------
c perform reduction
c ------------------
         ani = .5/float(nx)
         do 210 k = 1, kblok
            kb = k + ks
            koff = kxp*kb - 1
            do 190 j = 2, kxp
               jj = kxp2 - j
               arg = dnx*float(j + koff)
               g(j,k) = ani*((g(j,k) + conjg(f(jj,k))) 
     1            - (g(j,k) 
     2            - conjg(f(jj,k)))*cmplx(sin(arg),cos(arg)))
  190       continue
            if (kb.eq.0) then
               g(1,k) = 2.*ani*cmplx(real(g(1,k)) + 
     1                   aimag(g(1,k)),real(g(1,k))
     2                   - aimag(g(1,k)))
            else
               arg = dnx*float(1 + koff)
               g(1,k) = ani*((g(1,k) + conjg(f(1,k))) - 
     1                   (g(1,k) - conjg(f(1,k)))*
     2                   cmplx(sin(arg),cos(arg)))
            endif
c copy result to destination
            do 200 j = 1, kxp
               f(j,k) = g(j,k)
  200       continue
  210 continue
      return
c ------------------
c forward fourier transform
c scramble coefficients
c copy distributed data
c ------------------
  220 do 240 k = 1, kblok
         kb = k + ks
         kk = kbs - kb
      do 230 j = 2, kxp
         g(j,k) = f(j,kk)
  230 continue
      if (kb.gt.0) g(1,k) = f(1,kk+1)

  240 continue
c ------------------
c perform reduction
c ------------------
      do 270 k = 1, kblok
         kb = k + ks
         koff = kxp*kb - 1
         do 250 j = 2, kxp
            jj = kxp2 - j
            arg = dnx*float(j + koff)
            f(j,k) = (f(j,k) + conjg(g(jj,k))) - 
     1               (f(j,k) - conjg(g(jj,k)))*cmplx(sin(arg),-cos(arg))
  250    continue
         if (kb.eq.0) then
            f(1,k) = cmplx(real(f(1,k)) + aimag(f(1,k)),real(f(1,k)) - 
     1                aimag(f(1,k)))
         else
            arg = dnx*float(1 + koff)
            f(1,k) = (f(1,k) + conjg(g(1,k))) - 
     1               (f(1,k) - conjg(g(1,k)))*cmplx(sin(arg),-cos(arg))
         endif
c copy result to temporary
         do 260 j = 1, kxp
            g(j,k) = f(j,k)
  260    continue
  270 continue
c ----------------------------
c bit-reverse array elements
c ----------------------------
      call bitrv1(g,f,mixup,isign,indx1,kstrt,kxp,kblok)
      do 370 l = 1, indx1
         nxs = 2**(l - 1)
         kxs = nxs/kxp
c -------------------------
c local calculation
c -------------------------
         if (kxs.eq.0) then
c copy local data
            do 320 k = 1, kblok
               do 300 j = 1, kxp
                  lb = (j - 1)/nxs
                  if (lb.eq.(2*(lb/2))) then
                     g(j,k) = f(j+nxs,k)
                  else
                     g(j,k) = f(j-nxs,k)
                  endif
  300          continue
c perform reduction
               km = kxp/nxs
               do 310 j = 1, kxp
                 kb = j - 1
                 lb = kb/nxs
                 t1 = conjg(sct(1+km*(kb-nxs*lb),k))
                 if (lb.eq.(2*(lb/2))) then
                    f(j,k) = f(j,k) + t1*g(j,k)
                 else
                    f(j,k) = g(j,k) - t1*f(j,k)
                 endif
  310          continue
  320       continue
c -------------------------
c distributed calculation
c -------------------------
         else
c copy distributed data
            do 340 k = 1, kblok
               kb = k + ks
               lb = kb/kxs
               kb = kb + 1
               lb = lb - 2*(lb/2)
        do 330 j = 1, kxp
           if (lb.eq.0) then
              g(j,k) = f(j,kb+kxs)
           else
              g(j,k) = f(j,kb-kxs)
           endif
 330    continue


  340    continue
c ------------------
c perform reduction
c ------------------
         km = nxh/nxs
         dns = dnx*float(km)
         do 360 k = 1, kblok
         kb = k + ks
         lb = kb/kxs
         kb = kxp*(kb - kxs*lb) - 1
         lb = lb - 2*(lb/2)
         do 350 j = 1, kxp
         arg = dns*float(j + kb)
         t1 = cmplx(cos(arg),sin(arg))
         if (lb.eq.0) then
            f(j,k) = f(j,k) + t1*g(j,k)
         else
            f(j,k) = g(j,k) - t1*f(j,k)
         endif
  350    continue
  360    continue
      endif
  370 continue
      return
      end
C =========================================================================
c this subroutine performs bit-reversal of distributed complex data,
c that is, it calculates g(j,k) = f(jj,kk), where
c kk = (mixup(j,k)-1)/kxp+1 and jj = mixup(j,k)-kxp*(kk-1), and
c mixup(j,k) = 1 + reversed bits of (j + kxp*(k + kstrt - 2) - 1)
c f = complex input array
c g = complex output array
c mixup = array of bit reversed addresses
c if isign = 0, bit-reverse array is prepared
c if isign = 1, bit-reversal is performed
c indx = power of 2 which determines length of data, nx = 2**indx
c kstrt = starting data block number
c kblok = number of data blocks
c kxp = number of data values per block
c written by viktor k. decyk, ucla
C =========================================================================
      subroutine bitrv1(f,g,mixup,isign,indx,kstrt,kxp,kblok)
      character*8 jbname
      common /pparms/ nproc, nprec, lgrp, jbname
      complex f, g
      dimension f(kxp,kblok), g(kxp,kblok), mixup(kxp,kblok)
      nx = 2**indx
      if (kstrt.gt.nx) return
      ks = kstrt - 2
      if (isign.ne.0) go to 40
c ---------------------------------
c prepare bit-reverse index table
c ---------------------------------
      do 30 k = 1, kblok
         koff = kxp*(k + ks) - 1
         do 20 j = 1, kxp
            lb = j + koff
            ll = 0
            do 10 l = 1, indx
               jb = lb/2
               it = lb - 2*jb
               lb = jb
               ll = 2*ll + it
   10       continue
            mixup(j,k) = ll + 1
   20    continue
   30 continue
      return
c ---------------------------------
c bit-reverse array elements
c ---------------------------------
  40  do 60 k = 1, kblok
      do 50 j = 1, kxp
        ll = mixup(j,k)
        kk = (ll - 1)/kxp + 1
        jj = ll - kxp*(kk - 1)
        g(j,k) = f(jj,kk)
  50  continue
  60  continue

      return
      end

c parallel sum of a vector for pvm (2 versions)
c     INCLUDE 'vector/psum.f'
c     INCLUDE 'vector/pisum.f'

      SUBROUTINE TIMER(T)
C   ==================================================================
C   ===                                                            ===
C   ===    This subroutine returns the elapsed time in seconds     ===
C   ===                    since its first CALL                    ===
C   ===                                                            ===
C   ===                     Fortran90 version                      ===
C   ===                                                            ===
C   ===                     vsg - October, 92                      ===
C   ==================================================================
 
      real T
      real secs, tarray(2)
      secs = etime(tarray)
      t = secs
      return

      END


