
C =======================================================================
c this subroutine performs a two dimensional real to complex fast
c fourier transform and its inverse, using complex arithmetic,
c for data which is distributed in blocks
c indx/indy = exponent which determines length in x/y direction,
c where nx=2**indx, ny=2**indy
c ntpose = (0,1) = (no,yes) input, output data are transposed
c if isign = 0, the fft tables are prepared
c if isign = -1, an inverse fourier transform is performed
c if ntpose = 0, f is the input and output array, g is a scratch array
c f(n,m,l) = (1/nx*ny)*sum(f(j,k,i)*
c       exp(-sqrt(-1)*2pi*n*j/nx)*exp(-sqrt(-1)*2pi*mm*kk/ny))
c where mm = m + kyp*(l - 1) and kk = k + kyp*(i - 1)
c if ntpose = 1, f is the input and g is the output
c g(m,n,l) = (1/nx*ny)*sum(f(j,k,i)*
c       exp(-sqrt(-1)*2pi*nn*j/nx)*exp(-sqrt(-1)*2pi*m*kk/ny))
c where nn = n + kxp*(l - 1) and kk = k + kyp*(i - 1)
c if isign = 1, a forward fourier transform is performed
c if ntpose = 0, f is the input and output array, g is a scratch array
c f(j,k,i) = sum(f(n,m,l)*exp(sqrt(-1)*2pi*n*j/nx)*
c       exp(sqrt(-1)*2pi*mm*kk/ny))
c where mm = m + kyp*(l - 1) and kk = k + kyp*(i - 1)
c if ntpose = 1, g is the input and f is the output
c f(j,k,i) = sum(g(m,n,l)*exp(sqrt(-1)*2pi*nn*j/nx)*
c       exp(sqrt(-1)*2pi*m*kk/ny))
c where nn = n + kxp*(l - 1) and kk = k + kyp*(i - 1)
c g(k,1,1) = mode nx/2,k-1, where ny/2+2 <= k <= ny, and
c imaginary part of g(1,1,1) = real part of mode nx/2,0 and
c imaginary part of g(ny/2+1,1,1) = real part of mode nx/2,ny/2
c written by viktor k. decyk, ucla
c parallel version
C =========================================================================
      subroutine pfft2r(f,g,bs,br,isign,ntpose,mixup,sct,indx,indy,kstrt
     1,nxvh,nyv,kxp,kyp,jblok,kblok,nxhyd,nxyhd)
      complex f, g, bs, br, sct, s, t, t1
      dimension f(nxvh,kyp,kblok), g(nyv,kxp,jblok)
      dimension bs(kxp,kyp,kblok), br(kxp,kyp,jblok)
      dimension mixup(nxhyd), sct(nxyhd)
      indx1 = indx - 1
      indx1y = max0(indx1,indy)
      nx = 2**indx
      nxh = nx/2
      nxhh = nx/4
      nxh2 = nxh + 2
      ny = 2**indy
      nyh = ny/2
      ny2 = ny + 2
      nxy = max0(nx,ny)
      nxhy = 2**indx1y
      ks = kstrt - 2
      if (isign) 50, 10, 300
c ------------------------------------------------------------------
c prepare fft tables
c bit-reverse index table: mixup(j) = 1 + reversed bits of (j - 1)
c ------------------------------------------------------------------
   10 do 30 j = 1, nxhy
      lb = j - 1
      ll = 0
      do 20 k = 1, indx1y
      jb = lb/2
      it = lb - 2*jb
      lb = jb
      ll = 2*ll + it
   20 continue
      mixup(j) = ll + 1
   30 continue
c ---------------------------------------------
c sine/cosine table for the angles 2*n*pi/nxy
c ---------------------------------------------
      nxyh = nxy/2
      dnxy = 6.28318530717959/float(nxy)
      do 40 j = 1, nxyh
      arg = dnxy*float(j - 1)
      sct(j) = cmplx(cos(arg),-sin(arg))
   40 continue
      return
c ----------------------------
c inverse fourier transform
c ----------------------------
   50 if (kstrt.gt.ny) go to 180
      nrx = nxhy/nxh
      do 80 l = 1, kblok
c --------------------------------
c bit-reverse array elements in x
c --------------------------------
      do 70 j = 1, nxh
      j1 = (mixup(j) - 1)/nrx + 1
      if (j.ge.j1) go to 70
      do 60 k = 1, kyp
      t = f(j1,k,l)
      f(j1,k,l) = f(j,k,l)
      f(j,k,l) = t
   60 continue
   70 continue
   80 continue
c -----------------------
c first transform in x
c -----------------------
      nrx = nxy/nxh
      do 130 m = 1, indx1
      ns = 2**(m - 1)
      ns2 = ns + ns
      km = nxhh/ns
      kmr = km*nrx
      do 120 l = 1, kblok
      do 110 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 100 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = sct(1+kmr*(j-1))
      do 90 i = 1, kyp
      t = s*f(j2,i,l)
      f(j2,i,l) = f(j1,i,l) - t
      f(j1,i,l) = f(j1,i,l) + t
   90 continue
  100 continue
  110 continue
  120 continue
  130 continue
c ---------------------------------------
c unscramble coefficients and normalize
c ---------------------------------------
      kmr = nxy/nx
      ani = 1./float(2*nx*ny)
      do 170 l = 1, kblok
      do 150 j = 2, nxhh
      t1 = cmplx(aimag(sct(1+kmr*(j-1))),-real(sct(1+kmr*(j-1))))
      do 140 k = 1, kyp
      t = conjg(f(nxh2-j,k,l))
      s = f(j,k,l) + t
      t = (f(j,k,l) - t)*t1
      f(j,k,l) = ani*(s + t)
      f(nxh2-j,k,l) = ani*conjg(s - t)
  140 continue
  150 continue
      do 160 k = 1, kyp
      f(nxhh+1,k,l) = 2.*ani*conjg(f(nxhh+1,k,l))
      f(1,k,l) = 2.*ani*cmplx(real(f(1,k,l)) + aimag(f(1,k,l)),real(f(1,
     1k,l)) - aimag(f(1,k,l)))
  160 continue
  170 continue
c ------------------------
c transpose f array to g
c ------------------------
  180 call ptpose(f,g,bs,br,nxh,ny,kstrt,nxvh,nyv,kxp,kyp,jblok,kblok)
      if (kstrt.gt.nxh) go to 290
      nry = nxhy/ny
      do 210 l = 1, jblok
c ---------------------------------
c bit-reverse array elements in y
c ---------------------------------
      do 200 k = 1, ny
      k1 = (mixup(k) - 1)/nry + 1
      if (k.ge.k1) go to 200
      do 190 j = 1, kxp
      t = g(k1,j,l)
      g(k1,j,l) = g(k,j,l)
      g(k,j,l) = t

  190 continue
  200 continue
  210 continue
c ---------------------
c then transform in y
c ---------------------
      nry = nxy/ny
      do 260 m = 1, indy
      ns = 2**(m - 1)
      ns2 = ns + ns
      km = nyh/ns
      kmr = km*nry
      do 250 l = 1, jblok
      do 240 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 230 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = sct(1+kmr*(j-1))
      do 220 i = 1, kxp
      t = s*g(j2,i,l)
      g(j2,i,l) = g(j1,i,l) - t
      g(j1,i,l) = g(j1,i,l) + t
  220 continue
  230 continue
  240 continue
  250 continue
  260 continue
c -------------------------------
c unscramble modes kx = 0, nx/2
c -------------------------------
      do 280 l = 1, jblok
      if ((l+ks).gt.0) go to 280
      do 270 k = 2, nyh
c     if ((l+ks).eq.0) then
      s = g(ny2-k,1,l)
      g(ny2-k,1,l) = .5*cmplx(aimag(g(k,1,l) + s),real(g(k,1,l) - s))
      g(k,1,l) = .5*cmplx(real(g(k,1,l) + s),aimag(g(k,1,l) - s))
c     endif
  270 continue
  280 continue
c ------------------------
c transpose g array to f
c ------------------------
  290 if (ntpose.eq.0) call ptpose(g,f,br,bs,ny,nxh,kstrt,nyv,nxvh,kyp,k
     1xp,kblok,jblok)
      return
c ---------------------------
c forward fourier transform
c transpose f array to g
c ---------------------------
  300 if (ntpose.eq.0) call ptpose(f,g,bs,br,nxh,ny,kstrt,nxvh,nyv,kxp,k
     1yp,jblok,kblok)
      if (kstrt.gt.nxh) go to 410
      nry = nxhy/ny
      do 350 l = 1, jblok
c ---------------------------
c scramble modes kx = 0, nx/2
c ---------------------------
      if ((l+ks).gt.0) go to 320
      do 310 k = 2, nyh
      s = cmplx(aimag(g(ny2-k,1,l)),real(g(ny2-k,1,l)))
      g(ny2-k,1,l) = conjg(g(k,1,l) - s)
      g(k,1,l) = g(k,1,l) + s
  310 continue
c ----------------------------------
c bit-reverse array elements in y
c ----------------------------------
  320 do 340 k = 1, ny
      k1 = (mixup(k) - 1)/nry + 1
      if (k.ge.k1) go to 340
      do 330 j = 1, kxp
      t = g(k1,j,l)
      g(k1,j,l) = g(k,j,l)
      g(k,j,l) = t
  330 continue
  340 continue
  350 continue
c ----------------------
c first transform in y
c ----------------------
      nry = nxy/ny
      do 400 m = 1, indy
      ns = 2**(m - 1)
      ns2 = ns + ns
      km = nyh/ns
      kmr = km*nry
      do 390 l = 1, jblok
      do 380 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 370 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = conjg(sct(1+kmr*(j-1)))
      do 360 i = 1, kxp
      t = s*g(j2,i,l)
      g(j2,i,l) = g(j1,i,l) - t
      g(j1,i,l) = g(j1,i,l) + t
  360 continue
  370 continue
  380 continue
  390 continue
  400 continue
c ------------------------
c transpose g array to f
c ------------------------
  410 call ptpose(g,f,br,bs,ny,nxh,kstrt,nyv,nxvh,kyp,kxp,kblok,jblok)
      if (kstrt.gt.ny) return
      nrx = nxhy/nxh
      kmr = nxy/nx
      do 470 l = 1, kblok
c ----------------------
c scramble coefficients
c ----------------------
      do 430 j = 2, nxhh
      t1 = cmplx(aimag(sct(1+kmr*(j-1))),real(sct(1+kmr*(j-1))))
      do 420 k = 1, kyp
      t = conjg(f(nxh2-j,k,l))
      s = f(j,k,l) + t
      t = (f(j,k,l) - t)*t1
      f(j,k,l) = s + t
      f(nxh2-j,k,l) = conjg(s - t)
  420 continue
  430 continue
      do 440 k = 1, kyp
      f(nxhh+1,k,l) = 2.*conjg(f(nxhh+1,k,l))
      f(1,k,l) = cmplx(real(f(1,k,l)) + aimag(f(1,k,l)),real(f(1,k,l)) -
     1 aimag(f(1,k,l)))
  440 continue
c ---------------------------------
c bit-reverse array elements in x
c ---------------------------------
      do 460 j = 1, nxh
      j1 = (mixup(j) - 1)/nrx + 1
      if (j.ge.j1) go to 460
      do 450 k = 1, kyp
      t = f(j1,k,l)
      f(j1,k,l) = f(j,k,l)
      f(j,k,l) = t
  450 continue
  460 continue
  470 continue
c ---------------------
c then transform in x
c ---------------------
      nrx = nxy/nxh
      do 520 m = 1, indx1
      ns = 2**(m - 1)
      ns2 = ns + ns
      km = nxhh/ns
      kmr = km*nrx
      do 510 l = 1, kblok
      do 500 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 490 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = conjg(sct(1+kmr*(j-1)))
      do 480 i = 1, kyp
      t = s*f(j2,i,l)
      f(j2,i,l) = f(j1,i,l) - t
      f(j1,i,l) = f(j1,i,l) + t
  480 continue
  490 continue
  500 continue
  510 continue
  520 continue
      return
      end
C ========================================================================
c this subroutine performs a transpose of a matrix f, distributed in y,
c to a matrix g, distributed in x, that is,
c g(k+kyp*(m-1),j,l) = f(j+kxp*(l-1),k,m), where
c 1 <= j <= kxp, 1 <= k <= kyp, 1 <= l <= nx/kxp, 1 <= m <= ny/kyp
c and where indices l and m can be distributed across processors.
c this subroutine sends and receives one message at a time, either
c synchronously or asynchronously. it uses a minimum of system resources
c f = complex input array
c g = complex output array
c s, t = complex scratch arrays
c nx/ny = number of points in x/y
c kstrt = starting data block number
c nxv/nyv = first dimension of f/g
c kxp/kyp = number of data values per block in x/y
c jblok/kblok = number of data blocks in x/y
c optimized version
C ========================================================================
      subroutine ptpose(f,g,s,t,nx,ny,kstrt,nxv,nyv,kxp,kyp,jblok,kblok)
      character*8 jbname
      common /pparms/ nproc, nprec, lgrp, jbname
      complex f, g, s, t
      dimension f(nxv,kyp,kblok), g(nyv,kxp,jblok)
      dimension s(kxp,kyp,kblok), t(kxp,kyp,jblok)
      ks = kstrt - 2
      kxb = nx/kxp
      kyb = ny/kyp
c this segment is used for shared memory computers
      if (kstrt.gt.nx) return
      do 40 l = 1, jblok
      joff = kxp*(l + ks)
      do 30 i = 1, kyb
      koff = kyp*(i - 1)
      do 20 k = 1, kyp
      do 10 j = 1, kxp
      g(k+koff,j,l) = f(j+joff,k,i)
   10 continue
   20 continue
   30 continue
   40 continue

      return
      end

      SUBROUTINE TIMER(T)
C   ==================================================================
C   ===                                                            ===
C   ===    This subroutine returns the elapsed time in seconds     ===
C   ===                    since its first CALL                    ===
C   ===                                                            ===
C   ===                     Fortran90 version                      ===
C   ===                                                            ===
C   ===                     vsg - October, 92                      ===
C   ==================================================================
 
      real T
      real secs, tarray(2)
      secs = etime(tarray)
      t = secs
      return

      END
