c * * * periodic 3d electrostatic particle simulation kernel code * * *
c this is a simple 3d skeleton particle-in-cell code designed for
c exploring new computer architectures.  it contains the critical pieces
c needed for depositing charge, advancing particles, and solving the
c field.  the code moves only electrons, with periodic electrostatic
c forces obtained by solving poisson's equation with fast fourier
c transforms.  the only diagnostic is particle and field energy.
c portable gcpic kernel code, using algorithm described in:
c p. c. liewer and v. k. decyk, j. computational phys. 85, 302 (1989).
c written by viktor k. decyk, ucla
c for ibm sp1
      program beps3k

      parameter( indx =   4, indy =   3, indz =   5)
c     parameter( indx =   5, indy =   4, indz =   6)
c     parameter( indx =   6, indy =   5, indz =   7)
c     parameter( indx =   7, indy =   6, indz =   8)

      parameter( npx =      24, npy =      12, npz =     48)
c     parameter( npx =      64, npy =      32, npz =    128)
c     parameter( npx =     192, npy =      96, npz =    384)
c     parameter( npx =     512, npy =     256, npz =   1024)

      parameter( npxb =  12, npyb =   6, npzb =  24)
c     parameter( npxb =  32, npyb =  16, npzb =  64)
c     parameter( npxb =  96, npyb =  48, npzb = 192)
c     parameter( npxb = 256, npyb = 128, npzb = 512)

      parameter( tend =  85.000, dt = 0.2000000e+00)
      parameter( vtz =   1.000, vdz =   5.000, vtdz =   0.500)
      parameter( avdz =   5.000)
      parameter( indnvp =   2, idps =    2, idimp =   6, mshare =   1)
      parameter(npxyz=npx*npy*npz,npxyzb=npxb*npyb*npzb,np=npxyz+npxyzb)
      parameter(nx=2**indx,ny=2**indy,nz=2**indz)
      parameter(nxh=nx/2,nyh=ny/2,nzh=nz/2)
      parameter(nxv=nx+2,nxvh=nxv/2,nyv=ny+2,nzv=nz+2)
      parameter(nloop=tend/dt+.0001)
      parameter(nvp=2**indnvp,nblok=1+mshare*(nvp-1))
      parameter(npmax=(np/nvp)*1.01+60000,nzpmx=(nz-1)/nvp+4)
      parameter(nxyvzp=nxv*nyv*nzpmx)
      parameter(kzp=(nz-1)/nvp+1,kxp=(nxh-1)/nvp+1)
      parameter(lblok=1+mshare*(nz/kzp-1),jblok=1+mshare*(nxh/kxp-1))
      parameter(nmyz=ny*(nz/ny)+nz*(ny/nz),nyz=nmyz/(2-ny/nmyz-nz/nmyz))
      parameter(nmx=nx*(nyz/nx)+nyz*(nx/nyz))
      parameter(nxyz=nmx/(2-nx/nmx-nyz/nmx),nxyzh=nxyz/2)
      parameter(nmxh=nxh*(nyz/nxh)+nyz*(nxh/nyz))
      parameter(nxhyz=nmxh/(2-nxh/nmxh-nyz/nmxh))
      parameter(nbmax=1+(2*(npxyz*vtz+npxyzb*vtdz)+1.4*npxyzb*avdz)*dt/n
     1z)
      parameter(ntmax=2*nbmax)
      parameter(npd=128,n27=27)
      complex qc, fxc, fyc, fzc, qt, fxt, fyt, fzt
      complex ffc, bs, br, scr, sct
      common /large/ part
      dimension part(idimp,npmax,nblok)
      dimension maskp(npmax,nblok)
      dimension q(nxv,nyv,nzpmx,nblok)
      dimension fx(nxv,nyv,nzpmx,nblok), fy(nxv,nyv,nzpmx,nblok)
      dimension fz(nxv,nyv,nzpmx,nblok)
      dimension qc(nxvh,nyv,kzp,lblok)
      dimension fxc(nxvh,nyv,kzp,lblok), fyc(nxvh,nyv,kzp,lblok)
      dimension fzc(nxvh,nyv,kzp,lblok)
      dimension qt(nzv,nyv,kxp,jblok)
      dimension fxt(nzv,nyv,kxp,jblok), fyt(nzv,nyv,kxp,jblok)
      dimension fzt(nzv,nyv,kxp,jblok)
      dimension ffc(nzh,nyh,kxp,jblok)
      dimension mixup(nxhyz), sct(nxyzh)
      dimension bs(kxp,ny,kzp,lblok), br(kxp,ny,kzp,jblok)
      dimension edges(idps,nblok)
      dimension nzp(nblok), noff(nblok)
      dimension npp(nblok), nps(nblok)
      dimension sbufl(idimp,nbmax,nblok), sbufr(idimp,nbmax,nblok)
      dimension rbufl(idimp,nbmax,nblok), rbufr(idimp,nbmax,nblok)
      dimension ihole(ntmax,nblok)
      dimension jsl(idps,nblok), jsr(idps,nblok)
      dimension jss(idps,nblok), scr(nxvh,nyv,idps,nblok)
      dimension nn(npd,n27,nblok), amxyz(npd,n27,nblok)
      dimension dx(npd,nblok), dy(npd,nblok), dz(npd,nblok)


  991 format (5h t = ,i7)
  992 format (19h * * * q.e.d. * * *)
  993 format (34h field, kinetic, total energies = ,3e14.7)
      data vtx,vty,qme,vdx,vdy,vtdx,vtdy /1.,1.,-1.,0.,0.,.5,.5/
      data ax,ay,az /.866667,.866667,.866667/
      data ntpose /1/

      kstrt = 1
c      if (kstrt.eq.1) then
c         open(unit=6,file='output3',form='formatted',status='unknown')
c      endif
c initialize timer

c initialize constants
      itime = 0
      qtme = qme*dt
      affp = float(nx*ny*nz)/float(np)
      zero = 0.
c calculate partition variables
       print *,'begin...',elapsed_time() 
      call dcomp3(edges,nzp,noff,nz,kstrt,nvp,idps,nblok)
      print *,'after pdost2',elapsed_time()

c prepare fft tables
      isign = 0
      call pfft3r(qc,qt,bs,br,isign,ntpose,mixup,sct,indx,indy,indz,kstr
     1t,nxvh,nyv,nzv,kxp,kzp,jblok,lblok,ny,nxhyz,nxyzh)
      print *,'after pfft3',elapsed_time()

c calculate form factors
      call ppois3 (qt,fxt,fyt,fzt,isign,ffc,ax,ay,az,affp,we,nx,ny,nz,ks
     1trt,nyv,nzv,kxp,jblok,nyh,nzh)
      print *,'after pppis3 ',elapsed_time()

c initialize density profile and velocity distribution
c background electrons
      do 140 m = 1, nblok
      nps(m) = 1
      npp(m) = 0
      nzp3 = nzp(m) + 3
      do 130 l = 1, nzp3
      do 120 k = 1, ny
      do 110 j = 1, nx
      q(j,k,l,m) = 0.
  110 continue
  120 continue
  130 continue
  140 continue
      if (npxyz.gt.0) call pistr3 (part,edges,npp,nps,vtx,vty,vtz,zero,z
     1ero,zero,npx,npy,npz,nx,ny,nz,idimp,npmax,nblok,idps)
      print *,'after pistr3 ',elapsed_time()

c beam electrons
      do 150 m = 1, nblok
      nps(m) = npp(m) + 1
  150 continue
      if (npxyzb.gt.0) call pistr3 (part,edges,npp,nps,vtdx,vtdy,vtdz,vd
     1x,vdy,vdz,npxb,npyb,npzb,nx,ny,nz,idimp,npmax,nblok,idps)
      print *,'after pistr3',elapsed_time()

c deposit charge for initial distribution
      call pdost3(part,q,npp,noff,qme,nx,ny,idimp,npmax,nblok,nxv,nyv,nz
     1pmx)
      print *,'after pdost3',elapsed_time()

c add background ion density
      qi0 = -qme/affp
      do 250 m = 1, nblok
      do 240 l = 1, nzp(m)
      do 230 k = 1, ny
      do 220 j = 1, nx
      q(j,k,l+1,m) = q(j,k,l+1,m) + qi0
  220 continue
  230 continue
  240 continue
  250 continue
c
c * * * start main iteration loop * * *
c

  500 if (nloop.le.itime) go to 2000
      if (kstrt.eq.1) write (6,991) itime
      isign = -1
c copy data from particle to field partition, and add up guard cells
      call cppfp3 (q,qc,isign,scr,kstrt,nvp,nx,ny,nxv,nyv,nzpmx,nblok,nx
     1vh,kzp,lblok,idps)
      print *,'after cppfp3',elapsed_time()

c transform charge to fourier space
      call pfft3r(qc,qt,bs,br,isign,ntpose,mixup,sct,indx,indy,indz,kstr
     1t,nxvh,nyv,nzv,kxp,kzp,jblok,lblok,ny,nxhyz,nxyzh)
      print *,'after pfft3r',elapsed_time()

c calculate force/charge in fourier space
      call ppois3 (qt,fxt,fyt,fzt,isign,ffc,ax,ay,az,affp,we,nx,ny,nz,ks
     1trt,nyv,nzv,kxp,jblok,nyh,nzh)
      print *,'after ppois3',elapsed_time()

c transform force/charge to real space
      isign = 1
      call pfft3r(fxc,fxt,bs,br,isign,ntpose,mixup,sct,indx,indy,indz,ks
     1trt,nxvh,nyv,nzv,kxp,kzp,jblok,lblok,ny,nxhyz,nxyzh)
      print *,'after pfft3r',elapsed_time()
      call pfft3r(fyc,fyt,bs,br,isign,ntpose,mixup,sct,indx,indy,indz,ks
     1trt,nxvh,nyv,nzv,kxp,kzp,jblok,lblok,ny,nxhyz,nxyzh)
      print *,'after pfft3r',elapsed_time()
      call pfft3r(fzc,fzt,bs,br,isign,ntpose,mixup,sct,indx,indy,indz,ks
     1trt,nxvh,nyv,nzv,kxp,kzp,jblok,lblok,ny,nxhyz,nxyzh)
      print *,'after pfft3r',elapsed_time()

c copy data from field to particle partition, and copy to guard cells
      call cppfp3 (fx,fxc,isign,scr,kstrt,nvp,nx,ny,nxv,nyv,nzpmx,nblok,
     1nxvh,kzp,lblok,idps)
      print *,'after cppfp3',elapsed_time() 
      call cppfp3 (fy,fyc,isign,scr,kstrt,nvp,nx,ny,nxv,nyv,nzpmx,nblok,
     1nxvh,kzp,lblok,idps)
      print *,'after cppfp3',elapsed_time() 
      call cppfp3 (fz,fzc,isign,scr,kstrt,nvp,nx,ny,nxv,nyv,nzpmx,nblok,
     1nxvh,kzp,lblok,idps)
      print *,'after cppfp3',elapsed_time() 
c particle push and charge density update

c initialize charge density to zero
      do 1260 m = 1, nblok
      nzp3 = nzp(m) + 3
      do 1250 l = 1, nzp3
      do 1240 k = 1, ny
      do 1230 j = 1, nx
      q(j,k,l,m) = 0.
 1230 continue
 1240 continue
 1250 continue
 1260 continue
      wke = 0.

c push particles
      call ppush3 (part,fx,fy,fz,npp,noff,qtme,dt,wke,nx,ny,idimp,npmax,
     1nblok,nxv,nyv,nzpmx)
      print *,'after ppush3',elapsed_time() 

c move particles into appropriate spatial regions
      call pmove3 (part,edges,npp,sbufr,sbufl,rbufr,rbufl,ihole,jsr,jsl,
     1jss,nz,kstrt,nvp,idimp,npmax,nblok,idps,nbmax,ntmax,ierr)
      print *,'after pmove3',elapsed_time() 

c deposit charge
      call pdost3(part,q,npp,noff,qme,nx,ny,idimp,npmax,nblok,nxv,nyv,nz
     1pmx)
      print *,'after pdost3',elapsed_time()  

c add background charge
      do 1300 m = 1, nblok
      do 1290 l = 1, nzp(m)
      do 1280 k = 1, ny
      do 1270 j = 1, nx
      q(j,k,l+1,m) = q(j,k,l+1,m) + qi0
 1270 continue
 1280 continue
 1290 continue
 1300 continue

c energy diagnostic
      wt = we + wke
      if (kstrt.eq.1) write (6,993) we, wke, wt
      itime = itime + 1
      go to 500
 2000 continue
c
c * * * end main iteration loop * * *
c
      if (kstrt.eq.1) write (6,992)
      stop
      end
       
C--------------------------------------------------------------------------
C--------------------------------------------------------------------------
  
      real function elapsed_time()
      integer :: clock_count, clock_rate, clock_max
      integer, save :: last_count
      logical, save :: system_clock_not_yet_called = .true.
 
      call system_clock(count=clock_count,
     1                    count_rate=clock_rate,
     1                  count_max=clock_max)
      if (system_clock_not_yet_called) then
         system_clock_not_yet_called = .false.
         elapsed_time = 0.0
      else if (clock_count > last_count) then
         elapsed_time = real(clock_count-last_count)/real(clock_rate)
      else
         elapsed_time = real(clock_max-last_count+clock_count)
     1       /real(clock_rate)
      end if
      last_count = clock_count
      return
      end
C--------------------------------------------------------------------------
C--------------------------------------------------------------------------
  


      subroutine dcomp3(edges,nzp,noff,nz,kstrt,nvp,idps,nblok)

      dimension edges(idps,nblok)
      dimension nzp(nblok), noff(nblok)
      ks = kstrt - 2
      at1 = float(nz)/float(nvp)
      do 10 m = 1, nblok
      kb = m + ks
      edges(1,m) = at1*float(kb)
      noff(m) = edges(1,m) + .5
      edges(2,m) = at1*float(kb + 1)
      kr = edges(2,m) + .5
      nzp(m) = kr - noff(m)
   10 continue
      return
      end


      subroutine pistr3 (part,edges,npp,nps,vtx,vty,vtz,vdx,vdy,vdz,npx,
     1npy,npz,nx,ny,nz,idimp,npmax,nblok,idps)
c for 3d code, this subroutine calculates initial particle co-ordinates
c and velocities with uniform density and maxwellian velocity with drift
c for distributed data

      character*8 jbname
      double precision ranorm
      common /pparms/ nproc, nprec, lgrp, jbname
      external d_vadd
      dimension part(idimp,npmax,nblok)
      dimension edges(idps,nblok), npp(nblok), nps(nblok)
      dimension sum3(3), work3(3)
      npxyz = npx*npy*npz
      at1 = float(nx)/float(npx)
      at2 = float(ny)/float(npy)
      at3 = float(nz)/float(npz)
      do 40 l = 1, npz
      zt = at3*(float(l) - .5)
      do 30 k = 1, npy
      yt = at2*(float(k) - .5)
      do 20 j = 1, npx
c uniform density profile
      xt = at1*(float(j) - .5)
c maxwellian velocity distribution
      vxt = vtx*ranorm(d)
      vyt = vty*ranorm(d)
      vzt = vtz*ranorm(d)
      do 10 m = 1, nblok
      if ((zt.ge.edges(1,m)).and.(zt.lt.edges(2,m))) then
         npt = npp(m) + 1
         part(1,npt,m) = xt
         part(2,npt,m) = yt
         part(3,npt,m) = zt
         part(4,npt,m) = vxt
         part(5,npt,m) = vyt
         part(6,npt,m) = vzt
         npp(m) = npt
      endif
   10 continue
   20 continue
   30 continue
   40 continue
c add correct drift
      sum3(1) = 0.
      sum3(2) = 0.
      sum3(3) = 0.
      do 60 m = 1, nblok
      sum0 = 0.
      sum1 = 0.
      sum2 = 0.
      do 50 j = nps(m), npp(m)
c     do 50 j = 1, npmax
c     if ((j.ge.nps(m)).and.(j.le.npp(m))) then
      sum0 = sum0 + part(4,j,m)
      sum1 = sum1 + part(5,j,m)
      sum2 = sum2 + part(6,j,m)
c     endif
   50 continue
      sum3(1) = sum3(1) + sum0
      sum3(2) = sum3(2) + sum1
      sum3(3) = sum3(3) + sum2
   60 continue

      at1 = 1./float(npxyz)
      sum3(1) = at1*sum3(1) - vdx
      sum3(2) = at1*sum3(2) - vdy
      sum3(3) = at1*sum3(3) - vdz
      do 80 m = 1, nblok
      do 70 j = nps(m), npp(m)
c     do 70 j = 1, npmax
c     if ((j.ge.nps(m)).and.(j.le.npp(m))) then
      part(4,j,m) = part(4,j,m) - sum3(1)
      part(5,j,m) = part(5,j,m) - sum3(2)
      part(6,j,m) = part(6,j,m) - sum3(3)
c     endif
   70 continue
   80 continue
      return
      end


      subroutine pdost3(part,q,npp,noff,qm,nx,ny,idimp,npmax,nblok,nxv,n
     1yv,nzpmx)
c for 3d code, this subroutine calculates particle charge density
c using second-order spline interpolation, periodic boundaries
c and distributed data.

      dimension part(idimp,npmax,nblok), q(nxv,nyv,nzpmx,nblok)
      dimension npp(nblok), noff(nblok)
      do 20 m = 1, nblok
      lnoff = noff(m) - 2
c find interpolation weights
      do 10 j = 1, npp(m)
c     do 10 j = 1, npmax
c     if (j.le.npp(m)) then
      nn = part(1,j,m) + .5
      dxl = part(1,j,m) - float(nn)
      nn = nn + 1
      if (nn.gt.nx) nn = nn - nx
      amx = qm*(.75 - dxl*dxl)
      np = nn + 1
      if (np.gt.nx) np = np - nx
      dxp = .5*qm*(.5 + dxl)**2
      nl = nn - 1
      if (nl.lt.1) nl = nl + nx
      dxl = .5*qm*(.5 - dxl)**2
      mm = part(2,j,m) + .5
      dyl = part(2,j,m) - float(mm)
      mm = mm + 1
      if (mm.gt.ny) mm = mm - ny
      amy = .75 - dyl*dyl
      mp = mm + 1
      if (mp.gt.ny) mp = mp - ny
      dyp = .5*(.5 + dyl)**2
      ml = mm - 1
      if (ml.lt.1) ml = ml + ny
      dyl = .5*(.5 - dyl)**2
      ll = part(3,j,m) + .5
      dzl = part(3,j,m) - float(ll)
      ll = ll - lnoff
      amz = .75 - dzl*dzl
      lp = ll + 1
      dzp = .5*(.5 + dzl)**2
      lm = ll - 1
      dzl = .5*(.5 - dzl)**2
c deposit charge
      q(nn,mm,ll,m) = q(nn,mm,ll,m) + amx*amy*amz
      q(np,mm,ll,m) = q(np,mm,ll,m) + dxp*amy*amz
      q(nl,mm,ll,m) = q(nl,mm,ll,m) + dxl*amy*amz
      q(nn,mp,ll,m) = q(nn,mp,ll,m) + amx*dyp*amz
      q(np,mp,ll,m) = q(np,mp,ll,m) + dxp*dyp*amz
      q(nl,mp,ll,m) = q(nl,mp,ll,m) + dxl*dyp*amz
      q(nn,ml,ll,m) = q(nn,ml,ll,m) + amx*dyl*amz
      q(np,ml,ll,m) = q(np,ml,ll,m) + dxp*dyl*amz
      q(nl,ml,ll,m) = q(nl,ml,ll,m) + dxl*dyl*amz
      q(nn,mm,lp,m) = q(nn,mm,lp,m) + amx*amy*dzp
      q(np,mm,lp,m) = q(np,mm,lp,m) + dxp*amy*dzp
      q(nl,mm,lp,m) = q(nl,mm,lp,m) + dxl*amy*dzp
      q(nn,mp,lp,m) = q(nn,mp,lp,m) + amx*dyp*dzp
      q(np,mp,lp,m) = q(np,mp,lp,m) + dxp*dyp*dzp
      q(nl,mp,lp,m) = q(nl,mp,lp,m) + dxl*dyp*dzp
      q(nn,ml,lp,m) = q(nn,ml,lp,m) + amx*dyl*dzp
      q(np,ml,lp,m) = q(np,ml,lp,m) + dxp*dyl*dzp
      q(nl,ml,lp,m) = q(nl,ml,lp,m) + dxl*dyl*dzp
      q(nn,mm,lm,m) = q(nn,mm,lm,m) + amx*amy*dzl
      q(np,mm,lm,m) = q(np,mm,lm,m) + dxp*amy*dzl
      q(nl,mm,lm,m) = q(nl,mm,lm,m) + dxl*amy*dzl
      q(nn,mp,lm,m) = q(nn,mp,lm,m) + amx*dyp*dzl
      q(np,mp,lm,m) = q(np,mp,lm,m) + dxp*dyp*dzl
      q(nl,mp,lm,m) = q(nl,mp,lm,m) + dxl*dyp*dzl
      q(nn,ml,lm,m) = q(nn,ml,lm,m) + amx*dyl*dzl
      q(np,ml,lm,m) = q(np,ml,lm,m) + dxp*dyl*dzl
      q(nl,ml,lm,m) = q(nl,ml,lm,m) + dxl*dyl*dzl
c     endif
   10 continue
   20 continue
      return
      end

 
      subroutine cppfp3 (f,fc,isign,scr,kstrt,nvp,nx,ny,nxv,nyv,nzpmx,nb
     1lok,nxvh,kzp,lblok,idps)

      common /pparms/ nproc, nprec, lgrp, jbname
      complex fc, scr
      dimension f(nxv,nyv,nzpmx,nblok), fc(nxvh,nyv,kzp,lblok)
      dimension scr(nxvh,nyv,idps,nblok)
      nxh = nx/2
      nxyv = nxv*nyv
      nxyvh = nxvh*nyv
      ks = kstrt - 2
      moff = nzpmx*nvp
      if (isign) 10, 150, 160
c copy from particle to field partition
   10 do 50 m = 1, lblok
      do 40 l = 1, kzp
      do 30 k = 1, ny
      do 20 j = 1, nxh
      fc(j,k,l,m) = cmplx(f(2*j-1,k,l+1,m),f(2*j,k,l+1,m))
   20 continue
   30 continue
   40 continue
   50 continue
c add guard cells
      do 140 m = 1, lblok
      kr = m + ks + 2
      if (kr.gt.nvp) then
         kr = kr - nvp
      endif
      krr = kr
      kl = m + ks
      if (kl.lt.1) then
         kl = kl + nvp
      endif
      kll = kl
      ngc = 2
c special case of only one grid per processor
      if (kzp.eq.1) then
         krr = krr + 1
         if (krr.gt.nvp) then
            krr = krr - nvp
         endif
         kll = kll - 1
         if (kll.lt.1) then
            kll = kll + nvp
         endif
         ngc = 1
      endif
c first add guard cells from right
c this segment is used for shared memory computers
      do 70 k = 1, ny
      do 60 j = 1, nxh
      scr(j,k,1,m) = cmplx(f(2*j-1,k,kzp+2,kl),f(2*j,k,kzp+2,kl))
      scr(j,k,2,m) = cmplx(f(2*j-1,k,kzp+3,kll),f(2*j,k,kzp+3,kll))
   60 continue
   70 continue

      do 90 k = 1, ny
      do 80 j = 1, nxh
      fc(j,k,1,m) = fc(j,k,1,m) + scr(j,k,1,m)
      fc(j,k,ngc,m) = fc(j,k,ngc,m) + scr(j,k,2,m)
   80 continue
   90 continue
c then add guard cells from left
c this segment is used for shared memory computers
      do 110 k = 1, ny
      do 100 j = 1, nxh
      scr(j,k,1,m) = cmplx(f(2*j-1,k,1,kr),f(2*j,k,1,kr))
  100 continue
  110 continue
      do 130 k = 1, ny
      do 120 j = 1, nxh
      fc(j,k,kzp,m) = fc(j,k,kzp,m) + scr(j,k,1,m)
  120 continue
  130 continue
  140 continue
  150 return
c copy from field to particle partition
  160 do 200 m = 1, lblok
      do 190 l = 1, kzp
      do 180 k = 1, ny
      do 170 j = 1, nxh
      f(2*j-1,k,l+1,m) = real(fc(j,k,l,m))
      f(2*j,k,l+1,m) = aimag(fc(j,k,l,m))
  170 continue
  180 continue
  190 continue
  200 continue
c copy to guard cells
      do 250 m = 1, lblok
      kr = m + ks + 2
      if (kr.gt.nvp) then
         kr = kr - nvp
      endif
      krr = kr
      kl = m + ks
      if (kl.lt.1) then
         kl = kl + nvp
      endif
      kll = kl
      ngc = 2
c special case of only one grid per processor
      if (kzp.eq.1) then
         krr = krr + 1
         if (krr.gt.nvp) then
            krr = krr - nvp
         endif
         kll = kll - 1
         if (kll.lt.1) then
            kll = kll + nvp
         endif
         ngc = 1
      endif
c copy to left guard cell
c this segment is used for shared memory computers
      do 220 k = 1, ny
      do 210 j = 1, nxh
      f(2*j-1,k,1,m) = real(fc(j,k,kzp,kl))
      f(2*j,k,1,m) = aimag(fc(j,k,kzp,kl))
  210 continue
  220 continue
c copy to right guard cells
c this segment is used for shared memory computers
      do 240 k = 1, ny
      do 230 j = 1, nxh
      f(2*j-1,k,kzp+2,m) = real(fc(j,k,1,kr))
      f(2*j,k,kzp+2,m) = aimag(fc(j,k,1,kr))
      f(2*j-1,k,kzp+3,m) = real(fc(j,k,ngc,krr))
      f(2*j,k,kzp+3,m) = aimag(fc(j,k,ngc,krr))
  230 continue
  240 continue
  250 continue
      return
      end


      subroutine ppois3(q,fx,fy,fz,isign,ffc,ax,ay,az,affp,wp,nx,ny,nz,k
     1strt,nyv,nzv,kxp,jblok,nyhd,nzhd)
c this subroutine solves 3d poisson's equation in fourier space for
c force/charge (or convolution of electric field over particle shape)
c with periodic boundary conditions, for distributed data.

      character*8 jbname
      common /pparms/ nproc, nprec, lgrp, jbname
      complex q, fx, fy, fz, ffc, zero
      external d_vadd
      dimension q(nzv,nyv,kxp,jblok)
      dimension fx(nzv,nyv,kxp,jblok), fy(nzv,nyv,kxp,jblok)
      dimension fz(nzv,nyv,kxp,jblok)
      dimension ffc(nzhd,nyhd,kxp,jblok)
      nxh = nx/2
      nyh = ny/2
      nzh = nz/2
      ny2 = ny + 2
      nz2 = nz + 2
      ks = kstrt - 2
      dnx = 6.28318530717959/float(nx)
      dny = 6.28318530717959/float(ny)
      dnz = 6.28318530717959/float(nz)
      zero = cmplx(0.,0.)
      if (isign.ne.0) go to 50
      if (kstrt.gt.nxh) return
c prepare form factor array

      do 40 m = 1, jblok
      joff = kxp*(m + ks) - 1
      do 30 j = 1, kxp
      dkx = dnx*float(j + joff)
      at1 = dkx*dkx
      at2 = (dkx*ax)**2
      do 20 k = 1, nyh
      dky = dny*float(k - 1)
      at3 = dky*dky + at1
      at4 = (dky*ay)**2 + at2
      do 10 l = 1, nzh
      dkz = dnz*float(l - 1)
      at5 = dkz*dkz + at3
      at6 = exp(-.5*((dkz*az)**2 + at4))
      if (at5.eq.0.) then
         ffc(l,k,j,m) = cmplx(0.,1.)
      else
         ffc(l,k,j,m) = cmplx(affp*at6/at5,at6)
      endif
   10 continue
   20 continue
   30 continue
   40 continue
      return
c calculate force/charge and sum field energy
   50 wp = 0.
      if (kstrt.gt.nxh) go to 160
      do 150 m = 1, jblok
c mode numbers 0 < kx < nx/2, 0 < ky < ny/2, and 0 < kz < nz/2
      joff = kxp*(m + ks) - 1
      do 100 j = 1, kxp
      dkx = dnx*float(j + joff)
      if ((j+joff).gt.0) then
         do 70 k = 2, nyh
         k1 = ny2 - k
         dky = dny*float(k - 1)
         do 60 l = 2, nzh
         l1 = nz2 - l
         at1 = real(ffc(l,k,j,m))*aimag(ffc(l,k,j,m))
         at2 = dkx*at1
         at3 = dky*at1
         at4 = dnz*float(l - 1)*at1
         fx(l,k,j,m) = at2*cmplx(aimag(q(l,k,j,m)),-real(q(l,k,j,m)))
         fx(l,k1,j,m) = at2*cmplx(aimag(q(l,k1,j,m)),-real(q(l,k1,j,m)))
         fx(l1,k,j,m) = at2*cmplx(aimag(q(l1,k,j,m)),-real(q(l1,k,j,m)))
         fx(l1,k1,j,m) = at2*cmplx(aimag(q(l1,k1,j,m)),-real(q(l1,k1,j,m
     1)))
         fy(l,k,j,m) = at3*cmplx(aimag(q(l,k,j,m)),-real(q(l,k,j,m)))
         fy(l,k1,j,m) = at3*cmplx(-aimag(q(l,k1,j,m)),real(q(l,k1,j,m)))
         fy(l1,k,j,m) = at3*cmplx(aimag(q(l1,k,j,m)),-real(q(l1,k,j,m)))
         fy(l1,k1,j,m) = at3*cmplx(-aimag(q(l1,k1,j,m)),real(q(l1,k1,j,m
     1)))
         fz(l,k,j,m) = at4*cmplx(aimag(q(l,k,j,m)),-real(q(l,k,j,m)))
         fz(l,k1,j,m) = at4*cmplx(aimag(q(l,k1,j,m)),-real(q(l,k1,j,m)))
         fz(l1,k,j,m) = at4*cmplx(-aimag(q(l1,k,j,m)),real(q(l1,k,j,m)))
         fz(l1,k1,j,m) = at4*cmplx(-aimag(q(l1,k1,j,m)),real(q(l1,k1,j,m
     1)))
         wp = wp + at1*(q(l,k,j,m)*conjg(q(l,k,j,m)) + q(l,k1,j,m)*conjg
     1(q(l,k1,j,m)) + q(l1,k,j,m)*conjg(q(l1,k,j,m)) + q(l1,k1,j,m)*conj
     2g(q(l1,k1,j,m)))
   60    continue
   70    continue

c mode numbers kz = 0, nz/2
         l1 = nzh + 1
         do 80 k = 2, nyh
         k1 = ny2 - k
         at1 = real(ffc(1,k,j,m))*aimag(ffc(1,k,j,m))
         at2 = dkx*at1
         at3 = dny*float(k - 1)*at1
         fx(1,k,j,m) = at2*cmplx(aimag(q(1,k,j,m)),-real(q(1,k,j,m)))
         fx(1,k1,j,m) = at2*cmplx(aimag(q(1,k1,j,m)),-real(q(1,k1,j,m)))
         fx(l1,k,j,m) = zero
         fx(l1,k1,j,m) = zero
         fy(1,k,j,m) = at3*cmplx(aimag(q(1,k,j,m)),-real(q(1,k,j,m)))
         fy(1,k1,j,m) = at3*cmplx(-aimag(q(1,k1,j,m)),real(q(1,k1,j,m)))
         fy(l1,k,j,m) = zero
         fy(l1,k1,j,m) = zero
         fz(1,k,j,m) = zero
         fz(1,k1,j,m) = zero
         fz(l1,k,j,m) = zero
         fz(l1,k1,j,m) = zero
         wp = wp + at1*(q(1,k,j,m)*conjg(q(1,k,j,m)) + q(1,k1,j,m)*conjg
     1(q(1,k1,j,m)))
   80    continue
c mode numbers ky = 0, ny/2
         k1 = nyh + 1
         do 90 l = 2, nzh
         l1 = nz2 - l
         at1 = real(ffc(l,1,j,m))*aimag(ffc(l,1,j,m))
         at2 = dkx*at1
         at4 = dnz*float(l - 1)*at1
         fx(l,1,j,m) = at2*cmplx(aimag(q(l,1,j,m)),-real(q(l,1,j,m)))
         fx(l,k1,j,m) = zero
         fx(l1,1,j,m) = at2*cmplx(aimag(q(l1,1,j,m)),-real(q(l1,1,j,m)))
         fx(l1,k1,j,m) = zero
         fy(l,1,j,m) = zero
         fy(l,k1,j,m) = zero
         fy(l1,1,j,m) = zero
         fy(l1,k1,j,m) = zero
         fz(l,1,j,m) = at4*cmplx(aimag(q(l,1,j,m)),-real(q(l,1,j,m)))
         fz(l,k1,j,m) = zero
         fz(l1,1,j,m) = at4*cmplx(-aimag(q(l1,1,j,m)),real(q(l1,1,j,m)))
         fz(l1,k1,j,m) = zero
         wp = wp + at1*(q(l,1,j,m)*conjg(q(l,1,j,m)) + q(l1,1,j,m)*conjg
     1(q(l1,1,j,m)))
   90    continue
      endif
c mode numbers kz = 0, nz/2
      k1 = nyh + 1
      l1 = nzh + 1
      at1 = real(ffc(1,1,j,m))*aimag(ffc(1,1,j,m))
      at2 = dkx*at1
      fx(1,1,j,m) = at2*cmplx(aimag(q(1,1,j,m)),-real(q(1,1,j,m)))
      fx(1,k1,j,m) = zero
      fx(l1,1,j,m) = zero
      fx(l1,k1,j,m) = zero
      fy(1,1,j,m) = zero
      fy(1,k1,j,m) = zero
      fy(l1,1,j,m) = zero
      fy(l1,k1,j,m) = zero
      fz(1,1,j,m) = zero
      fz(1,k1,j,m) = zero
      fz(l1,1,j,m) = zero
      fz(l1,k1,j,m) = zero
      wp = wp + at1*(q(1,1,j,m)*conjg(q(1,1,j,m)))
  100 continue
c mode numbers kx = 0, nx/2
      if ((m+ks).eq.0) then
         do 120 k = 2, nyh
         k1 = ny2 - k
         dky = dny*float(k - 1)
         do 110 l = 2, nzh
c        if ((m+ks).eq.0) then
         l1 = nz2 - l
         at1 = real(ffc(l,k,1,m))*aimag(ffc(l,k,1,m))
         at3 = dky*at1
         at4 = dnz*float(l - 1)*at1
         fx(l,k,1,m) = zero
         fx(l,k1,1,m) = zero
         fx(l1,k,1,m) = zero
         fx(l1,k1,1,m) = zero
         fy(l,k,1,m) = at3*cmplx(aimag(q(l,k,1,m)),-real(q(l,k,1,m)))
         fy(l,k1,1,m) = zero
         fy(l1,k,1,m) = at3*cmplx(aimag(q(l1,k,1,m)),-real(q(l1,k,1,m)))
         fy(l1,k1,1,m) = zero
         fz(l,k,1,m) = at4*cmplx(aimag(q(l,k,1,m)),-real(q(l,k,1,m)))
         fz(l,k1,1,m) = zero
         fz(l1,k,1,m) = at4*cmplx(-aimag(q(l1,k,1,m)),real(q(l1,k,1,m)))
         fz(l1,k1,1,m) = zero
         wp = wp + at1*(q(l,k,1,m)*conjg(q(l,k,1,m)) + q(l1,k,1,m)*conjg
     1(q(l1,k,1,m)))
c        endif
  110    continue
  120    continue
c mode numbers kz = 0, nz/2
         l1 = nzh + 1
         do 130 k = 2, nyh
         k1 = ny2 - k
         at1 = real(ffc(1,k,1,m))*aimag(ffc(1,k,1,m))
         at3 = dny*float(k - 1)*at1
         fx(1,k,1,m) = zero
         fx(1,k1,1,m) = zero
         fx(l1,k,1,m) = zero
         fx(l1,k1,1,m) = zero
         fy(1,k,1,m) = at3*cmplx(aimag(q(1,k,1,m)),-real(q(1,k,1,m)))
         fy(1,k1,1,m) = zero
         fy(l1,k,1,m) = zero
         fy(l1,k1,1,m) = zero
         fz(1,k,1,m) = zero
         fz(1,k1,1,m) = zero
         fz(l1,k,1,m) = zero
         fz(l1,k1,1,m) = zero
         wp = wp + at1*(q(1,k,1,m)*conjg(q(1,k,1,m)))
  130    continue
c mode numbers ky = 0, ny/2
         k1 = nyh + 1
         do 140 l = 2, nzh
         l1 = nz2 - l
         at1 = real(ffc(l,1,1,m))*aimag(ffc(l,1,1,m))
         at4 = dnz*float(l - 1)*at1
         fx(l,1,1,m) = zero
         fx(l,k1,1,m) = zero
         fx(l1,1,1,m) = zero
         fx(l1,k1,1,m) = zero
         fy(l,1,1,m) = zero
         fy(l,k1,1,m) = zero
         fy(l1,1,1,m) = zero
         fy(l1,k1,1,m) = zero
         fz(l,1,1,m) = at4*cmplx(aimag(q(l,1,1,m)),-real(q(l,1,1,m)))
         fz(l,k1,1,m) = zero
         fz(l1,1,1,m) = zero
         fz(l1,k1,1,m) = zero
         wp = wp + at1*(q(l,1,1,m)*conjg(q(l,1,1,m)))
  140    continue
      endif
  150 continue
  160 continue

      wp = float(nx*ny*nz)*wp
      return
      end


      subroutine ppush3 (part,fx,fy,fz,npp,noff,qtm,dt,ek,nx,ny,idimp,np
     1max,nblok,nxv,nyv,nzpmx)
c for 3d code, this subroutine updates particle co-ordinates and
c velocities using leap-frog scheme in time and second-order spline
c interpolation in space, with periodic boundary conditions,

      character*8 jbname
      common /pparms/ nproc, nprec, lgrp, jbname
      external d_vadd
      dimension part(idimp,npmax,nblok)
      dimension fx(nxv,nyv,nzpmx,nblok), fy(nxv,nyv,nzpmx,nblok)
      dimension fz(nxv,nyv,nzpmx,nblok)
      dimension npp(nblok), noff(nblok)
      zero = 0.
      anx = float(nx)
      any = float(ny)
      sum1 = 0.
      do 20 m = 1, nblok
      lnoff = noff(m) - 2
c find interpolation weights
      do 10 j = 1, npp(m)
c     do 10 j = 1, npmax
      nn = part(1,j,m) + .5
      dx = part(1,j,m) - float(nn)
      nn = nn + 1
      if (nn.gt.nx) nn = nn - nx
      amx = .75 - dx*dx
      np = nn + 1
      if (np.gt.nx) np = np - nx
      dxp = .5*(.5 + dx)**2
      nl = nn - 1
      if (nl.lt.1) nl = nl + nx
      dxl = .5*(.5 - dx)**2
      mm = part(2,j,m) + .5
      dy = part(2,j,m) - float(mm)
      mm = mm + 1
      if (mm.gt.ny) mm = mm - ny
      amy = .75 - dy*dy
      mp = mm + 1
      if (mp.gt.ny) mp = mp - ny
      dyp = .5*(.5 + dy)**2
      ml = mm - 1
      if (ml.lt.1) ml = ml + ny
      dyl = .5*(.5 - dy)**2
      ll = part(3,j,m) + .5
      dz = part(3,j,m) - float(ll)
      ll = ll - lnoff
      amz = .75 - dz*dz
      lp = ll + 1
      dzp = .5*(.5 + dz)**2
      lm = ll - 1
      dzl = .5*(.5 - dz)**2
c find acceleration

      dx = amz*(amy*(amx*fx(nn,mm,ll,m) + dxp*fx(np,mm,ll,m) + dxl*fx(nl
     1,mm,ll,m)) + dyp*(amx*fx(nn,mp,ll,m) + dxp*fx(np,mp,ll,m) + dxl*fx
     2(nl,mp,ll,m)) + dyl*(amx*fx(nn,ml,ll,m) + dxp*fx(np,ml,ll,m) + dxl
     3*fx(nl,ml,ll,m))) + dzp*(amy*(amx*fx(nn,mm,lp,m) + dxp*fx(np,mm,lp
     4,m) + dxl*fx(nl,mm,lp,m)) + dyp*(amx*fx(nn,mp,lp,m) + dxp*fx(np,mp
     5,lp,m) + dxl*fx(nl,mp,lp,m)) + dyl*(amx*fx(nn,ml,lp,m) + dxp*fx(np
     6,ml,lp,m) + dxl*fx(nl,ml,lp,m))) + dzl*(amy*(amx*fx(nn,mm,lm,m) +
     7dxp*fx(np,mm,lm,m) + dxl*fx(nl,mm,lm,m)) + dyp*(amx*fx(nn,mp,lm,m)
     8 + dxp*fx(np,mp,lm,m) + dxl*fx(nl,mp,lm,m)) + dyl*(amx*fx(nn,ml,lm
     9,m) + dxp*fx(np,ml,lm,m) + dxl*fx(nl,ml,lm,m)))
      dy = amz*(amy*(amx*fy(nn,mm,ll,m) + dxp*fy(np,mm,ll,m) + dxl*fy(nl
     1,mm,ll,m)) + dyp*(amx*fy(nn,mp,ll,m) + dxp*fy(np,mp,ll,m) + dxl*fy
     2(nl,mp,ll,m)) + dyl*(amx*fy(nn,ml,ll,m) + dxp*fy(np,ml,ll,m) + dxl
     3*fy(nl,ml,ll,m))) + dzp*(amy*(amx*fy(nn,mm,lp,m) + dxp*fy(np,mm,lp
     4,m) + dxl*fy(nl,mm,lp,m)) + dyp*(amx*fy(nn,mp,lp,m) + dxp*fy(np,mp
     5,lp,m) + dxl*fy(nl,mp,lp,m)) + dyl*(amx*fy(nn,ml,lp,m) + dxp*fy(np
     6,ml,lp,m) + dxl*fy(nl,ml,lp,m))) + dzl*(amy*(amx*fy(nn,mm,lm,m) +
     7dxp*fy(np,mm,lm,m) + dxl*fy(nl,mm,lm,m)) + dyp*(amx*fy(nn,mp,lm,m)
     8 + dxp*fy(np,mp,lm,m) + dxl*fy(nl,mp,lm,m)) + dyl*(amx*fy(nn,ml,lm
     9,m) + dxp*fy(np,ml,lm,m) + dxl*fy(nl,ml,lm,m)))
      dz = amz*(amy*(amx*fz(nn,mm,ll,m) + dxp*fz(np,mm,ll,m) + dxl*fz(nl
     1,mm,ll,m)) + dyp*(amx*fz(nn,mp,ll,m) + dxp*fz(np,mp,ll,m) + dxl*fz
     2(nl,mp,ll,m)) + dyl*(amx*fz(nn,ml,ll,m) + dxp*fz(np,ml,ll,m) + dxl
     3*fz(nl,ml,ll,m))) + dzp*(amy*(amx*fz(nn,mm,lp,m) + dxp*fz(np,mm,lp
     4,m) + dxl*fz(nl,mm,lp,m)) + dyp*(amx*fz(nn,mp,lp,m) + dxp*fz(np,mp
     5,lp,m) + dxl*fz(nl,mp,lp,m)) + dyl*(amx*fz(nn,ml,lp,m) + dxp*fz(np
     6,ml,lp,m) + dxl*fz(nl,ml,lp,m))) + dzl*(amy*(amx*fz(nn,mm,lm,m) +
     7dxp*fz(np,mm,lm,m) + dxl*fz(nl,mm,lm,m)) + dyp*(amx*fz(nn,mp,lm,m)
     8 + dxp*fz(np,mp,lm,m) + dxl*fz(nl,mp,lm,m)) + dyl*(amx*fz(nn,ml,lm
     9,m) + dxp*fz(np,ml,lm,m) + dxl*fz(nl,ml,lm,m)))
c new velocity
      dx = part(4,j,m) + qtm*dx
      dy = part(5,j,m) + qtm*dy
      dz = part(6,j,m) + qtm*dz
c average kinetic energy
      sum1 = sum1 + (dx + part(4,j,m))**2 + (dy + part(5,j,m))**2 + (dz
     1+ part(6,j,m))**2
      part(4,j,m) = dx
      part(5,j,m) = dy
      part(6,j,m) = dz
c new position
      dx = part(1,j,m) + dx*dt
      dy = part(2,j,m) + dy*dt
      dz = part(3,j,m) + dz*dt
c periodic boundary conditions
      if (dx.lt.zero) dx = dx + anx
      if (dx.ge.anx) dx = dx - anx
      part(1,j,m) = dx
      if (dy.lt.zero) dy = dy + any
      if (dy.ge.any) dy = dy - any
      part(2,j,m) = dy
      part(3,j,m) = dz
   10 continue
   20 continue

c normalize kinetic energy
      ek = ek + .125*sum1
      return
      end



      subroutine pmove3 (part,edges,npp,sbufr,sbufl,rbufr,rbufl,ihole,js
     1r,jsl,jss,nz,kstrt,nvp,idimp,npmax,nblok,idps,nbmax,ntmax,ierr)
c this subroutine moves particles into appropriate spatial regions

      character*8 jbname

      common /pparms/ nproc, nprec, lgrp, jbname
      external i_vadd
      dimension part(idimp,npmax,nblok)
      dimension edges(idps,nblok), npp(nblok)
      dimension sbufl(idimp,nbmax,nblok), sbufr(idimp,nbmax,nblok)
      dimension rbufl(idimp,nbmax,nblok), rbufr(idimp,nbmax,nblok)
      dimension jsl(idps,nblok), jsr(idps,nblok), jss(idps,nblok)
      dimension ihole(ntmax,nblok)
      anz = float(nz)
      ks = kstrt - 2
      iter = 4
c debugging section: count total number of particles before move
      npr = 0
      do 5 m = 1, nblok
      npr = npr + npp(m)
    5 continue

c buffer outgoing particles
   10 do 30 m = 1, nblok
      kb = m + ks
      jsl(1,m) = 0
      jsr(1,m) = 0
      jss(2,m) = 0
      do 20 j = 1, npp(m)

      zt = part(3,j,m)
c particles going forward
      if (zt.ge.edges(2,m)) then
         if (jsr(1,m).lt.nbmax) then
            jsr(1,m) = jsr(1,m) + 1
            if ((kb+1).eq.nvp) zt = zt - anz
            sbufr(1,jsr(1,m),m) = part(1,j,m)
            sbufr(2,jsr(1,m),m) = part(2,j,m)
            sbufr(3,jsr(1,m),m) = zt
            sbufr(4,jsr(1,m),m) = part(4,j,m)
            sbufr(5,jsr(1,m),m) = part(5,j,m)
            sbufr(6,jsr(1,m),m) = part(6,j,m)
            ihole(jsl(1,m)+jsr(1,m),m) = j
         else
            jss(2,m) = 1
         endif
c particles going backward
      elseif (zt.lt.edges(1,m)) then
         if (jsl(1,m).lt.nbmax) then
            jsl(1,m) = jsl(1,m) + 1
            if (kb.eq.0) zt = zt + anz
            sbufl(1,jsl(1,m),m) = part(1,j,m)
            sbufl(2,jsl(1,m),m) = part(2,j,m)
            sbufl(3,jsl(1,m),m) = zt
            sbufl(4,jsl(1,m),m) = part(4,j,m)
            sbufl(5,jsl(1,m),m) = part(5,j,m)
            sbufl(6,jsl(1,m),m) = part(6,j,m)
            ihole(jsl(1,m)+jsr(1,m),m) = j
         else
            jss(2,m) = 1
         endif
      endif
   20 continue
      jss(1,m) = jsl(1,m) + jsr(1,m)
   30 continue
c check for full buffer condition
      ibflg = 0
      do 40 m = 1, nblok
      ibflg = ibflg + jss(2,m)
   40 continue

c copy particle buffers
   50 iter = iter + 4
      do 80 m = 1, nblok
c get particles from behind
c this segment is used for shared memory computers
      kl = m + ks
      if (kl.lt.1) then
         jsl(2,m) = jsr(1,kl+nvp)
      else
         jsl(2,m) = jsr(1,kl)
      endif
      
c this loop is used for shared memory computers
      do 60 j = 1, jsl(2,m)
      if (kl.lt.1) then
         rbufl(1,j,m) = sbufr(1,j,kl+nvp)
         rbufl(2,j,m) = sbufr(2,j,kl+nvp)
         rbufl(3,j,m) = sbufr(3,j,kl+nvp)
         rbufl(4,j,m) = sbufr(4,j,kl+nvp)
         rbufl(5,j,m) = sbufr(5,j,kl+nvp)
         rbufl(6,j,m) = sbufr(6,j,kl+nvp)
      else
         rbufl(1,j,m) = sbufr(1,j,kl)
         rbufl(2,j,m) = sbufr(2,j,kl)
         rbufl(3,j,m) = sbufr(3,j,kl)
         rbufl(4,j,m) = sbufr(4,j,kl)
         rbufl(5,j,m) = sbufr(5,j,kl)
         rbufl(6,j,m) = sbufr(6,j,kl)
      endif
   60 continue

c get particles from ahead
c this segment is used for shared memory computers
      kr = m + ks + 2
      if (kr.gt.nvp) then
         jsr(2,m) = jsl(1,kr-nvp)
      else
         jsr(2,m) = jsl(1,kr)
      endif

c this loop is used for shared memory computers
      do 70 j = 1,jsr(2,m)
      if (kr.gt.nvp) then
         rbufr(1,j,m) = sbufl(1,j,kr-nvp)
         rbufr(2,j,m) = sbufl(2,j,kr-nvp)
         rbufr(3,j,m) = sbufl(3,j,kr-nvp)
         rbufr(4,j,m) = sbufl(4,j,kr-nvp)
         rbufr(5,j,m) = sbufl(5,j,kr-nvp)
         rbufr(6,j,m) = sbufl(6,j,kr-nvp)
      else
         rbufr(1,j,m) = sbufl(1,j,kr)
         rbufr(2,j,m) = sbufl(2,j,kr)
         rbufr(3,j,m) = sbufl(3,j,kr)
         rbufr(4,j,m) = sbufl(4,j,kr)
         rbufr(5,j,m) = sbufl(5,j,kr)
         rbufr(6,j,m) = sbufl(6,j,kr)
      endif
   70 continue

   80 continue
c check if particles must be passed further
      nps = 0
      do 110 m = 1, nblok
c check if any particles coming from ahead must go further back
      jsl(1,m) = 0
      do 90 j = 1, jsr(2,m)
      if (rbufr(3,j,m).lt.edges(1,m)) jsl(1,m) = 1
   90 continue
c check if any particles coming from behind must go further forward
      jsr(1,m) = 0
      do 100 j = 1, jsl(2,m)
      if (rbufl(3,j,m).ge.edges(2,m)) jsr(1,m) = 1
  100 continue
      nps = nps + (jsl(1,m) + jsr(1,m))
  110 continue

      if (nps.eq.0) go to 150
c remove particles which do not belong here
      do 140 m = 1, nblok
      kb = m + ks
c first check particles coming from ahead
      jsl(1,m) = 0
      jss(2,m) = 0
      do 120 j = 1, jsr(2,m)
      zt = rbufr(3,j,m)
c particles going backward
      if (zt.lt.edges(1,m)) then
         jsl(1,m) = jsl(1,m) + 1
         if (kb.eq.0) zt = zt + anz
         sbufl(1,jsl(1,m),m) = rbufr(1,j,m)
         sbufl(2,jsl(1,m),m) = rbufr(2,j,m)
         sbufl(3,jsl(1,m),m) = zt
         sbufl(4,jsl(1,m),m) = rbufr(4,j,m)
         sbufl(5,jsl(1,m),m) = rbufr(5,j,m)
         sbufl(6,jsl(1,m),m) = rbufr(6,j,m)
c particles staying here
      else
         jss(2,m) = jss(2,m) + 1
         rbufr(1,jss(2,m),m) = rbufr(1,j,m)
         rbufr(2,jss(2,m),m) = rbufr(2,j,m)
         rbufr(3,jss(2,m),m) = zt
         rbufr(4,jss(2,m),m) = rbufr(4,j,m)
         rbufr(5,jss(2,m),m) = rbufr(5,j,m)
         rbufr(6,jss(2,m),m) = rbufr(6,j,m)
      endif
  120 continue
      jsr(2,m) = jss(2,m)
c next check particles coming from behind
      jsr(1,m) = 0
      jss(2,m) = 0
      do 130 j = 1, jsl(2,m)
      zt = rbufl(3,j,m)
c particles going forward
      if (zt.ge.edges(2,m)) then
         jsr(1,m) = jsr(1,m) + 1
         if ((kb+1).eq.nvp) zt = zt - anz
         sbufr(1,jsr(1,m),m) = rbufl(1,j,m)
         sbufr(2,jsr(1,m),m) = rbufl(2,j,m)
         sbufr(3,jsr(1,m),m) = zt
         sbufr(4,jsr(1,m),m) = rbufl(4,j,m)
         sbufr(5,jsr(1,m),m) = rbufl(5,j,m)
         sbufr(6,jsr(1,m),m) = rbufl(6,j,m)
c particles staying here
      else
         jss(2,m) = jss(2,m) + 1
         rbufl(1,jss(2,m),m) = rbufl(1,j,m)
         rbufl(2,jss(2,m),m) = rbufl(2,j,m)
         rbufl(3,jss(2,m),m) = zt
         rbufl(4,jss(2,m),m) = rbufl(4,j,m)
         rbufl(5,jss(2,m),m) = rbufl(5,j,m)
         rbufl(6,jss(2,m),m) = rbufl(6,j,m)
      endif
  130 continue
      jsl(2,m) = jss(2,m)
  140 continue
c check if move would overflow particle array
  150 ierr = 0
      do 160 m = 1, nblok
      if ((npp(m)+jsl(2,m)+jsr(2,m)-jss(1,m)).gt.npmax) then
         jss(2,m) = 1
      else
         jss(2,m) = 0
      endif
      ierr = ierr + jss(2,m)
  160 continue

      if (ierr.gt.0) then
         write (6,*) 'particle overflow error, ierr =', ierr
         return
      endif
      do 200 m = 1, nblok
c distribute incoming particles from buffers
c distribute particles coming from behind into holes
      jss(2,m) = min0(jss(1,m),jsl(2,m))
      do 170 j = 1, jss(2,m)

      part(1,ihole(j,m),m) = rbufl(1,j,m)
      part(2,ihole(j,m),m) = rbufl(2,j,m)
      part(3,ihole(j,m),m) = rbufl(3,j,m)
      part(4,ihole(j,m),m) = rbufl(4,j,m)
      part(5,ihole(j,m),m) = rbufl(5,j,m)
      part(6,ihole(j,m),m) = rbufl(6,j,m)
  170 continue
      if (jss(1,m).gt.jsl(2,m)) then
         jss(2,m) = min0(jss(1,m)-jsl(2,m),jsr(2,m))
      else
         jss(2,m) = jsl(2,m) - jss(1,m)
      endif
      do 180 j = 1, jss(2,m)

c no more particles coming from behind
c distribute particles coming from ahead into holes
      if (jss(1,m).gt.jsl(2,m)) then
         part(1,ihole(j+jsl(2,m),m),m) = rbufr(1,j,m)
         part(2,ihole(j+jsl(2,m),m),m) = rbufr(2,j,m)
         part(3,ihole(j+jsl(2,m),m),m) = rbufr(3,j,m)
         part(4,ihole(j+jsl(2,m),m),m) = rbufr(4,j,m)
         part(5,ihole(j+jsl(2,m),m),m) = rbufr(5,j,m)
         part(6,ihole(j+jsl(2,m),m),m) = rbufr(6,j,m)
      else
c no more holes
c distribute remaining particles from behind into bottom
         part(1,j+npp(m),m) = rbufl(1,j+jss(1,m),m)
         part(2,j+npp(m),m) = rbufl(2,j+jss(1,m),m)
         part(3,j+npp(m),m) = rbufl(3,j+jss(1,m),m)
         part(4,j+npp(m),m) = rbufl(4,j+jss(1,m),m)
         part(5,j+npp(m),m) = rbufl(5,j+jss(1,m),m)
         part(6,j+npp(m),m) = rbufl(6,j+jss(1,m),m)
      endif
  180 continue
      if (jss(1,m).le.jsl(2,m)) then
         npp(m) = npp(m) + (jsl(2,m) - jss(1,m))
         jss(1,m) = jsl(2,m)
      endif
      jss(2,m) = jss(1,m) - (jsl(2,m) + jsr(2,m))
      if (jss(2,m).gt.0) then
         jss(1,m) = (jsl(2,m) + jsr(2,m))
         jsr(2,m) = jss(2,m)
      else
         jss(1,m) = jss(1,m) - jsl(2,m)
         jsr(2,m) = -jss(2,m)
      endif
      do 190 j = 1, jsr(2,m)

c holes left over
c fill up remaining holes in particle array with particles from bottom
      if (jss(2,m).gt.0) then
         j1 = npp(m) - j + 1
         j2 = jss(1,m) + jss(2,m) - j + 1
         if (j1.gt.ihole(j2,m)) then
c move particle only if it is below current hole
            part(1,ihole(j2,m),m) = part(1,j1,m)
            part(2,ihole(j2,m),m) = part(2,j1,m)
            part(3,ihole(j2,m),m) = part(3,j1,m)
            part(4,ihole(j2,m),m) = part(4,j1,m)
            part(5,ihole(j2,m),m) = part(5,j1,m)
            part(6,ihole(j2,m),m) = part(6,j1,m)
         endif
      else
c no more holes
c distribute remaining particles from ahead into bottom
         part(1,j+npp(m),m) = rbufr(1,j+jss(1,m),m)
         part(2,j+npp(m),m) = rbufr(2,j+jss(1,m),m)
         part(3,j+npp(m),m) = rbufr(3,j+jss(1,m),m)
         part(4,j+npp(m),m) = rbufr(4,j+jss(1,m),m)
         part(5,j+npp(m),m) = rbufr(5,j+jss(1,m),m)
         part(6,j+npp(m),m) = rbufr(6,j+jss(1,m),m)
      endif
  190 continue
      if (jss(2,m).gt.0) then
         npp(m) = npp(m) - jsr(2,m)
      else
         npp(m) = npp(m) + jsr(2,m)
      endif
      jss(1,m) = 0
  200 continue
c check if any particles have to be passed further
      if (nps.gt.0) go to 50
c check if buffer overflowed and more particles remain to be checked
      if (ibflg.gt.0) go to 10
c debugging section: count total number of particles after move
      nps = 0
      do 205 m = 1, nblok
      nps = nps + npp(m)
  205 continue

      if (npr.ne.nps) write (6,*) 'number error, np-old,np-new=',npr,nps
      return
      end



      subroutine pfft3r(f,g,bs,br,isign,ntpose,mixup,sct,indx,indy,indz,
     1kstrt,nxv,nyv,nzv,kxp,kzp,jblok,lblok,nyd,nxhyzd,nxyzhd)
c this subroutine performs a three dimensional real to complex fast
c fourier transform and its inverse, using complex arithmetic,
c for data which is distributed in blocks

      complex f, g, bs, br, sct, s, t, t1
      dimension f(nxv,nyv,kzp,lblok), g(nzv,nyv,kxp,jblok)
      dimension bs(kxp,nyd,kzp,lblok), br(kxp,nyd,kzp,jblok)
      dimension mixup(nxhyzd), sct(nxyzhd)
      indx1 = indx - 1
      ndx1yz = max0(indx1,indy,indz)
      nx = 2**indx
      nxh = nx/2
      nxhh = nx/4
      nxh2 = nxh + 2
      ny = 2**indy
      nyh = ny/2
      ny2 = ny + 2
      nz = 2**indz
      nzh = nz/2
      nz2 = nz + 2
      nxyz = max0(nx,ny,nz)
      nxhyz = 2**ndx1yz
      ks = kstrt - 2
      if (isign) 50, 10, 460
c prepare fft tables
c bit-reverse index table: mixup(j) = 1 + reversed bits of (j - 1)
   10 do 30 j = 1, nxhyz
      lb = j - 1
      ll = 0
      do 20 k = 1, ndx1yz
      jb = lb/2
      it = lb - 2*jb
      lb = jb
      ll = 2*ll + it
   20 continue
      mixup(j) = ll + 1
   30 continue
c sine/cosine table for the angles 2*n*pi/nxyz
      nxyzh = nxyz/2
      dnxyz = 6.28318530717959/float(nxyz)
      do 40 j = 1, nxyzh
      arg = dnxyz*float(j - 1)
      sct(j) = cmplx(cos(arg),-sin(arg))
   40 continue
      return
c inverse fourier transform
   50 if (kstrt.gt.nz) go to 320
      nrx = nxhyz/nxh
      do 90 m = 1, lblok
c bit-reverse array elements in x
      do 80 n = 1, kzp
      do 70 j = 1, nxh
      j1 = (mixup(j) - 1)/nrx + 1
      if (j.ge.j1) go to 70
      do 60 i = 1, ny
c     if (j.lt.j1) then
      t = f(j1,i,n,m)
      f(j1,i,n,m) = f(j,i,n,m)
      f(j,i,n,m) = t
c     endif
   60 continue
   70 continue
   80 continue
   90 continue
c first transform in x
      nrx = nxyz/nxh
      do 150 l = 1, indx1
      ns = 2**(l - 1)
      ns2 = ns + ns
      km = nxhh/ns
      kmr = km*nrx
      do 140 m = 1, lblok
      do 130 n = 1, kzp
      do 120 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 110 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = sct(1+kmr*(j-1))
      do 100 i = 1, ny
      t = s*f(j2,i,n,m)
      f(j2,i,n,m) = f(j1,i,n,m) - t
      f(j1,i,n,m) = f(j1,i,n,m) + t
  100 continue
  110 continue
  120 continue
  130 continue
  140 continue
  150 continue
c unscramble coefficients and normalize
      kmr = nxyz/nx
      ani = 1./float(2*nx*ny*nz)
      nry = nxhyz/ny
      do 220 m = 1, lblok
      do 210 n = 1, kzp
      do 170 j = 2, nxhh
      t1 = cmplx(aimag(sct(1+kmr*(j-1))),-real(sct(1+kmr*(j-1))))
      do 160 k = 1, ny
      t = conjg(f(nxh2-j,k,n,m))
      s = f(j,k,n,m) + t
      t = (f(j,k,n,m) - t)*t1
      f(j,k,n,m) = ani*(s + t)
      f(nxh2-j,k,n,m) = ani*conjg(s - t)
  160 continue
  170 continue
      do 180 k = 1, ny
      f(nxhh+1,k,n,m) = 2.*ani*conjg(f(nxhh+1,k,n,m))
      f(1,k,n,m) = 2.*ani*cmplx(real(f(1,k,n,m)) + aimag(f(1,k,n,m)),rea
     1l(f(1,k,n,m)) - aimag(f(1,k,n,m)))
  180 continue
c bit-reverse array elements in y
      do 200 k = 1, ny
      k1 = (mixup(k) - 1)/nry + 1
      if (k.ge.k1) go to 200
      do 190 i = 1, nxh
c     if (k.lt.k1) then
      t = f(i,k1,n,m)
      f(i,k1,n,m) = f(i,k,n,m)
      f(i,k,n,m) = t
c     endif
  190 continue
  200 continue
  210 continue
  220 continue
c then transform in y
      nry = nxyz/ny
      do 280 l = 1, indy
      ns = 2**(l - 1)
      ns2 = ns + ns
      km = nyh/ns
      kmr = km*nry
      do 270 m = 1, lblok
      do 260 n = 1, kzp
      do 250 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 240 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = sct(1+kmr*(j-1))
      do 230 i = 1, nxh
      t = s*f(i,j2,n,m)
      f(i,j2,n,m) = f(i,j1,n,m) - t
      f(i,j1,n,m) = f(i,j1,n,m) + t
  230 continue
  240 continue
  250 continue
  260 continue
  270 continue
  280 continue
c unscramble modes kx = 0, nx/2
      do 310 m = 1, lblok
      do 300 n = 1, kzp
      do 290 k = 2, nyh
      s = f(1,ny2-k,n,m)
      f(1,ny2-k,n,m) = .5*cmplx(aimag(f(1,k,n,m) + s),real(f(1,k,n,m) -
     1s))
      f(1,k,n,m) = .5*cmplx(real(f(1,k,n,m) + s),aimag(f(1,k,n,m) - s))
  290 continue
  300 continue
  310 continue
c transpose f array to g
  320 call ptpos2(f,g,bs,br,nxh,ny,nz,kstrt,nxv,nyv,nzv,kxp,kzp,jblok,lb
     1lok)
      if (kstrt.gt.nxh) go to 450
      nrz = nxhyz/nz
      do 360 m = 1, jblok
c bit-reverse array elements in z
      do 350 l = 1, nz
      l1 = (mixup(l) - 1)/nrz + 1
      if (l.ge.l1) go to 350
      do 340 i = 1, kxp
      do 330 n = 1, ny
c     if (l.lt.l1) then
      t = g(l1,n,i,m)
      g(l1,n,i,m) = g(l,n,i,m)
      g(l,n,i,m) = t
c     endif
  330 continue
  340 continue
  350 continue
  360 continue
c finally transform in z
      nrz = nxyz/nz
      do 420 l = 1, indz
      ns = 2**(l - 1)
      ns2 = ns + ns
      km = nzh/ns
      kmr = km*nrz
      do 410 m = 1, jblok
      do 400 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 390 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = sct(1+kmr*(j-1))
      do 380 i = 1, kxp
      do 370 n = 1, ny
      t = s*g(j2,n,i,m)
      g(j2,n,i,m) = g(j1,n,i,m) - t
      g(j1,n,i,m) = g(j1,n,i,m) + t
  370 continue
  380 continue
  390 continue
  400 continue
  410 continue
  420 continue
c unscramble modes kx = 0, nx/2
      do 440 m = 1, jblok
      if ((m+ks).gt.0) go to 440
      do 430 n = 2, nzh
c     if ((m+ks).eq.0) then
      s = g(nz2-n,1,1,m)
      g(nz2-n,1,1,m) = .5*cmplx(aimag(g(n,1,1,m) + s),real(g(n,1,1,m) -
     1s))
      g(n,1,1,m) = .5*cmplx(real(g(n,1,1,m) + s),aimag(g(n,1,1,m) - s))
      s = g(nz2-n,nyh+1,1,m)
      g(nz2-n,nyh+1,1,m) = .5*cmplx(aimag(g(n,nyh+1,1,m) + s),real(g(n,n
     1yh+1,1,m) - s))
      g(n,nyh+1,1,m) = .5*cmplx(real(g(n,nyh+1,1,m) + s),aimag(g(n,nyh+1
     1,1,m) - s))
c     endif
  430 continue
  440 continue
c transpose g array to f
  450 if (ntpose.eq.0) call ptpos2(g,f,br,bs,nz,ny,nxh,kstrt,nzv,nyv,nxv
     1,kzp,kxp,lblok,jblok)
      return
c forward fourier transform
c transpose f array to g
  460 if (ntpose.eq.0) call ptpos2(f,g,bs,br,nxh,ny,nz,kstrt,nxv,nyv,nzv
     1,kxp,kzp,jblok,lblok)
      if (kstrt.gt.nxh) go to 590
      nrz = nxhyz/nz
c scramble modes kx = 0, nx/2
      do 480 m = 1, jblok
      if ((m+ks).gt.0) go to 480
      do 470 n = 2, nzh
c     if ((m+ks).eq.0) then
      s = cmplx(aimag(g(nz2-n,1,1,m)),real(g(nz2-n,1,1,m)))
      g(nz2-n,1,1,m) = conjg(g(n,1,1,m) - s)
      g(n,1,1,m) = g(n,1,1,m) + s
      s = cmplx(aimag(g(nz2-n,nyh+1,1,m)),real(g(nz2-n,nyh+1,1,m)))
      g(nz2-n,nyh+1,1,m) = conjg(g(n,nyh+1,1,m) - s)
      g(n,nyh+1,1,m) = g(n,nyh+1,1,m) + s
c     endif
  470 continue
  480 continue
      do 520 m = 1, jblok
c bit-reverse array elements in z
      do 510 l = 1, nz
      l1 = (mixup(l) - 1)/nrz + 1
      if (l.ge.l1) go to 510
      do 500 i = 1, kxp
      do 490 n = 1, ny
c     if (l.lt.l1) then
      t = g(l1,n,i,m)
      g(l1,n,i,m) = g(l,n,i,m)
      g(l,n,i,m) = t
c     endif
  490 continue
  500 continue
  510 continue
  520 continue
c first transform in z
      nrz = nxyz/nz
      do 580 l = 1, indz
      ns = 2**(l - 1)
      ns2 = ns + ns
      km = nzh/ns
      kmr = km*nrz
      do 570 m = 1, jblok
      do 560 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 550 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = conjg(sct(1+kmr*(j-1)))
      do 540 i = 1, kxp
      do 530 n = 1, ny
      t = s*g(j2,n,i,m)
      g(j2,n,i,m) = g(j1,n,i,m) - t
      g(j1,n,i,m) = g(j1,n,i,m) + t
  530 continue
  540 continue
  550 continue
  560 continue
  570 continue
  580 continue
c transpose g array to f
  590 call ptpos2(g,f,br,bs,nz,ny,nxh,kstrt,nzv,nyv,nxv,kzp,kxp,lblok,jb
     1lok)
      if (kstrt.gt.nz) return
      nry = nxhyz/ny
c scramble modes kx = 0, nx/2
      do 620 m = 1, lblok
      do 610 n = 1, kzp
      do 600 k = 2, nyh
      s = cmplx(aimag(f(1,ny2-k,n,m)),real(f(1,ny2-k,n,m)))
      f(1,ny2-k,n,m) = conjg(f(1,k,n,m) - s)
      f(1,k,n,m) = f(1,k,n,m) + s
  600 continue
  610 continue
  620 continue
      do 660 m = 1, lblok
c bit-reverse array elements in y
      do 650 n = 1, kzp
      do 640 k = 1, ny
      k1 = (mixup(k) - 1)/nry + 1
      if (k.ge.k1) go to 640
      do 630 i = 1, nxh
c     if (k.lt.k1) then
      t = f(i,k1,n,m)
      f(i,k1,n,m) = f(i,k,n,m)
      f(i,k,n,m) = t
c     endif
  630 continue
  640 continue
  650 continue
  660 continue
c then transform in y
      nry = nxyz/ny
      do 720 l = 1, indy
      ns = 2**(l - 1)
      ns2 = ns + ns
      km = nyh/ns
      kmr = km*nry
      do 710 m = 1, lblok
      do 700 n = 1, kzp
      do 690 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 680 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = conjg(sct(1+kmr*(j-1)))
      do 670 i = 1, nxh
      t = s*f(i,j2,n,m)
      f(i,j2,n,m) = f(i,j1,n,m) - t
      f(i,j1,n,m) = f(i,j1,n,m) + t
  670 continue
  680 continue
  690 continue
  700 continue
  710 continue
  720 continue
      nrx = nxhyz/nxh
c scramble coefficients
      kmr = nxyz/nx
      do 790 m = 1, lblok
      do 780 n = 1, kzp
      do 740 j = 2, nxhh
      t1 = cmplx(aimag(sct(1+kmr*(j-1))),real(sct(1+kmr*(j-1))))
      do 730 k = 1, ny
      t = conjg(f(nxh2-j,k,n,m))
      s = f(j,k,n,m) + t
      t = (f(j,k,n,m) - t)*t1
      f(j,k,n,m) = s + t
      f(nxh2-j,k,n,m) = conjg(s - t)
  730 continue
  740 continue
      do 750 k = 1, ny
      f(nxhh+1,k,n,m) = 2.*conjg(f(nxhh+1,k,n,m))
      f(1,k,n,m) = cmplx(real(f(1,k,n,m)) + aimag(f(1,k,n,m)),real(f(1,k
     1,n,m)) - aimag(f(1,k,n,m)))
  750 continue
c bit-reverse array elements in x
      do 770 j = 1, nxh
      j1 = (mixup(j) - 1)/nrx + 1
      if (j.ge.j1) go to 770
      do 760 i = 1, ny
c     if (j.lt.j1) then
      t = f(j1,i,n,m)
      f(j1,i,n,m) = f(j,i,n,m)
      f(j,i,n,m) = t
c     endif
  760 continue
  770 continue
  780 continue
  790 continue
c finally transform in x
      nrx = nxyz/nxh
      do 850 l = 1, indx1
      ns = 2**(l - 1)
      ns2 = ns + ns
      km = nxhh/ns
      kmr = km*nrx
      do 840 m = 1, lblok
      do 830 n = 1, kzp
      do 820 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 810 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = conjg(sct(1+kmr*(j-1)))
      do 800 i = 1, ny
      t = s*f(j2,i,n,m)
      f(j2,i,n,m) = f(j1,i,n,m) - t
      f(j1,i,n,m) = f(j1,i,n,m) + t
  800 continue
  810 continue
  820 continue
  830 continue
  840 continue
  850 continue
      return
      end

      subroutine ptpos2(f,g,s,t,nx,ny,nz,kstrt,nxv,nyv,nzv,kxp,kzp,jblok
     1,lblok)

      character*8 jbname

      common /pparms/ nproc, nprec, lgrp, jbname
      complex f, g, s, t
      dimension f(nxv,nyv,kzp,lblok), g(nzv,nyv,kxp,jblok)
      dimension s(kxp,ny,kzp,lblok), t(kxp,ny,kzp,jblok)
      ks = kstrt - 2
      kxb = nx/kxp
      kzb = nz/kzp
c this segment is used for shared memory computers
      if (kstrt.gt.nx) return
      do 50 m = 1, jblok
      joff = kxp*(m + ks)
      do 40 i = 1, kzb
      loff = kzp*(i - 1)   
      do 30 l = 1, kzp
      do 20 k = 1, ny
      do 10 j = 1, kxp
      g(l+loff,k,j,m) = f(j+joff,k,l,i)
   10 continue
   20 continue
   30 continue
   40 continue
   50 continue

      return
      end

      function ranorm(d)
c this program calculates a random number y from a gaussian distribution
c with zero mean and unit variance, according to the method of
c mueller and box:

      integer r1,r2,r4,r5
      double precision ranorm,h1l,h1u,h2l,r0,r3,asc,bsc,temp
      save iflg,r1,r2,r4,r5,h1l,h1u,h2l,r0
      data r1,r2,r4,r5 /885098780,1824280461,1396483093,55318673/
      data h1l,h1u,h2l /65531.0d0,32767.0d0,65525.0d0/
      data iflg,r0 /0,0.0d0/
      if (iflg.eq.0) go to 10
      ranorm = r0
      r0 = 0.0d0
      iflg = 0
      return
   10 isc = 65536
      asc = dfloat(isc)
      bsc = asc*asc
      i1 = r1 - (r1/isc)*isc
      r3 = h1l*dfloat(r1) + asc*h1u*dfloat(i1)
      i1 = r3/bsc
      r3 = r3 - dfloat(i1)*bsc
      bsc = 0.5d0*bsc
      i1 = r2/isc
      isc = r2 - i1*isc
      r0 = h1l*dfloat(r2) + asc*h1u*dfloat(isc)
      asc = 1.0d0/bsc
      isc = r0*asc
      r2 = r0 - dfloat(isc)*bsc
      r3 = r3 + (dfloat(isc) + 2.0d0*h1u*dfloat(i1))
      isc = r3*asc
      r1 = r3 - dfloat(isc)*bsc
      temp = dsqrt(-2.0d0*dlog((dfloat(r1) + dfloat(r2)*asc)*asc))
      isc = 65536
      asc = dfloat(isc)
      bsc = asc*asc
      i1 = r4 - (r4/isc)*isc
      r3 = h2l*dfloat(r4) + asc*h1u*dfloat(i1)
      i1 = r3/bsc
      r3 = r3 - dfloat(i1)*bsc
      bsc = 0.5d0*bsc
      i1 = r5/isc
      isc = r5 - i1*isc
      r0 = h2l*dfloat(r5) + asc*h1u*dfloat(isc)
      asc = 1.0d0/bsc
      isc = r0*asc
      r5 = r0 - dfloat(isc)*bsc
      r3 = r3 + (dfloat(isc) + 2.0d0*h1u*dfloat(i1))
      isc = r3*asc
      r4 = r3 - dfloat(isc)*bsc
      r0 = 6.28318530717959d0*((dfloat(r4) + dfloat(r5)*asc)*asc)
      ranorm = temp*dsin(r0)
      r0 = temp*dcos(r0)
      iflg = 1
      return
      end


