
C* -------------------------------------------------------------------
C*
C* The following code is Dr. Decyk's 2-D PIC code for single-processor
C* computers. [KD]
C*
C* -------------------------------------------------------------------

c * * * periodic 2d electrostatic particle simulation kernel code * * *
c This is a simple 2d skeleton particle-in-cell code designed for
c exploring new computer architectures.  It contains the critical pieces
c needed for depositing charge, advancing particles, and solving the
c field.  The code moves only electrons, with periodic electrostatic
c forces obtained by solving poisson's equation with fast fourier
c transforms.  The only diagnostic is particle and field energy.
c portable gcpic kernel code, using algorithm described in:
c p. c. liewer and v. k. decyk, j. computational phys. 85, 302 (1989).
c written by viktor k. decyk, ucla

      program beps2k

      parameter( indx =   5, indy =   6, npx =      96, npy =     192)
cbenc parameter( indx =   6, indy =   7, npx =     384, npy =     768)
cbenc parameter( indx =   7, indy =   8, npx =    1280, npy =    2560)

      parameter( npxb =  32, npyb =  64)
cbenc parameter( npxb = 128, npyb = 256)
cbenc parameter( npxb = 384, npyb = 768)

      parameter( tend =  5.000, dt = 0.2000000e+00)

      parameter( vty =   1.000, vdy =   5.000, vtdy =   0.500)

      parameter( avdy =   5.000)

      parameter( indnvp =   0, idps =    2, idimp =   4)

      parameter(npxy=npx*npy,npxyb=npxb*npyb,np=npxy+npxyb)
      parameter(nx=2**indx,ny=2**indy,nxh=nx/2,nyh=ny/2)
      parameter(nxv=nx+2,nyv=ny+2,nxvh=nxv/2,nxvy=nxv*ny)

      parameter(nloop=tend/dt+.0001)

      parameter(nvp=2**indnvp,nblok=1+(nvp-1))

      parameter(npmax=(np/nvp)*1.01+7000,nypmx=(ny-1)/nvp+4)
      parameter(nxvyp=nxv*nypmx)

      parameter(kyp=(ny-1)/nvp+1,kxp=(nxh-1)/nvp+1)

      parameter(kblok=1+(ny/kyp-1),jblok=1+(nxh/kxp-1))

      parameter(nmx=nx*(ny/nx)+ny*(nx/ny),nxy=nmx/(2-nx/nmx-ny/nmx))
      parameter(nxyh=nxy/2,nmxh=nxh*(ny/nxh)+ny*(nxh/ny))
      parameter(nxhy=nmxh/(2-nxh/nmxh-ny/nmxh))

      parameter(nbmax=1+(2*(npxy*vty+npxyb*vtdy)+1.4*npxyb*avdy)*dt/ny)

      parameter(ntmax=2*nbmax)

      parameter(npd=128,nine=9)
      complex qc, fxc, fyc, qt, fxt, fyt, ffc, bs, br, scr, sct
      common /large/ part

      dimension part(idimp,npmax,nblok)

      dimension q(nxv,nypmx,nblok)
      dimension fx(nxv,nypmx,nblok), fy(nxv,nypmx,nblok)

      dimension qc(nxvh,kyp,kblok)
      dimension fxc(nxvh,kyp,kblok), fyc(nxvh,kyp,kblok)

      dimension qt(nyv,kxp,jblok)
      dimension fxt(nyv,kxp,jblok), fyt(nyv,kxp,jblok)

      dimension ffc(nyh,kxp,jblok)

      dimension mixup(nxhy), sct(nxyh)

      dimension bs(kxp,kyp,kblok), br(kxp,kyp,jblok)

      dimension edges(idps,nblok)

      dimension nyp(nblok), noff(nblok)

      dimension npp(nblok), nps(nblok)

      dimension sbufl(idimp,nbmax,nblok), sbufr(idimp,nbmax,nblok)

      dimension rbufl(idimp,nbmax,nblok), rbufr(idimp,nbmax,nblok)

      dimension ihole(ntmax,nblok)

      dimension jsl(idps,nblok), jsr(idps,nblok)

      dimension jss(idps,nblok), scr(nxvh,idps,nblok)

      real TIME_INIT, TIME_TOTAL
      real tbegin,timeend

  991 format (5h t = ,i7)
  992 format (19h * * * q.e.d. * * *)
  993 format (34h field, kinetic, total energies = ,3e14.7)

      data vtx,qme,vdx,vtdx /1.,-1.,0.,.5/

      data ax,ay /.866667,.866667/

c ntpose = (0,1) = (no,yes) input, output data are transposed in pfft2r
      data ntpose /1/

c -----------
c initialize
c -----------
      kstrt = 1
c      open(unit=6,file='output2',form='formatted',status='unknown')
c -------------------------------------
c initialize timer
c -------------------------------------
      call TIMERA(TIME_INIT)
c -------------------------------------
c initialize constants
c -------------------------------------
      itime = 0
      qtme = qme*dt
      affp = float(nx*ny)/float(np)
      zero = 0.
c -------------------------------------
c calculate partition variables
c -------------------------------------
      print *,'begin..',elapsed_time()
      call dcomp2(edges,nyp,noff,ny,kstrt,nvp,idps,nblok)
      print *,'after dcomp2',elapsed_time()

c -------------------------------------
c prepare fft tables
c -------------------------------------
      isign = 0
      call pfft2r(qc,qt,bs,br,isign,ntpose,mixup,sct,indx,indy,kstrt,nxv
     1h,nyv,kxp,kyp,jblok,kblok,nxhy,nxyh)
      print *,'after bitrv',elapsed_time()

c       do l=1,nxhy        
c            print *,'sct-mixup',l,sct(l),mixup(l)
c           enddo
 
c -------------------------------------
c calculate form factors
c -------------------------------------
      call ppois2(qt,fxt,fyt,isign,ffc,ax,ay,affp,we,nx,ny,kstrt,nyv,kxp
     1,jblok,nyh)

      print *,'after ppois2',elapsed_time() 


c ------------------------------------------------------
c initialize density profile and velocity distribution
c background electrons
c ------------------------------------------------------
      do 130 l = 1, nblok
      nps(l) = 1
      npp(l) = 0
      nyp3 = nyp(l) + 3
      do 120 k = 1, nyp3
       do 110 j = 1, nx
        q(j,k,l) = 0.
  110  continue
  120 continue
  130 continue
      if (npxy.gt.0) call pistr2 (part,edges,npp,nps,vtx,vty,zero,zero,n
     1px,npy,nx,ny,idimp,npmax,nblok,idps)
      
       print *,'after pistr2',elapsed_time() 
 

   
c ----------------
c beam electrons
c ----------------
      do 140 l = 1, nblok
      nps(l) = npp(l) + 1
  140 continue
      if (npxyb.gt.0) call pistr2 (part,edges,npp,nps,vtdx,vtdy,vdx,vdy,
     1npxb,npyb,nx,ny,idimp,npmax,nblok,idps)

       print *,'after pistr2',elapsed_time()
c ----------------------------------------
c deposit charge for initial distribution
c ----------------------------------------
      call pdost2(part,q,npp,noff,qme,nx,idimp,npmax,nblok,nxv,nypmx)
        print *,'after pdost2',elapsed_time()
  
c -----------------------------
c add background ion density
c -----------------------------
      qi0 = -qme/affp
      do 240 l = 1, nblok
      do 230 k = 1, nyp(l)
      do 220 j = 1, nx
      q(j,k+1,l) = q(j,k+1,l) + qi0
  220 continue
  230 continue
  240 continue
c =======================================
c
c * * * start main iteration loop * * *
c
c =======================================
        call timer(timebegin)
  500 if (nloop.le.itime) go to 2000
      write (6,991) itime
      isign = -1
c -------------------------------------------
c copy data from particle to field partition
c -------------------------------------------
      call cppfp2 (q,qc,isign,scr,kstrt,nvp,nx,nxv,nypmx,nblok,nxvh,kyp,
     1kblok,idps)
      print *,'after cppfp2',elapsed_time()
     
c       do k=1,kblok
c        do i=1,kyp
c         do j=1,nxvh
c          print *,'qc',j,i,k,qc(j,i,k)
c          enddo
c          enddo
c          enddo

c ----------------------------------
c transform charge to fourier space
c ----------------------------------
      call pfft2r(qc,qt,bs,br,isign,ntpose,mixup,sct,indx,indy,kstrt,nxv
     1h,nyv,kxp,kyp,jblok,kblok,nxhy,nxyh)
       print *,'after pfft2r',elapsed_time()

c       do k=1,jblok
c        do i=1,kxp
c         do j=1,nyv
c           print *,'fft-qc',j,i,k,qt(j,i,k)
c           enddo
c           enddo
c           enddo
c           stop
c -----------------------------------------
c calculate force/charge in fourier space
c -----------------------------------------
      call ppois2(qt,fxt,fyt,isign,ffc,ax,ay,affp,we,nx,ny,kstrt,nyv,kxp
     1,jblok,nyh)
         print *,'after ppois2',elapsed_time()
       
c -------------------------------------
c transform force/charge to real space
c -------------------------------------
      isign = 1
      call pfft2r(fxc,fxt,bs,br,isign,ntpose,mixup,sct,indx,indy,kstrt,n
     1xvh,nyv,kxp,kyp,jblok,kblok,nxhy,nxyh)
      print *,'after pfft2r',elapsed_time()

c        do k=1,kblok
c        do i=1,kyp
c         do j=1,nxvh
c           print *,'fxc',j,i,k,fxc(j,i,k)
c           enddo
c           enddo
c           enddo
c           stop
      
  
      call pfft2r(fyc,fyt,bs,br,isign,ntpose,mixup,sct,indx,indy,kstrt,n
     1xvh,nyv,kxp,kyp,jblok,kblok,nxhy,nxyh)
        print *,'after pfft2r',elapsed_time()
c       do k=1,kblok
c        do i=1,kyp
c         do j=1,nxvh
c           print *,'fc',j,i,k,fxc(j,i,k),fyc(j,i,k)
c           enddo
c           enddo
c           enddo
c           stop
      
c -------------------------------------------
c copy data from field to particle partition
c -------------------------------------------
      call cppfp2 (fx,fxc,isign,scr,kstrt,nvp,nx,nxv,nypmx,nblok,nxvh,ky
     1p,kblok,idps)
      print *,'after cppfp2',elapsed_time()
      call cppfp2 (fy,fyc,isign,scr,kstrt,nvp,nx,nxv,nypmx,nblok,nxvh,ky
     1p,kblok,idps)
      print *,'after cppfp2',elapsed_time()

c       do  k=1,nblok
c        do i=1,nypmx
c         do j=1,nxv
c           print *,'fxyc',j,i,k,fx(j,i,k),fy(j,i,k)
c           enddo
c           enddo
c           enddo
    

c -----------------------------------------
c particle push and charge density update
c -----------------------------------------
c     call timera(TIME_PUSH)
c -----------------------------------
c initialize charge density to zero
c -----------------------------------
      do 1250 l = 1, nblok
      nyp3 = nyp(l) + 3
      do 1240 k = 1, nyp3
      do 1230 j = 1, nx
      q(j,k,l) = 0.
 1230 continue
 1240 continue
 1250 continue
      wke = 0.
c ----------------
c push particles
c ----------------
      call ppush2 (part,fx,fy,npp,noff,qtme,dt,wke,nx,idimp,npmax,nblok,
     1nxv,nypmx,edges,ny)
      print *,'after ppush2',elapsed_time()
c -------------------------------------------------
c move particles into appropriate spatial regions
c -------------------------------------------------
c      call pmove2 (part,edges,npp,sbufr,sbufl,rbufr,rbufl,ihole,jsr,jsl,
c     1jss,ny,kstrt,nvp,idimp,npmax,nblok,idps,nbmax,ntmax,ierr)
c       print *,'after pmove2',elapsed_time() 

c        do  k=1,nblok
c        do i=1,npp(k)
c         do j=1,idimp
c           print *,'part',j,i,k,part(j,i,k)
c           enddo
c           enddo
c           enddo
c           stop

c ----------------
c deposit charge
c ----------------
      call pdost2(part,q,npp,noff,qme,nx,idimp,npmax,nblok,nxv,nypmx)
        print *,'after pdost2',elapsed_time()
c ----------------------
c add background charge
c ----------------------
      do 1280 l = 1, nblok
      do 1270 k = 1, nyp(l)
      do 1260 j = 1, nx
      q(j,k+1,l) = q(j,k+1,l) + qi0
 1260 continue
 1270 continue
 1280 continue
c     call timera(1,'push    ')
      print *,'we-wke',we,wke
 
c --------------------
c energy diagnostic
c --------------------
      wt = we + wke
      write (6,993) we, wke, wt
      itime = itime + 1
      go to 500
 2000 continue
c =====================================
c
c * * * end main iteration loop * * *
c
c =====================================
      write (6,992)
      call timer(timeend)

      print *,'total time = ',timeend-timebegin, 'seconds'

      stop
      end

      real function elapsed_time()
      integer :: clock_count, clock_rate, clock_max
      integer, save :: last_count
      logical, save :: system_clock_not_yet_called = .true.

      call system_clock(count=clock_count,                              
     1                    count_rate=clock_rate, 
     1                  count_max=clock_max)
      if (system_clock_not_yet_called) then
         system_clock_not_yet_called = .false.
         elapsed_time = 0.0
      else if (clock_count > last_count) then
         elapsed_time = real(clock_count-last_count)/real(clock_rate)
      else
         elapsed_time = real(clock_max-last_count+clock_count)          
     1       /real(clock_rate)
      end if
      last_count = clock_count
      return
       end 

      subroutine timer(t)

      integer val(8)
      double precision t

      call date_and_time(values=val)
      t = dble(val(8))*1d-3 + dble(val(7)) +          
     1   dble(val(6))*60d0 + dble(val(5))*3600d0

      end subroutine timer




C =======================================================================
c
c  This subroutine determines spatial boundaries for particle
c  decomposition, calculates number of grid points in each spatial
c  region, and the offset of these grid points from the global address
c
C =======================================================================
      subroutine dcomp2(edges,nyp,noff,ny,kstrt,nvp,idps,nblok)
      dimension edges(idps,nblok)
      dimension nyp(nblok), noff(nblok)
      ks = kstrt - 2
      at1 = float(ny)/float(nvp)
      do 10 l = 1, nblok
      kb = l + ks
      edges(1,l) = at1*float(kb)
      noff(l) = edges(1,l) + .5
      edges(2,l) = at1*float(kb + 1)
      kr = edges(2,l) + .5
      nyp(l) = kr - noff(l)
   10 continue
      return
      end
C =======================================================================
c
c  This subroutine calculates initial particle co-ordinates
c  and velocities with uniform density and maxwellian velocity with drift.
c
C =======================================================================
      subroutine pistr2 (part,edges,npp,nps,vtx,vty,vdx,vdy,npx,npy,nx,n
     1y,idimp,npmax,nblok,idps)
      double precision ranorm
      dimension part(idimp,npmax,nblok)
      dimension edges(idps,nblok), npp(nblok), nps(nblok)
      dimension sum2(2)
      npxy = npx*npy
      at1 = float(nx)/float(npx)
      at2 = float(ny)/float(npy)
      do 30 k = 1, npy
      yt = at2*(float(k) - .5)
      do 20 j = 1, npx
c ---------------------------------
c uniform density profile
c ---------------------------------
      xt = at1*(float(j) - .5)
c ---------------------------------
c maxwellian velocity distribution
c ---------------------------------
      vxt = vtx*ranorm(d)
      vyt = vty*ranorm(d)
      do 10 l = 1, nblok
      if ((yt.ge.edges(1,l)).and.(yt.lt.edges(2,l))) then
         npt = npp(l) + 1
         part(1,npt,l) = xt
         part(2,npt,l) = yt
         part(3,npt,l) = vxt
         part(4,npt,l) = vyt
         npp(l) = npt
      endif
   10 continue
   20 continue
   30 continue
c -------------------
c add correct drift
c -------------------
      sum2(1) = 0.
      sum2(2) = 0.
      do 50 l = 1, nblok
      sum0 = 0.
      sum1 = 0.
      do 40 j = nps(l), npp(l)
      sum0 = sum0 + part(3,j,l)
      sum1 = sum1 + part(4,j,l)
   40 continue
      sum2(1) = sum2(1) + sum0
      sum2(2) = sum2(2) + sum1
   50 continue

      at1 = 1./float(npxy)
      sum2(1) = at1*sum2(1) - vdx
      sum2(2) = at1*sum2(2) - vdy
      do 70 l = 1, nblok
      do 60 j = nps(l), npp(l)
      part(3,j,l) = part(3,j,l) - sum2(1)
      part(4,j,l) = part(4,j,l) - sum2(2)
   60 continue
   70 continue
      return
      end
C =========================================================================
c
c  This subroutine calculates particle charge density
c  using second-order spline interpolation, periodic boundaries.
c
c charge density is approximated by values at the nearest grid points
c
C =========================================================================
      subroutine pdost2(part,q,npp,noff,qm,nx,idimp,npmax,nblok,nxv,
     1   nypmx)

      dimension part(idimp,npmax,nblok), q(nxv,nypmx,nblok)
      dimension npp(nblok), noff(nblok)
      do 20 l = 1, nblok
      mnoff = noff(l) - 2
c -------------------------------
c find interpolation weights
c -------------------------------
      do 10 j = 1, npp(l)
      nn = part(1,j,l) + .5
      dxl = part(1,j,l) - float(nn)
      nn = nn + 1
      if (nn.gt.nx) nn = nn - nx
      amx = qm*(.75 - dxl*dxl)
      np = nn + 1
      if (np.gt.nx) np = np - nx
      dxp = .5*qm*(.5 + dxl)**2
      nl = nn - 1
      if (nl.lt.1) nl = nl + nx
      dxl = .5*qm*(.5 - dxl)**2
      mm = part(2,j,l) + .5
      dyl = part(2,j,l) - float(mm)
      mm = mm - mnoff
      amy = .75 - dyl*dyl
      mp = mm + 1
      dyp = .5*(.5 + dyl)**2
      ml = mm - 1
      dyl = .5*(.5 - dyl)**2
c ----------------
c deposit charge
c ----------------
      q(nn,mm,l) = q(nn,mm,l) + amx*amy
      q(np,mm,l) = q(np,mm,l) + dxp*amy
      q(nl,mm,l) = q(nl,mm,l) + dxl*amy
      q(nn,mp,l) = q(nn,mp,l) + amx*dyp
      q(np,mp,l) = q(np,mp,l) + dxp*dyp
      q(nl,mp,l) = q(nl,mp,l) + dxl*dyp
      q(nn,ml,l) = q(nn,ml,l) + amx*dyl
      q(np,ml,l) = q(np,ml,l) + dxp*dyl
      q(nl,ml,l) = q(nl,ml,l) + dxl*dyl
c     endif
   10 continue
   20 continue
      return
      end

C =========================================================================
c
c  This subroutine copies data between particle and field partitions,
c  where the field and particle partitions are assumed to be the same.
c
c    f(j,k,l) = real data for grid j,k in particle partition l.  
c    fc(j,k,l) = complex data for grid j,k in field partition l.  

c    if isign = -1, data is copied from particle to field partition.
c    if isign = 1, data is copied from field to particle partition.
C =========================================================================
      subroutine cppfp2 (f,fc,isign,scr,kstrt,nvp,nx,nxv,nypmx,nblok,
     1  nxvh,kyp,kblok,idps)

      complex fc, scr
      dimension f(nxv,nypmx,nblok), fc(nxvh,kyp,kblok)
      dimension scr(nxvh,idps,nblok)
      nxh = nx/2
      ks = kstrt - 2
      moff = nypmx*nvp
      if (isign) 10, 100, 110
c --------------------------------------
c copy from particle to field partition
c --------------------------------------
   10 do 40 l = 1, kblok
      do 30 k = 1, kyp
      do 20 j = 1, nxh
      fc(j,k,l) = cmplx(f(2*j-1,k+1,l),f(2*j,k+1,l))
   20 continue
   30 continue
   40 continue

      do 90 l = 1, kblok
      kr = l + ks + 2
      if (kr.gt.nvp) then
         kr = kr - nvp
      endif
      krr = kr
      kl = l + ks
      if (kl.lt.1) then
         kl = kl + nvp
      endif
      kll = kl
      ngc = 2
c ---------------------------------------------
c special case of only one grid per processor
c ---------------------------------------------
      if (kyp.eq.1) then
         krr = krr + 1
         if (krr.gt.nvp) then
            krr = krr - nvp
         endif
         kll = kll - 1
         if (kll.lt.1) then
            kll = kll + nvp
         endif
         ngc = 1
      endif

      do 50 j = 1, nxh
      scr(j,1,l) = cmplx(f(2*j-1,kyp+2,kl),f(2*j,kyp+2,kl))
      scr(j,2,l) = cmplx(f(2*j-1,kyp+3,kll),f(2*j,kyp+3,kll))
   50 continue

      do 60 j = 1, nxh
      fc(j,1,l) = fc(j,1,l) + scr(j,1,l)
      fc(j,ngc,l) = fc(j,ngc,l) + scr(j,2,l)
   60 continue

      do 70 j = 1, nxh
      scr(j,1,l) = cmplx(f(2*j-1,1,kr),f(2*j,1,kr))
   70 continue

      do 80 j = 1, nxh
      fc(j,kyp,l) = fc(j,kyp,l) + scr(j,1,l)
   80 continue
   90 continue
  100 return
c ---------------------------------------
c copy from field to particle partition
c ---------------------------------------
  110 do 140 l = 1, kblok
      do 130 k = 1, kyp
      do 120 j = 1, nxh
      f(2*j-1,k+1,l) = real(fc(j,k,l))
      f(2*j,k+1,l) = aimag(fc(j,k,l))
  120 continue
  130 continue
  140 continue

      do 170 l = 1, kblok
      kr = l + ks + 2
      if (kr.gt.nvp) then
         kr = kr - nvp
      endif
      krr = kr
      kl = l + ks
      if (kl.lt.1) then
         kl = kl + nvp
      endif
      kll = kl
      ngc = 2
c --------------------------------------------
c special case of only one grid per processor
c --------------------------------------------
      if (kyp.eq.1) then
         krr = krr + 1
         if (krr.gt.nvp) then
            krr = krr - nvp
         endif
         kll = kll - 1
         if (kll.lt.1) then
            kll = kll + nvp
         endif
         ngc = 1
      endif

      do 150 j = 1, nxh
      f(2*j-1,1,l) = real(fc(j,kyp,kl))
      f(2*j,1,l) = aimag(fc(j,kyp,kl))
  150 continue

      do 160 j = 1, nxh
      f(2*j-1,kyp+2,l) = real(fc(j,1,kr))
      f(2*j,kyp+2,l) = aimag(fc(j,1,kr))
      f(2*j-1,kyp+3,l) = real(fc(j,ngc,krr))
      f(2*j,kyp+3,l) = aimag(fc(j,ngc,krr))
  160 continue

  170 continue
      return
      end

C ========================================================================
c
c  This subroutine solves 2d poisson's equation in fourier space for
c  force/charge (or convolution of electric field over particle shape)
c  with periodic boundary conditions.
c
c  The equation used is:
c    fx(kx,ky) = -sqrt(-1)*kx*g(kx,ky)*s(kx,ky),
c    fy(kx,ky) = -sqrt(-1)*ky*g(kx,ky)*s(kx,ky),
c    where kx = 2pi*j/nx, ky = 2pi*k/ny, and j,k = fourier mode numbers,
c      g(kx,ky) = (affp/(kx**2+ky**2))*s(kx,ky),
c      s(kx,ky) = exp(-((kx*ax)**2+(ky*ay)**2)/2), except for
c    for fourier mode (jj-1,k-1), where jj = j + kxp*(l - 1)
c
c    if isign = 0, form factor array is prepared
c    if isign is not equal to 0, force/charge is calculated.
c
C ========================================================================
      subroutine ppois2(q,fx,fy,isign,ffc,ax,ay,affp,wp,nx,ny,kstrt,nyv,
     1kxp,jblok,nyhd)
      complex q, fx, fy, ffc, zero
      dimension q(nyv,kxp,jblok)
      dimension fx(nyv,kxp,jblok), fy(nyv,kxp,jblok)
      dimension ffc(nyhd,kxp,jblok)
      nxh = nx/2
      nyh = ny/2
      ny2 = ny + 2
      ks = kstrt - 2
      dnx = 6.28318530717959/float(nx)
      dny = 6.28318530717959/float(ny)
      zero = cmplx(0.,0.)
      if (isign.ne.0) go to 40
      if (kstrt.gt.nxh) return
c ---------------------------
c prepare form factor array
c ---------------------------
      do 30 l = 1, jblok
      joff = kxp*(l + ks) - 1
      do 20 j = 1, kxp
      dkx = dnx*float(j + joff)
      at1 = dkx*dkx
      at2 = (dkx*ax)**2
      do 10 k = 1, nyh
      dky = dny*float(k - 1)
      at3 = dky*dky + at1
      at4 = exp(-.5*((dky*ay)**2 + at2))
      if (at3.eq.0.) then
         ffc(k,j,l) = cmplx(0.,1.)
      else
         ffc(k,j,l) = cmplx(affp*at4/at3,at4)
      endif
   10 continue
   20 continue
   30 continue
      return
c ---------------------------------------------
c calculate force/charge and sum field energy
c ---------------------------------------------
   40 wp = 0.
      if (kstrt.gt.nxh) go to 90
      do 80 l = 1, jblok
c mode numbers 0 < kx < nx/2 and 0 < ky < ny/2
      joff = kxp*(l + ks) - 1
      do 60 j = 1, kxp
      dkx = dnx*float(j + joff)
      if ((j+joff).gt.0) then
         do 50 k = 2, nyh
c        if ((j+joff).gt.0) then
         k1 = ny2 - k
         at1 = real(ffc(k,j,l))*aimag(ffc(k,j,l))
         at2 = dkx*at1
         at3 = dny*float(k - 1)*at1
         fx(k,j,l) = at2*cmplx(aimag(q(k,j,l)),-real(q(k,j,l)))
         fx(k1,j,l) = at2*cmplx(aimag(q(k1,j,l)),-real(q(k1,j,l)))
         fy(k,j,l) = at3*cmplx(aimag(q(k,j,l)),-real(q(k,j,l)))
         fy(k1,j,l) = at3*cmplx(-aimag(q(k1,j,l)),real(q(k1,j,l)))
         wp = wp + at1*(q(k,j,l)*conjg(q(k,j,l)) + q(k1,j,l)*conjg(q(k1,
     1j,l)))
c        endif
   50    continue
      endif
c ---------------------------
c mode numbers ky = 0, ny/2
c ---------------------------
      k1 = nyh + 1
      at1 = real(ffc(1,j,l))*aimag(ffc(1,j,l))
      fx(1,j,l) = dkx*at1*cmplx(aimag(q(1,j,l)),-real(q(1,j,l)))
      fx(k1,j,l) = zero
      fy(1,j,l) = zero
      fy(k1,j,l) = zero
      wp = wp + at1*(q(1,j,l)*conjg(q(1,j,l)))
   60 continue
c ---------------------------
c mode numbers kx = 0, nx/2
c ---------------------------
      if ((l+ks).eq.0) then
         do 70 k = 2, nyh
         k1 = ny2 - k
         at1 = real(ffc(k,1,l))*aimag(ffc(k,1,l))
         fx(k,1,l) = zero
         fx(k1,1,l) = zero
         fy(k,1,l) = dny*float(k - 1)*at1*cmplx(aimag(q(k,1,l)),-real(q(
     1k,1,l)))
         fy(k1,1,l) = zero
         wp = wp + at1*(q(k,1,l)*conjg(q(k,1,l)))
   70    continue
      endif
   80 continue
   90 continue

      wp = float(nx*ny)*wp
      return
      end

C =========================================================================
c
c  This subroutine updates particle co-ordinates and
c  velocities using leap-frog scheme in time and second-order spline
c  interpolation in space, with periodic boundary conditions,
c  The equations used are:
c      vx(t+dt/2) = vx(t-dt/2) + (q/m)*fx(x(t),y(t))*dt,
c      vy(t+dt/2) = vy(t-dt/2) + (q/m)*fy(x(t),y(t))*dt,
c  where q/m is charge/mass, and
c      x(t+dt) = x(t) + vx(t+dt/2)*dt, y(t+dt) = y(t) + vy(t+dt/2)*dt
c      fx(x(t),y(t)) and fy(x(t),y(t)) are approximated by interpolation from
c the nearest grid points.
c 
C =========================================================================
      subroutine ppush2 (part,fx,fy,npp,noff,qtm,dt,ek,nx,idimp,npmax,nb
     1lok,nxv,nypmx,edges,ny)

      dimension part(idimp,npmax,nblok)
      dimension fx(nxv,nypmx,nblok), fy(nxv,nypmx,nblok)
      dimension npp(nblok), noff(nblok),edges(idimp,nblok)
      zero = 0.
      anx = float(nx)
      any = float(ny)
      sum1 = 0.
      do 20 l = 1, nblok
      mnoff = noff(l) - 2
c ---------------------------
c find interpolation weights
c ---------------------------
      do 10 j = 1, npp(l)
      nn = part(1,j,l) + .5
      dx = part(1,j,l) - float(nn)
      nn = nn + 1
      if (nn.gt.nx) nn = nn - nx
      amx = .75 - dx*dx
      np = nn + 1
      if (np.gt.nx) np = np - nx
      dxp = .5*(.5 + dx)**2
      nl = nn - 1
      if (nl.lt.1) nl = nl + nx
      dxl = .5*(.5 - dx)**2
      mm = part(2,j,l) + .5
      dy = part(2,j,l) - float(mm)
      mm = mm - mnoff
      amy = .75 - dy*dy
      mp = mm + 1
      dyp = .5*(.5 + dy)**2
      ml = mm - 1
      dyl = .5*(.5 - dy)**2
c ------------------
c find acceleration
c ------------------
      dx = amy*(amx*fx(nn,mm,l) + dxp*fx(np,mm,l) + dxl*fx(nl,mm,l)) + d
     1yp*(amx*fx(nn,mp,l) + dxp*fx(np,mp,l) + dxl*fx(nl,mp,l)) + dyl*(am
     2x*fx(nn,ml,l)+ dxp*fx(np,ml,l) + dxl*fx(nl,ml,l))
      dy = amy*(amx*fy(nn,mm,l) + dxp*fy(np,mm,l) + dxl*fy(nl,mm,l)) + d
     1yp*(amx*fy(nn,mp,l) + dxp*fy(np,mp,l) + dxl*fy(nl,mp,l)) + dyl*(am
     2x*fy(nn,ml,l)+ dxp*fy(np,ml,l) + dxl*fy(nl,ml,l))
c --------------
c new velocity
c --------------
      dx = part(3,j,l) + qtm*dx
      dy = part(4,j,l) + qtm*dy
c -----------------------
c average kinetic energy
c -----------------------
      sum1 = sum1 + (dx + part(3,j,l))**2 + (dy + part(4,j,l))**2
      part(3,j,l) = dx
      part(4,j,l) = dy
c -------------
c new position
c -------------
      dx = part(1,j,l) + dx*dt
      dy = part(2,j,l) + dy*dt
c ------------------------------
c periodic boundary conditions
c ------------------------------
      if (dx.lt.zero) dx = dx + anx
      if (dx.ge.anx) dx = dx - anx

C For one processor case ,add these following lines
      if (dy.ge.edges(2,l)) then
          dy = dy - any
      else if (dy.lt.edges(1,l)) then
          dy = dy + any
      endif
        
      part(1,j,l) = dx
      part(2,j,l) = dy
   10 continue
   20 continue
c --------------------------
c normalize kinetic energy
c --------------------------
      ek = ek + .125*sum1
      return
      end


C ========================================================================
c
c  This subroutine moves particles into appropriate spatial regions
c   ierr = (0,1) = (no,yes) error condition exists
c
C ========================================================================
      subroutine pmove2 (part,edges,npp,sbufr,sbufl,rbufr,rbufl,ihole,
     1 jsr,jsl,jss,ny,kstrt,nvp,idimp,npmax,nblok,idps,nbmax,ntmax,ierr)
      dimension part(idimp,npmax,nblok)
      dimension edges(idps,nblok), npp(nblok)
      dimension sbufl(idimp,nbmax,nblok), sbufr(idimp,nbmax,nblok)
      dimension rbufl(idimp,nbmax,nblok), rbufr(idimp,nbmax,nblok)
      dimension jsl(idps,nblok), jsr(idps,nblok), jss(idps,nblok)
      dimension ihole(ntmax,nblok)
      any = float(ny)
      ks = kstrt - 2
      iter = 4
c debugging section: count total number of particles before move
      npr = 0
      do 5 l = 1, nblok
      npr = npr + npp(l)
    5 continue
c --------------------------
c buffer outgoing particles
c --------------------------
   10 do 30 l = 1, nblok
      kb = l + ks
      jsl(1,l) = 0
      jsr(1,l) = 0
      jss(2,l) = 0
      do 20 j = 1, npp(l)
      yt = part(2,j,l)
c --------------------
c particles going up
c --------------------
      if (yt.ge.edges(2,l)) then
         if (jsr(1,l).lt.nbmax) then
            jsr(1,l) = jsr(1,l) + 1
            if ((kb+1).eq.nvp) yt = yt - any
            sbufr(1,jsr(1,l),l) = part(1,j,l)
            sbufr(2,jsr(1,l),l) = yt
            sbufr(3,jsr(1,l),l) = part(3,j,l)
            sbufr(4,jsr(1,l),l) = part(4,j,l)
            ihole(jsl(1,l)+jsr(1,l),l) = j
         else
            jss(2,l) = 1
         endif
c ---------------------
c particles going down
c ---------------------
      elseif (yt.lt.edges(1,l)) then
         if (jsl(1,l).lt.nbmax) then
            jsl(1,l) = jsl(1,l) + 1
            if (kb.eq.0) yt = yt + any
            sbufl(1,jsl(1,l),l) = part(1,j,l)
            sbufl(2,jsl(1,l),l) = yt
            sbufl(3,jsl(1,l),l) = part(3,j,l)
            sbufl(4,jsl(1,l),l) = part(4,j,l)
            ihole(jsl(1,l)+jsr(1,l),l) = j
         else
            jss(2,l) = 1
         endif
      endif
   20 continue
      jss(1,l) = jsl(1,l) + jsr(1,l)
   30 continue
c ---------------------------------
c check for full buffer condition
c ---------------------------------
      ibflg = 0
      do 40 l = 1, nblok
      ibflg = ibflg + jss(2,l)
   40 continue
c -----------------------
c copy particle buffers
c -----------------------
   50 iter = iter + 4
      do 80 l = 1, nblok
c --------------------------
c get particles from below
c --------------------------
      kl = l + ks
      if (kl.lt.1) then
         jsl(2,l) = jsr(1,kl+nvp)
      else
         jsl(2,l) = jsr(1,kl)
      endif

      do 60 j = 1, jsl(2,l)
      if (kl.lt.1) then
         rbufl(1,j,l) = sbufr(1,j,kl+nvp)
         rbufl(2,j,l) = sbufr(2,j,kl+nvp)
         rbufl(3,j,l) = sbufr(3,j,kl+nvp)
         rbufl(4,j,l) = sbufr(4,j,kl+nvp)
      else
         rbufl(1,j,l) = sbufr(1,j,kl)
         rbufl(2,j,l) = sbufr(2,j,kl)
         rbufl(3,j,l) = sbufr(3,j,kl)
         rbufl(4,j,l) = sbufr(4,j,kl)
      endif
   60 continue
c ---------------------------
c get particles from above
c ---------------------------
      kr = l + ks + 2
      if (kr.gt.nvp) then
         jsr(2,l) = jsl(1,kr-nvp)
      else
         jsr(2,l) = jsl(1,kr)
      endif
      do 70 j = 1, jsr(2,l)
      if (kr.gt.nvp) then
         rbufr(1,j,l) = sbufl(1,j,kr-nvp)
         rbufr(2,j,l) = sbufl(2,j,kr-nvp)
         rbufr(3,j,l) = sbufl(3,j,kr-nvp)
         rbufr(4,j,l) = sbufl(4,j,kr-nvp)
      else
         rbufr(1,j,l) = sbufl(1,j,kr)
         rbufr(2,j,l) = sbufl(2,j,kr)
         rbufr(3,j,l) = sbufl(3,j,kr)
         rbufr(4,j,l) = sbufl(4,j,kr)
      endif
   70 continue

   80 continue
c ------------------------------------------
c check if particles must be passed further
c ------------------------------------------
      nps = 0
      do 110 l = 1, nblok
c --------------------------------------------------------------
c check if any particles coming from above must go further down
c --------------------------------------------------------------
      jsl(1,l) = 0
      do 90 j = 1, jsr(2,l)
      if (rbufr(2,j,l).lt.edges(1,l)) jsl(1,l) = 1
   90 continue
c ------------------------------------------------------------
c check if any particles coming from below must go further up
c ------------------------------------------------------------
      jsr(1,l) = 0
      do 100 j = 1, jsl(2,l)
      if (rbufl(2,j,l).ge.edges(2,l)) jsr(1,l) = 1
  100 continue
      nps = nps + (jsl(1,l) + jsr(1,l))
  110 continue

      if (nps.eq.0) go to 150
c -------------------------------------------
c remove particles which do not belong here
c -------------------------------------------
      do 140 l = 1, nblok
      kb = l + ks
c -----------------------------------------
c first check particles coming from above
c -----------------------------------------
      jsl(1,l) = 0
      jss(2,l) = 0
      do 120 j = 1, jsr(2,l)
      yt = rbufr(2,j,l)
c ----------------------
c particles going down
c ----------------------
      if (yt.lt.edges(1,l)) then
         jsl(1,l) = jsl(1,l) + 1
         if (kb.eq.0) yt = yt + any
         sbufl(1,jsl(1,l),l) = rbufr(1,j,l)
         sbufl(2,jsl(1,l),l) = yt
         sbufl(3,jsl(1,l),l) = rbufr(3,j,l)
         sbufl(4,jsl(1,l),l) = rbufr(4,j,l)
c ----------------------
c particles staying here
c ----------------------
      else
         jss(2,l) = jss(2,l) + 1
         rbufr(1,jss(2,l),l) = rbufr(1,j,l)
         rbufr(2,jss(2,l),l) = yt
         rbufr(3,jss(2,l),l) = rbufr(3,j,l)
         rbufr(4,jss(2,l),l) = rbufr(4,j,l)
      endif
  120 continue
      jsr(2,l) = jss(2,l)
c ---------------------------------------
c next check particles coming from below
c ---------------------------------------
      jsr(1,l) = 0
      jss(2,l) = 0
      do 130 j = 1, jsl(2,l)
      yt = rbufl(2,j,l)
c -------------------
c particles going up
c -------------------
      if (yt.ge.edges(2,l)) then
         jsr(1,l) = jsr(1,l) + 1
         if ((kb+1).eq.nvp) yt = yt - any
         sbufr(1,jsr(1,l),l) = rbufl(1,j,l)
         sbufr(2,jsr(1,l),l) = yt
         sbufr(3,jsr(1,l),l) = rbufl(3,j,l)
         sbufr(4,jsr(1,l),l) = rbufl(4,j,l)
c ------------------------
c particles staying here
c ------------------------
      else
         jss(2,l) = jss(2,l) + 1
         rbufl(1,jss(2,l),l) = rbufl(1,j,l)
         rbufl(2,jss(2,l),l) = yt
         rbufl(3,jss(2,l),l) = rbufl(3,j,l)
         rbufl(4,jss(2,l),l) = rbufl(4,j,l)
      endif
  130 continue
      jsl(2,l) = jss(2,l)
  140 continue
c ---------------------------------------------
c check if move would overflow particle array
c ---------------------------------------------
  150 ierr = 0
      do 160 l = 1, nblok
      if ((npp(l)+jsl(2,l)+jsr(2,l)-jss(1,l)).gt.npmax) then
         jss(2,l) = 1
      else
         jss(2,l) = 0
      endif
      ierr = ierr + jss(2,l)
  160 continue

      if (ierr.gt.0) then
         write (6,*) 'particle overflow error, ierr =', ierr
         return
      endif
      do 200 l = 1, nblok
c ---------------------------------------------------
c distribute incoming particles from buffers
c ---------------------------------------------------
      jss(2,l) = min0(jss(1,l),jsl(2,l))
      do 170 j = 1, jss(2,l)
      part(1,ihole(j,l),l) = rbufl(1,j,l)
      part(2,ihole(j,l),l) = rbufl(2,j,l)
      part(3,ihole(j,l),l) = rbufl(3,j,l)
      part(4,ihole(j,l),l) = rbufl(4,j,l)
  170 continue
      if (jss(1,l).gt.jsl(2,l)) then
         jss(2,l) = min0(jss(1,l)-jsl(2,l),jsr(2,l))
      else
         jss(2,l) = jsl(2,l) - jss(1,l)
      endif
      do 180 j = 1, jss(2,l)
      if (jss(1,l).gt.jsl(2,l)) then
         part(1,ihole(j+jsl(2,l),l),l) = rbufr(1,j,l)
         part(2,ihole(j+jsl(2,l),l),l) = rbufr(2,j,l)
         part(3,ihole(j+jsl(2,l),l),l) = rbufr(3,j,l)
         part(4,ihole(j+jsl(2,l),l),l) = rbufr(4,j,l)
      else
c ------------------------------------------------------
c no more holes
c distribute remaining particles from below into bottom
c ------------------------------------------------------
         part(1,j+npp(l),l) = rbufl(1,j+jss(1,l),l)
         part(2,j+npp(l),l) = rbufl(2,j+jss(1,l),l)
         part(3,j+npp(l),l) = rbufl(3,j+jss(1,l),l)
         part(4,j+npp(l),l) = rbufl(4,j+jss(1,l),l)
      endif
  180 continue
      if (jss(1,l).le.jsl(2,l)) then
         npp(l) = npp(l) + (jsl(2,l) - jss(1,l))
         jss(1,l) = jsl(2,l)
      endif
      jss(2,l) = jss(1,l) - (jsl(2,l) + jsr(2,l))
      if (jss(2,l).gt.0) then
         jss(1,l) = (jsl(2,l) + jsr(2,l))
         jsr(2,l) = jss(2,l)
      else
         jss(1,l) = jss(1,l) - jsl(2,l)
         jsr(2,l) = -jss(2,l)
      endif
      do 190 j = 1, jsr(2,l)
c ---------------------------------------------------------------------
c holes left over
c fill up remaining holes in particle array with particles from bottom
c ---------------------------------------------------------------------
      if (jss(2,l).gt.0) then
         j1 = npp(l) - j + 1
         j2 = jss(1,l) + jss(2,l) - j + 1
         if (j1.gt.ihole(j2,l)) then
c ------------------------------------------------
c move particle only if it is below current hole
c ------------------------------------------------
            part(1,ihole(j2,l),l) = part(1,j1,l)
            part(2,ihole(j2,l),l) = part(2,j1,l)
            part(3,ihole(j2,l),l) = part(3,j1,l)
            part(4,ihole(j2,l),l) = part(4,j1,l)
         endif
      else
c --------------------------------------------------------
c no more holes
c distribute remaining particles from above into bottom
c --------------------------------------------------------
         part(1,j+npp(l),l) = rbufr(1,j+jss(1,l),l)
         part(2,j+npp(l),l) = rbufr(2,j+jss(1,l),l)
         part(3,j+npp(l),l) = rbufr(3,j+jss(1,l),l)
         part(4,j+npp(l),l) = rbufr(4,j+jss(1,l),l)
      endif
  190 continue
      if (jss(2,l).gt.0) then
         npp(l) = npp(l) - jsr(2,l)
      else
         npp(l) = npp(l) + jsr(2,l)
      endif
      jss(1,l) = 0
  200 continue
c -------------------------------------------------
c check if any particles have to be passed further
c -------------------------------------------------
      if (nps.gt.0) go to 50
c -------------------------------------------------------------------
c check if buffer overflowed and more particles remain to be checked
c -------------------------------------------------------------------
      if (ibflg.gt.0) go to 10
c ---------------------------------------------------------------
c debugging section: count total number of particles after move
c ---------------------------------------------------------------
      nps = 0
      do 205 l = 1, nblok
      nps = nps + npp(l)
  205 continue
      if (npr.ne.nps) write (6,*) 'number error, np-old,np-new=',npr,nps
      return
      end

C =======================================================================
c
c This subroutine performs a two dimensional real to complex fast
c fourier transform and its inverse, using complex arithmetic,
c
c   indx/indy = exponent which determines length in x/y direction,
c   where nx=2**indx, ny=2**indy
c   ntpose = (0,1) = (no,yes) input, output data are transposed
c
c   if isign = 0, the fft tables are prepared
c   if isign = -1, an inverse fourier transform is performed
c   if ntpose = 0, f is the input and output array, g is a scratch array
c       f(n,m,l) = (1/nx*ny)*sum(f(j,k,i)*
c       exp(-sqrt(-1)*2pi*n*j/nx)*exp(-sqrt(-1)*2pi*mm*kk/ny))
c       where mm = m + kyp*(l - 1) and kk = k + kyp*(i - 1)
c   if ntpose = 1, f is the input and g is the output
c       g(m,n,l) = (1/nx*ny)*sum(f(j,k,i)*
c       exp(-sqrt(-1)*2pi*nn*j/nx)*exp(-sqrt(-1)*2pi*m*kk/ny))
c       where nn = n + kxp*(l - 1) and kk = k + kyp*(i - 1)
c   if isign = 1, a forward fourier transform is performed
c   if ntpose = 0, f is the input and output array, g is a scratch array
c   f(j,k,i) = sum(f(n,m,l)*exp(sqrt(-1)*2pi*n*j/nx)*
c       exp(sqrt(-1)*2pi*mm*kk/ny))
c   where mm = m + kyp*(l - 1) and kk = k + kyp*(i - 1)
c   if ntpose = 1, g is the input and f is the output
c   f(j,k,i) = sum(g(m,n,l)*exp(sqrt(-1)*2pi*nn*j/nx)*
c       exp(sqrt(-1)*2pi*m*kk/ny))
c   where nn = n + kxp*(l - 1) and kk = k + kyp*(i - 1)
c   g(k,1,1) = mode nx/2,k-1, where ny/2+2 <= k <= ny, and
c   imaginary part of g(1,1,1) = real part of mode nx/2,0 and
c   imaginary part of g(ny/2+1,1,1) = real part of mode nx/2,ny/2
c   written by viktor k. decyk, ucla
C =========================================================================
      subroutine pfft2r(f,g,bs,br,isign,ntpose,mixup,sct,indx,indy,
     1 kstrt,nxvh,nyv,kxp,kyp,jblok,kblok,nxhyd,nxyhd)
      complex f, g, bs, br, sct, s, t, t1
      dimension f(nxvh,kyp,kblok), g(nyv,kxp,jblok)
      dimension bs(kxp,kyp,kblok), br(kxp,kyp,jblok)
      dimension mixup(nxhyd), sct(nxyhd)
      indx1 = indx - 1
      indx1y = max0(indx1,indy)
      nx = 2**indx
      nxh = nx/2
      nxhh = nx/4
      nxh2 = nxh + 2
      ny = 2**indy
      nyh = ny/2
      ny2 = ny + 2
      nxy = max0(nx,ny)
      nxhy = 2**indx1y
      ks = kstrt - 2
      if (isign) 50, 10, 300
c ------------------------------------------------------------------
c prepare fft tables
c bit-reverse index table: mixup(j) = 1 + reversed bits of (j - 1)
c ------------------------------------------------------------------
   10 do 30 j = 1, nxhy
      lb = j - 1
      ll = 0
      do 20 k = 1, indx1y
      jb = lb/2
      it = lb - 2*jb
      lb = jb
      ll = 2*ll + it
   20 continue
      mixup(j) = ll + 1
   30 continue
c ---------------------------------------------
c sine/cosine table for the angles 2*n*pi/nxy
c ---------------------------------------------
      nxyh = nxy/2
      dnxy = 6.28318530717959/float(nxy)
      do 40 j = 1, nxyh
      arg = dnxy*float(j - 1)
      sct(j) = cmplx(cos(arg),-sin(arg))
   40 continue
      return
c ----------------------------
c inverse fourier transform
c ----------------------------
   50 if (kstrt.gt.ny) go to 180
      nrx = nxhy/nxh
      do 80 l = 1, kblok
c --------------------------------
c bit-reverse array elements in x
c --------------------------------
      do 70 j = 1, nxh
      j1 = (mixup(j) - 1)/nrx + 1
      if (j.ge.j1) go to 70
      do 60 k = 1, kyp
      t = f(j1,k,l)
      f(j1,k,l) = f(j,k,l)
      f(j,k,l) = t
   60 continue
   70 continue
   80 continue
c -----------------------
c first transform in x
c -----------------------
      nrx = nxy/nxh
      do 130 m = 1, indx1
      ns = 2**(m - 1)
      ns2 = ns + ns
      km = nxhh/ns
      kmr = km*nrx
      do 120 l = 1, kblok
      do 110 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 100 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = sct(1+kmr*(j-1))
      do 90 i = 1, kyp
      t = s*f(j2,i,l)
      f(j2,i,l) = f(j1,i,l) - t
      f(j1,i,l) = f(j1,i,l) + t
   90 continue
  100 continue
  110 continue
  120 continue
  130 continue
c---------------------------------------
c unscramble coefficients and normalize
c ---------------------------------------
      kmr = nxy/nx
      ani = 1./float(2*nx*ny)
      do 170 l = 1, kblok
      do 150 j = 2, nxhh
      t1 = cmplx(aimag(sct(1+kmr*(j-1))),-real(sct(1+kmr*(j-1))))
      do 140 k = 1, kyp
      t = conjg(f(nxh2-j,k,l))
      s = f(j,k,l) + t
      t = (f(j,k,l) - t)*t1
      f(j,k,l) = ani*(s + t)
      f(nxh2-j,k,l) = ani*conjg(s - t)
  140 continue
  150 continue
      do 160 k = 1, kyp
      f(nxhh+1,k,l) = 2.*ani*conjg(f(nxhh+1,k,l))
      f(1,k,l) = 2.*ani*cmplx(real(f(1,k,l)) + aimag(f(1,k,l)),real(f(1,
     1k,l)) - aimag(f(1,k,l)))
  160 continue
  170 continue
c ------------------------
c transpose f array to g
c ------------------------
     
  180 call ptpose(f,g,bs,br,nxh,ny,kstrt,nxvh,nyv,kxp,kyp,jblok,kblok)
    
      if (kstrt.gt.nxh) go to 290
      nry = nxhy/ny
      do 210 l = 1, jblok
c ---------------------------------
c bit-reverse array elements in y
c ---------------------------------
      do 200 k = 1, ny
      k1 = (mixup(k) - 1)/nry + 1
      if (k.ge.k1) go to 200
      do 190 j = 1, kxp
      t = g(k1,j,l)
      g(k1,j,l) = g(k,j,l)
      g(k,j,l) = t

  190 continue
  200 continue
  210 continue
c ---------------------
c then transform in y
c ---------------------
      nry = nxy/ny
      do 260 m = 1, indy
      ns = 2**(m - 1)
      ns2 = ns + ns
      km = nyh/ns
      kmr = km*nry
      do 250 l = 1, jblok
      do 240 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 230 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = sct(1+kmr*(j-1))
      do 220 i = 1, kxp
      t = s*g(j2,i,l)
      g(j2,i,l) = g(j1,i,l) - t
      g(j1,i,l) = g(j1,i,l) + t
  220 continue
  230 continue
  240 continue
  250 continue
  260 continue
c -------------------------------
c unscramble modes kx = 0, nx/2
c -------------------------------
      do 280 l = 1, jblok
      if ((l+ks).gt.0) go to 280
      do 270 k = 2, nyh
c     if ((l+ks).eq.0) then
      s = g(ny2-k,1,l)
      g(ny2-k,1,l) = .5*cmplx(aimag(g(k,1,l) + s),real(g(k,1,l) - s))
      g(k,1,l) = .5*cmplx(real(g(k,1,l) + s),aimag(g(k,1,l) - s))
c     endif
  270 continue
  280 continue
c ------------------------
c transpose g array to f
c ------------------------
  290 if (ntpose.eq.0) call ptpose(g,f,br,bs,ny,nxh,kstrt,nyv,nxvh,kyp,k
     1xp,kblok,jblok)
      
c ---------------------------
c forward fourier transform
c transpose f array to g
c ---------------------------
      return
  300 if (ntpose.eq.0) call ptpose(f,g,bs,br,nxh,ny,kstrt,nxvh,nyv,kxp,k
     1yp,jblok,kblok)
      
      if (kstrt.gt.nxh) go to 410
      nry = nxhy/ny
      do 350 l = 1, jblok
c ---------------------------
c scramble modes kx = 0, nx/2
c ---------------------------
      if ((l+ks).gt.0) go to 320
      do 310 k = 2, nyh
      s = cmplx(aimag(g(ny2-k,1,l)),real(g(ny2-k,1,l)))
      g(ny2-k,1,l) = conjg(g(k,1,l) - s)
      g(k,1,l) = g(k,1,l) + s
  310 continue
c ----------------------------------
c bit-reverse array elements in y
c ----------------------------------
  320 do 340 k = 1, ny
      k1 = (mixup(k) - 1)/nry + 1
      if (k.ge.k1) go to 340
      do 330 j = 1, kxp
      t = g(k1,j,l)
      g(k1,j,l) = g(k,j,l)
      g(k,j,l) = t
  330 continue
  340 continue
  350 continue
c ----------------------
c first transform in y
c ----------------------
      nry = nxy/ny
      do 400 m = 1, indy
      ns = 2**(m - 1)
      ns2 = ns + ns
      km = nyh/ns
      kmr = km*nry
      do 390 l = 1, jblok
      do 380 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 370 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = conjg(sct(1+kmr*(j-1)))
      do 360 i = 1, kxp
      t = s*g(j2,i,l)
      g(j2,i,l) = g(j1,i,l) - t
      g(j1,i,l) = g(j1,i,l) + t
  360 continue
  370 continue
  380 continue
  390 continue
  400 continue
c ------------------------
c transpose g array to f
c ------------------------
    
  410 call ptpose(g,f,br,bs,ny,nxh,kstrt,nyv,nxvh,kyp,kxp,kblok,jblok)
      
      if (kstrt.gt.ny) return
      nrx = nxhy/nxh
      kmr = nxy/nx
      do 470 l = 1, kblok
c ----------------------
c scramble coefficients
c ----------------------
      do 430 j = 2, nxhh
      t1 = cmplx(aimag(sct(1+kmr*(j-1))),real(sct(1+kmr*(j-1))))
      do 420 k = 1, kyp
      t = conjg(f(nxh2-j,k,l))
      s = f(j,k,l) + t
      t = (f(j,k,l) - t)*t1
      f(j,k,l) = s + t
      f(nxh2-j,k,l) = conjg(s - t)
  420 continue
  430 continue
      do 440 k = 1, kyp
      f(nxhh+1,k,l) = 2.*conjg(f(nxhh+1,k,l))
      f(1,k,l) = cmplx(real(f(1,k,l)) + aimag(f(1,k,l)),real(f(1,k,l)) -
     1 aimag(f(1,k,l)))
  440 continue
c ---------------------------------
c bit-reverse array elements in x
c ---------------------------------
      do 460 j = 1, nxh
      j1 = (mixup(j) - 1)/nrx + 1
      if (j.ge.j1) go to 460
      do 450 k = 1, kyp
      t = f(j1,k,l)
      f(j1,k,l) = f(j,k,l)
      f(j,k,l) = t
  450 continue
  460 continue
  470 continue
c ---------------------
c then transform in x
c ---------------------
      nrx = nxy/nxh
      do 520 m = 1, indx1
      ns = 2**(m - 1)
      ns2 = ns + ns
      km = nxhh/ns
      kmr = km*nrx
      do 510 l = 1, kblok
      do 500 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 490 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = conjg(sct(1+kmr*(j-1)))
      do 480 i = 1, kyp
      t = s*f(j2,i,l)
      f(j2,i,l) = f(j1,i,l) - t
      f(j1,i,l) = f(j1,i,l) + t
  480 continue
  490 continue
  500 continue
  510 continue
  520 continue
      return
      end
C ========================================================================
c
c  This subroutine performs a transpose of a matrix f, distributed in y,
c  to a matrix g, distributed in x, that is,
c    g(k+kyp*(m-1),j,l) = f(j+kxp*(l-1),k,m), where
c    1 <= j <= kxp, 1 <= k <= kyp, 1 <= l <= nx/kxp, 1 <= m <= ny/kyp
c  and where indices l and m can be distributed across processors.
c    f = complex input array
c    g = complex output array
c    s, t = complex scratch arrays
C ========================================================================
      subroutine ptpose(f,g,s,t,nx,ny,kstrt,nxv,nyv,kxp,kyp,jblok,kblok)
      complex f, g, s, t
      dimension f(nxv,kyp,kblok), g(nyv,kxp,jblok)
      dimension s(kxp,kyp,kblok), t(kxp,kyp,jblok)
      ks = kstrt - 2
      kxb = nx/kxp
      kyb = ny/kyp


      if (kstrt.gt.nx) return
      do 40 l = 1, jblok
      joff = kxp*(l + ks)
      do 30 i = 1, kyb
      koff = kyp*(i - 1)
      do 20 k = 1, kyp
      do 10 j = 1, kxp
      g(k+koff,j,l) = f(j+joff,k,i)
   10 continue
   20 continue
   30 continue
   40 continue

      return
      end

      function ranorm(d)
c this program calculates a random number y from a gaussian distribution
c with zero mean and unit variance, according to the method of
c mueller and box:
c    y(k) = (-2*ln(x(k)))**1/2*sin(2*pi*x(k+1))
c    y(k+1) = (-2*ln(x(k)))**1/2*cos(2*pi*x(k+1)),
c where x is a random number uniformly distributed on (0,1).
c written for the ibm by viktor k. decyk, ucla
      integer r1,r2,r4,r5
      double precision ranorm,h1l,h1u,h2l,r0,r3,asc,bsc,temp
      save iflg,r1,r2,r4,r5,h1l,h1u,h2l,r0
      data r1,r2,r4,r5 /885098780,1824280461,1396483093,55318673/
      data h1l,h1u,h2l /65531.0d0,32767.0d0,65525.0d0/
      data iflg,r0 /0,0.0d0/
      if (iflg.eq.0) go to 10
      ranorm = r0
      r0 = 0.0d0
      iflg = 0
      return
   10 isc = 65536
      asc = dfloat(isc)
      bsc = asc*asc
      i1 = r1 - (r1/isc)*isc
      r3 = h1l*dfloat(r1) + asc*h1u*dfloat(i1)
      i1 = r3/bsc
      r3 = r3 - dfloat(i1)*bsc
      bsc = 0.5d0*bsc
      i1 = r2/isc
      isc = r2 - i1*isc
      r0 = h1l*dfloat(r2) + asc*h1u*dfloat(isc)
      asc = 1.0d0/bsc
      isc = r0*asc
      r2 = r0 - dfloat(isc)*bsc
      r3 = r3 + (dfloat(isc) + 2.0d0*h1u*dfloat(i1))
      isc = r3*asc
      r1 = r3 - dfloat(isc)*bsc
      temp = dsqrt(-2.0d0*dlog((dfloat(r1) + dfloat(r2)*asc)*asc))
      isc = 65536
      asc = dfloat(isc)
      bsc = asc*asc
      i1 = r4 - (r4/isc)*isc
      r3 = h2l*dfloat(r4) + asc*h1u*dfloat(i1)
      i1 = r3/bsc
      r3 = r3 - dfloat(i1)*bsc
      bsc = 0.5d0*bsc
      i1 = r5/isc
      isc = r5 - i1*isc
      r0 = h2l*dfloat(r5) + asc*h1u*dfloat(isc)
      asc = 1.0d0/bsc
      isc = r0*asc
      r5 = r0 - dfloat(isc)*bsc
      r3 = r3 + (dfloat(isc) + 2.0d0*h1u*dfloat(i1))
      isc = r3*asc
      r4 = r3 - dfloat(isc)*bsc
      r0 = 6.28318530717959d0*((dfloat(r4) + dfloat(r5)*asc)*asc)
      ranorm = temp*dsin(r0)
      r0 = temp*dcos(r0)
      iflg = 1
      return
      end

      SUBROUTINE TIMERA(T)
C   ==================================================================
C   ===                                                            ===
C   ===    This subroutine returns the elapsed time in seconds     ===
C   ===                    since its first CALL                    ===
C   ===                                                            ===
C   ==================================================================
 
      real T
      real secs, tarray(2)
c      secs = etime(tarray)
      t = secs
      return

      END
