C =======================================================================
c
c  This subroutine calculates initial particle co-ordinates
c  and velocities with uniform density and maxwellian velocity with drift.
c
C =======================================================================
      subroutine dist_particle3 (part,edges,npp,nps,vtx,vty,vtz,vdx,
     1 vdy,vdz,npx,npy,npz,nx,ny,nz,idimp,npmax,nblok,idps,sum0,sum1,
     1  sum2)
 
      double precision ranorm
      dimension part(idimp,npmax,nblok)
      dimension edges(idps,nblok), npp(nblok), nps(nblok)
      real sum0(nblok),sum1(nblok),sum2(nblok)

      integer myprocnum,p,offset
      include 'pglocal.f'

      myprocnum = pghpf_myprocnum()
      item_cnt = npz/nblok

    
      npxyz = npx*npy*npz
      at1 = float(nx)/float(npx)
      at2 = float(ny)/float(npy)
      at3 = float(nz)/float(npz)
     
      if (myprocnum .eq. nblok -1 ) then
      do 33 p = 0, nblok-1
      num_of_particle = 0
      offset = npp(1)
      do 44 ki = 1, item_cnt
      zt = at3*(float(ki+p*item_cnt) - .5)
      do 30 j = 1, npy
      yt = at2*(float(j) - .5)
      do 20 k = 1, npx
c ---------------------------------
c uniform density profile
c ---------------------------------
      xt = at1*(float(k) - .5)
c ---------------------------------
c maxwellian velocity distribution
c ---------------------------------
      vxt = vtx*ranorm(d)
      vyt = vty*ranorm(d)
      vzt = vtz*ranorm(d)

      num_of_particle = num_of_particle + 1 
      part(1,num_of_particle+offset,1) = xt
      part(2,num_of_particle+offset,1) = yt
      part(3,num_of_particle+offset,1) = zt
      part(4,num_of_particle+offset,1) = vxt
      part(5,num_of_particle+offset,1) = vyt
      part(6,num_of_particle+offset,1) = vzt

   20 continue
   30 continue
   44 continue

      if ( p .lt. nblok-1) then
         call pghpf_csend(p,num_of_particle,1,1,PGLCL_INT4)
         call pghpf_csend(p,part(1,offset+1,1),6*num_of_particle,
     *   1,PGLCL_REAL4)
      else
         npp(1) = npp(1) + num_of_particle
      endif
   33 continue  
      else
    
        call pghpf_crecv(nblok-1,num_of_particle,1,1,
     *         PGLCL_INT4)
        offset = npp(1)+1
        call pghpf_crecv(nblok-1,part(1,offset,1),6*num_of_particle,
     *     1, PGLCL_REAL4) 
        npp(1)= npp(1) + num_of_particle
      endif

      sum0(1) = 0.
      sum1(1) = 0.
      sum2(1) = 0.
      do j = nps(1), npp(1)
        sum0(1) = sum0(1) + part(4,j,1)
        sum1(1) = sum1(1) + part(5,j,1)
        sum2(1) = sum2(1) + part(6,j,1)
      enddo
      return
      end

C =========================================================================
C =========================================================================
       
      subroutine pist_extrinsic(part,npp,nps,idimp,nblok,npmax,sum3)
       
      real part(idimp,npmax,nblok),sum3(3)
      integer npp(nblok), nps(nblok)
      include 'pglocal.f'
       
      do j = nps(1), npp(1)
         part(4,j,1) = part(4,j,1) - sum3(1)
         part(5,j,1) = part(5,j,1) - sum3(2)
         part(6,j,1) = part(6,j,1) - sum3(3)
      enddo
      return
      end

C =========================================================================
C =========================================================================
 

C =========================================================================
c
c  This subroutine calculates particle charge density
c  using second-order spline interpolation, periodic boundaries.
c
c charge density is approximated by values at the nearest grid points
c
C =========================================================================
  
      subroutine pdost3(part,q,npp,noff,qm,nx,ny,idimp,npmax,nblok,nxv,
     1   nyv,nzpmx)

      dimension part(idimp,npmax,nblok), q(nxv,nyv,nzpmx,nblok)
      dimension npp(nblok), noff(nblok)
      include 'pglocal.f'
      integer ip

      ip = pghpf_myprocnum()
 
       
      do k=1,nxv
       do i=1,nyv
        do j=1,nzpmx
          q(k,i,j,1) = 0.0
        enddo
       enddo
      enddo


      l = 1
      mnoff = noff(l) - 2
c -------------------------------
c find interpolation weights
c -------------------------------
   
      do 10 j = 1, npp(l)
    

      nn = part(1,j,l) + .5
      dxl = part(1,j,l) - float(nn)
      nn = nn + 1
      if (nn.gt.nx) nn = nn - nx
      amx = qm*(.75 - dxl*dxl)
      np = nn + 1
      if (np.gt.nx) np = np - nx
      dxp = .5*qm*(.5 + dxl)**2
      nl = nn - 1
      if (nl.lt.1) nl = nl + nx
      dxl = .5*qm*(.5 - dxl)**2
      
      kk = part(2,j,l) + .5
      dyl = part(2,j,l) - float(kk)
      kk = kk + 1
      if (kk.gt.ny) kk = kk - ny
      amy = qm*(.75 - dyl*dyl)
      kp = kk + 1
      if (kp.gt.ny) kp = kp - ny
      dyp = .5*qm*(.5 + dyl)**2
      kl = kk - 1
      if (kl.lt.1) kl = kl + ny
      dyl = .5*qm*(.5 - dyl)**2


      mm = part(3,j,l) + .5
      dzl = part(3,j,l) - float(mm)
      mm = mm - mnoff
      amz = .75 - dzl*dzl
      mp = mm + 1
      dzp = .5*(.5 + dzl)**2
      ml = mm - 1
      dzl = .5*(.5 - dzl)**2
c ----------------
c deposit charge
c ----------------
      q(nn,kk,mm,l) = q(nn,kk,mm,l) + amx*amy*amz
      q(np,kk,mm,l) = q(np,kk,mm,l) + dxp*amy*amz
      q(nl,kk,mm,l) = q(nl,kk,mm,l) + dxl*amy*amz
      q(nn,kp,mm,l) = q(nn,kp,mm,l) + amx*dyp*amz
      q(np,kp,mm,l) = q(np,kp,mm,l) + dxp*dyp*amz
      q(nl,kp,mm,l) = q(nl,kp,mm,l) + dxl*dyp*amz
      q(nn,kl,mm,l) = q(nn,kl,mm,l) + amx*dyl*amz
      q(np,kl,mm,l) = q(np,kl,mm,l) + dxp*dyl*amz
      q(nl,kl,mm,l) = q(nl,kl,mm,l) + dxl*dyl*amz

      q(nn,kk,mp,l) = q(nn,kk,mp,l) + amx*amy*dzp
      q(np,kk,mp,l) = q(np,kk,mp,l) + dxp*amy*dzp
      q(nl,kk,mp,l) = q(nl,kk,mp,l) + dxl*amy*dzp
      q(nn,kp,mp,l) = q(nn,kp,mp,l) + amx*dyp*dzp
      q(np,kp,mp,l) = q(np,kp,mp,l) + dxp*dyp*dzp
      q(nl,kp,mp,l) = q(nl,kp,mp,l) + dxl*dyp*dzp
      q(nn,kl,mp,l) = q(nn,kl,mp,l) + amx*dyl*dzp
      q(np,kl,mp,l) = q(np,kl,mp,l) + dxp*dyl*dzp
      q(nl,kl,mp,l) = q(nl,kl,mp,l) + dxl*dyl*dzp

      q(nn,kk,ml,l) = q(nn,kk,ml,l) + amx*amy*dzl
      q(np,kk,ml,l) = q(np,kk,ml,l) + dxp*amy*dzl
      q(nl,kk,ml,l) = q(nl,kk,ml,l) + dxl*amy*dzl
      q(nn,kp,ml,l) = q(nn,kp,ml,l) + amx*dyp*dzl
      q(np,kp,ml,l) = q(np,kp,ml,l) + dxp*dyp*dzl
      q(nl,kp,ml,l) = q(nl,kp,ml,l) + dxl*dyp*dzl
      q(nn,kl,ml,l) = q(nn,kl,ml,l) + amx*dyl*dzl
      q(np,kl,ml,l) = q(np,kl,ml,l) + dxp*dyl*dzl
      q(nl,kl,ml,l) = q(nl,kl,ml,l) + dxl*dyl*dzl
  

   10 continue

      return
      end

C =========================================================================
C =========================================================================

      subroutine do_forall_1 (f,fc,nxv,nyv,nzpmx,nblok,nxvh,
     1 nyvh,kzp,kblok,nxh,nyh)
  
      complex fc
      dimension f(nxv,nyv,nzpmx,nblok), fc(nxvh,nyvh,kzp,kblok)
      include 'pglocal.f' 
c ---------------------------------------
c copy from field to particle partition
c ---------------------------------------

      l = 1
      do 130 k = 1, kzp
      do 120 j = 1, nyh
      do 122 i = 1, nxh
      f(2*i-1,2*j-1,k+1,l) = real(fc(i,j,k,l))
      f(2*i,2*j,k+1,l)     = aimag(fc(i,j,k,l))
  122 continue 
  120 continue
  130 continue
      return
      end

C =========================================================================
C =========================================================================
 
      subroutine do_forall_2 (f,scr,nxv,nyv,nxvh,nyvh,nzpmx,nblok,
     1 kzp,kblok,nxh,nyh,idps)
  
      complex scr
      dimension f(nxv,nyv,nzpmx,nblok), scr(nxvh,nyvh,idps,kblok)
      include 'pglocal.f'

      l = 1
      do 50 j = 1, nyh
      do 55 i = 1, nxh
      f(2*i-1,2*j-1,1,l) = real(scr(i,j,1,l))
      f(2*i,2*j,1,l) = aimag(scr(i,j,1,l))
   55 continue  
   50 continue
      return
      end

C =========================================================================
C =========================================================================

      subroutine do_forall_3 (f,scr,nxv,nyv,nxvh,nyvh,nzpmx,nblok,
     1 kzp,kblok,nxh,nyh,idps)
  
      complex scr
      dimension f(nxv,nyv,nzpmx,nblok), scr(nxvh,nyvh,idps,kblok)
      include 'pglocal.f'

      l = 1
      do 160 j = 1, nyh
      do 166 i = 1, nxh
      f(2*i-1,2*j-1,kzp+2,l) = real (scr(i,j,1,l))
      f(2*i,2*j,kzp+2,l)     = aimag(scr(i,j,1,l))
      f(2*i-1,2*j-1,kzp+3,l) = real (scr(i,j,2,l))
      f(2*i,2*j,kzp+3,l)     = aimag(scr(i,j,2,l))
  166 continue
  160 continue
      
      return
      end

C =========================================================================
C =========================================================================



C ========================================================================
c
c  This subroutine solves 2d poisson's equation in fourier space for
c  force/charge (or convolution of electric field over particle shape)
c  with periodic boundary conditions.
c
c  The equation used is:
c    fx(kx,ky) = -sqrt(-1)*kx*g(kx,ky)*s(kx,ky),
c    fy(kx,ky) = -sqrt(-1)*ky*g(kx,ky)*s(kx,ky),
c    where kx = 2pi*j/nx, ky = 2pi*k/ny, and j,k = fourier mode numbers,
c      g(kx,ky) = (affp/(kx**2+ky**2))*s(kx,ky),
c      s(kx,ky) = exp(-((kx*ax)**2+(ky*ay)**2)/2), except for
c    for fourier mode (jj-1,k-1), where jj = j + kxp*(l - 1)
c
c    if isign = 0, form factor array is prepared
c    if isign is not equal to 0, force/charge is calculated.
c
C ========================================================================
      subroutine extrinsic_ppois3(q,fx,fy,fz,isign,ffc,ax,ay,az,affp,
     1 wp,nx,ny,nz,kstrt,kxp,kyp,jblok,nzv,nzh,nyvh,sum1,nblok) 

      complex q, fx, fy, fz,ffc, zero
      dimension q(nzv,nyvh,kxp,jblok)
      dimension fx(nzv,nyvh,kxp,jblok),fy(nzv,nyvh,kxp,jblok)
      dimension fz(nzv,nyvh,kxp,jblok)
      dimension ffc(nzh,nyvh,kxp,jblok),sum1(nblok)
      integer myprocnum
      include 'pglocal.f'


      myprocnum = pghpf_myprocnum()
      sum1(1) = 0.0
 
      nxh = nx/2
      nyh = ny/2
C we have nzh as parameter in subroutine
      nz2 = nz  + 2
      ks = kstrt - 2
      dnx = 6.28318530717959/float(nx)
      dny = 6.28318530717959/float(ny)
      dnz = 6.28318530717959/float(nz)
      zero = cmplx(0.,0.)

      if (isign.ne.0) go to 40
c ---------------------------
c prepare form factor array
c ---------------------------
      l = 1

      joff = kxp*(myprocnum+1 + ks) - 1
      do 20 j = 1, kxp
        dkx = dnx*float(j + joff)
        xat1 = dkx*dkx
        xat2 = (dkx*ax)**2

      do 10 k = 1, nyvh
        dky = dny*float(k - 1)
        yat1 = dky*dky
        yat2 = (dky*ay)**2   
  
      do 11 i = 1, nzh
        dkz = dnz*float(i - 1)
        zat1 = dkz*dkz
        zat2 = (dkz*az)**2   

      at3 = xat1+yat1+zat1 
      at4 = exp(-.5*(xat2+yat2+zat2))
      if (at3.eq.0.) then
         ffc(i,k,j,l) = cmplx(0.,1.)
      else
         ffc(i,k,j,l) = cmplx(affp*at4/at3,at4)
      endif
   11 continue  
   10 continue
   20 continue


      return
c ---------------------------------------------
c calculate force/charge and sum field energy
c ---------------------------------------------
   40 wp = 0.
      l = 1
        
c mode numbers 0 < kx < nx/2 and 0 < kz < nz/2 and 0 < ky < ny/2
      joff = kxp*(myprocnum+1 + ks) - 1
      do 60 j = 2, kxp
      dkx = dnx*float(j + joff)
         do 50 k = 2, nyh
         dky = dny*float(k - 1)
         do 55 i = 2, nzh
         i1 = nz2 - i
         dkz = dnz*float(i - 1)
  
         at1 = real(ffc(i,k,j,l))*aimag(ffc(i,k,j,l))
         at2 = dkx*at1
         at3 = dky*at1
         at4 = dkz*at1

         fx(i,k,j,l) = at2*cmplx(aimag(q(i,k,j,l)),-real(q(i,k,j,l)))
         fx(i1,k,j,l) = at2*cmplx(aimag(q(i1,k,j,l)),-real(q(i1,k,j,l)))

         fy(i,k,j,l) = at3*cmplx(aimag(q(i,k,j,l)),-real(q(i,k,j,l)))
         fy(i1,k,j,l) = at3*cmplx(aimag(q(i1,k,j,l)),-real(q(i1,k,j,l)))

         fz(i,k,j,l) = at4*cmplx(aimag(q(i,k,j,l)),-real(q(i,k,j,l)))
         fz(i1,k,j,l) = at4*cmplx(aimag(q(i1,k,j,l)),-real(q(i1,k,j,l)))
               
         wp = wp + at1*(q(i,k,j,l)*conjg(q(i,k,j,l)) + 
     1    q(i1,k,j,l)*conjg(q(i1,k,j,l)))

   55    continue
          
c ---------------------------
c mode numbers kz = 0, nz/2
c ---------------------------
      i1 = nzh + 1

      fx(1,k,j,l)  = zero 
      fx(i1,k,j,l) = zero
      fy(1,k,j,l)  = zero
      fy(i1,k,j,l) = zero
      fz(1,k,j,l)  = zero
      fz(i1,k,j,l) = zero
     
      wp = wp + at1*(q(1,k,j,l)*conjg(q(1,k,j,l)))
   50 continue 

c -------------------------------------------
c mode numbers ky = 0, ny/2 
c -------------------------------------------
      do 90 i = 2, nzh  
         i1 = nz2 - i
       
         fx(i,1,j,l)  = zero
         fx(i1,1,j,l) = zero
         fy(i,1,j,l)  = zero
         fy(i1,1,j,l) = zero
         fz(i,1,j,l)  = zero
         fz(i1,1,j,l) = zero
     
         wp = wp + at1*(q(i,1,j,l)*conjg(q(i,1,j,l)))
   90  continue  

   60 continue
c -------------------------------------------
c mode numbers kx = 0, nx/2 
c -------------------------------------------
         if ((myprocnum+1+ks) .eq. 0) then
         do 70 k = 1, nyh
         do 77 i = 1, nzh
         i1 = nz2 - i
    
         fx(i,k,1,l)  = zero
         fx(i1,k,1,l) = zero
         fy(i,k,1,l)  = zero
         fy(i1,k,1,l) = zero
         fz(i,k,1,l)  = zero
         fz(i1,k,1,l) = zero
     
         wp = wp + at1*(q(i,k,1,l)*conjg(q(i,k,1,l)))
   77    continue  
   70    continue
         endif

      sum1(1) = wp
      return
      end         

C =========================================================================
c
c  This subroutine updates particle co-ordinates and
c  velocities using leap-frog scheme in time and second-order spline
c  interpolation in space, with periodic boundary conditions,
c  The equations used are:
c      vx(t+dt/2) = vx(t-dt/2) + (q/m)*fx(x(t),y(t))*dt,
c      vy(t+dt/2) = vy(t-dt/2) + (q/m)*fy(x(t),y(t))*dt,
c  where q/m is charge/mass, and
c      x(t+dt) = x(t) + vx(t+dt/2)*dt, y(t+dt) = y(t) + vy(t+dt/2)*dt
c      fx(x(t),y(t)) and fy(x(t),y(t)) are approximated by interpolation from
c the nearest grid points.
c 
C =========================================================================
      subroutine extrinsic_ppush3 (part,fx,fy,fz,npp,noff,qtm,dt,ek,
     1 nx,ny,idimp, npmax,nblok,nxv,nyv,nzpmx,sum1)
                                         
      dimension part(idimp,npmax,nblok)
      dimension fx(nxv,nyv,nzpmx,nblok), fy(nxv,nyv,nzpmx,nblok)
      dimension fz(nxv,nyv,nzpmx,nblok)
      dimension npp(nblok), noff(nblok),sum1(nblok)
      include 'pglocal.f'

      zero = 0.
      anx = float(nx)
      any = float(ny)
      sum0 = 0.
      
      l = 1 

      mnoff = noff(l) - 2
c ---------------------------
c find interpolation weights
c ---------------------------
      do 10 j = 1, npp(l)

      nn = part(1,j,l) + .5
      dx = part(1,j,l) - float(nn)
      nn = nn + 1
      if (nn.gt.nx) nn = nn - nx
      amx = .75 - dx*dx
      np = nn + 1
      if (np.gt.nx) np = np - nx
      dxp = .5*(.5 + dx)**2
      nl = nn - 1
      if (nl.lt.1) nl = nl + nx
      dxl = .5*(.5 - dx)**2


      kk = part(2,j,l) + .5
      dy = part(2,j,l) - float(kk)
      kk = kk + 1
      if (kk.gt.ny) kk = kk - ny
      amy = .75 - dy*dy
      kp = kk + 1
      if (kp.gt.ny) kp = kp - ny
      dyp = .5*(.5 + dy)**2
      kl = kk - 1
      if (kl.lt.1) kl = kl + ny
      dyl = .5*(.5 - dy)**2


      mm = part(3,j,l) + .5
      dz = part(3,j,l) - float(mm)
      mm = mm - mnoff
      amz = .75 - dz*dz
      mp = mm + 1
      dzp = .5*(.5 + dz)**2
      ml = mm - 1
      dzl = .5*(.5 - dz)**2


c ------------------
c find acceleration
c ------------------
      dx1 = amx*(amy*(amz*fx(nn,kk,mm,l) + dzp*fx(nn,kk,mp,l) + 
     *     dzl*fx(nn,kk,ml,l)) + dyp*(amz*fx(nn,kp,mm,l) +
     *     dzp*fx(nn,kp,mp,l) + dzl*fx(nn,kp,ml,l)) + 
     *     dyl*(amz*fx(nn,kl,mm,l)+ dzp*fx(nn,kl,mp,l) +
     *     dzl*fx(nn,kl,ml,l)) )

      dx2 = dxp*(amy*(amz*fx(np,kk,mm,l) + dzp*fx(np,kk,mp,l) + 
     *     dzl*fx(np,kk,ml,l)) + dyp*(amz*fx(np,kp,mm,l) +
     *     dzp*fx(np,kp,mp,l) + dzl*fx(np,kp,ml,l)) + 
     *     dyl*(amz*fx(np,kl,mm,l)+ dzp*fx(np,kl,mp,l) +
     *     dzl*fx(np,kl,ml,l)) )

      dx3 = dxl*(amy*(amz*fx(nl,kk,mm,l) + dzp*fx(nl,kk,mp,l) + 
     *     dzl*fx(nl,kk,ml,l)) + dyp*(amz*fx(nl,kp,mm,l) +
     *     dzp*fx(nl,kp,mp,l) + dzl*fx(nl,kp,ml,l)) + 
     *     dyl*(amz*fx(nl,kl,mm,l)+ dzp*fx(nl,kl,mp,l) +
     *     dzl*fx(nl,kl,ml,l)) )

      dx = dx1+dx2+dx3


       dy1 = amx*(amy*(amz*fy(nn,kk,mm,l) + dzp*fy(nn,kk,mp,l) + 
     *     dzl*fy(nn,kk,ml,l)) + dyp*(amz*fy(nn,kp,mm,l) +
     *     dzp*fy(nn,kp,mp,l) + dzl*fy(nn,kp,ml,l)) + 
     *     dyl*(amz*fy(nn,kl,mm,l)+ dzp*fy(nn,kl,mp,l) +
     *     dzl*fy(nn,kl,ml,l)) )

      dy2 = dxp*(amy*(amz*fy(np,kk,mm,l) + dzp*fy(np,kk,mp,l) + 
     *     dzl*fy(np,kk,ml,l)) + dyp*(amz*fy(np,kp,mm,l) +
     *     dzp*fy(np,kp,mp,l) + dzl*fy(np,kp,ml,l)) + 
     *     dyl*(amz*fy(np,kl,mm,l)+ dzp*fy(np,kl,mp,l) +
     *     dzl*fy(np,kl,ml,l)) )

      dy3 = dxl*(amy*(amz*fy(nl,kk,mm,l) + dzp*fy(nl,kk,mp,l) + 
     *     dzl*fy(nl,kk,ml,l)) + dyp*(amz*fy(nl,kp,mm,l) +
     *     dzp*fy(nl,kp,mp,l) + dzl*fy(nl,kp,ml,l)) + 
     *     dyl*(amz*fy(nl,kl,mm,l)+ dzp*fy(nl,kl,mp,l) +
     *     dzl*fy(nl,kl,ml,l)) )

      dy = dy1+dy2+dy3


       dz1 = amx*(amy*(amz*fz(nn,kk,mm,l) + dzp*fz(nn,kk,mp,l) + 
     *     dzl*fz(nn,kk,ml,l)) + dyp*(amz*fz(nn,kp,mm,l) +
     *     dzp*fz(nn,kp,mp,l) + dzl*fz(nn,kp,ml,l)) + 
     *     dyl*(amz*fz(nn,kl,mm,l)+ dzp*fz(nn,kl,mp,l) +
     *     dzl*fz(nn,kl,ml,l)) )

      dz2 = dxp*(amy*(amz*fz(np,kk,mm,l) + dzp*fz(np,kk,mp,l) + 
     *     dzl*fz(np,kk,ml,l)) + dyp*(amz*fz(np,kp,mm,l) +
     *     dzp*fz(np,kp,mp,l) + dzl*fz(np,kp,ml,l)) + 
     *     dyl*(amz*fz(np,kl,mm,l)+ dzp*fz(np,kl,mp,l) +
     *     dzl*fz(np,kl,ml,l)) )

      dz3 = dxl*(amy*(amz*fz(nl,kk,mm,l) + dzp*fz(nl,kk,mp,l) + 
     *     dzl*fz(nl,kk,ml,l)) + dyp*(amz*fz(nl,kp,mm,l) +
     *     dzp*fz(nl,kp,mp,l) + dzl*fz(nl,kp,ml,l)) + 
     *     dyl*(amz*fz(nl,kl,mm,l)+ dzp*fz(nl,kl,mp,l) +
     *     dzl*fz(nl,kl,ml,l)) )

      dz = dz1+dz2+dz3


c --------------
c new velocity
c --------------
      dx = part(4,j,l) + qtm*dx
      dy = part(5,j,l) + qtm*dy
      dz = part(6,j,l) + qtm*dz
c -----------------------
c average kinetic energy
c -----------------------
      sum0 = sum0 + (dx + part(4,j,l))**2 + (dy + part(5,j,l))**2 +
     *              (dz + part(6,j,l))**2

      part(4,j,l) = dx
      part(5,j,l) = dy
      part(6,j,l) = dz
c -------------
c new position
c -------------
      dx = part(1,j,l) + dx*dt
      dy = part(2,j,l) + dy*dt
      dz = part(3,j,l) + dz*dt
c ------------------------------
c periodic boundary conditions
c ------------------------------
      if (dx.lt.zero) dx = dx + anx
      if (dx.ge.anx) dx = dx - anx

      if (dy.lt.zero) dy = dy + any
      if (dy.ge.any) dy = dy - any

      part(1,j,l) = dx
      part(2,j,l) = dy
      part(3,j,l) = dz
   10 continue

c --------------------------
c normalize kinetic energy
c --------------------------
      sum1(1) = sum0
      return
      end

C ========================================================================
C ========================================================================

      
      subroutine extrinsic1_pmove3 (part,edges,npp,sbufr,sbufl,
     1 ihole,jsr,jsl,jss,nz,kstrt,nvp,idimp,npmax,nblok,idps,nbmax,
     1 ntmax,ierr)

      dimension part(idimp,npmax,nblok)
      dimension edges(idps,nblok), npp(nblok)
      dimension sbufl(idimp,nbmax,nblok), sbufr(idimp,nbmax,nblok)
      dimension jsl(idps,nblok), jsr(idps,nblok), jss(idps,nblok)
      dimension ihole(ntmax,nblok)
      integer myprocnum
      include 'pglocal.f'

      anz = float(nz)
      ks = kstrt - 2
      iter = 4

c --------------------------
c buffer outgoing particles
c --------------------------
      myprocnum  = pghpf_myprocnum()
      l = 1
      
      kb = myprocnum+ 1 + ks
      jsl(1,l) = 0
      jsr(1,l) = 0
      jss(2,l) = 0
      do 20 j = 1, npp(l)
      zt = part(3,j,l)
c --------------------
c particles going up
c --------------------
      if (zt.ge.edges(2,l)) then
         if (jsr(1,l).lt.nbmax) then
            jsr(1,l) = jsr(1,l) + 1
            if ((kb+1).eq.nvp) zt = zt - anz
            sbufr(1,jsr(1,l),l) = part(1,j,l)
            sbufr(2,jsr(1,l),l) = part(2,j,l)
            sbufr(3,jsr(1,l),l) = zt
            
            sbufr(4,jsr(1,l),l) = part(4,j,l)
            sbufr(5,jsr(1,l),l) = part(5,j,l)
            sbufr(6,jsr(1,l),l) = part(6,j,l)

            ihole(jsl(1,l)+jsr(1,l),l) = j
         else
            jss(2,l) = 1
         endif
c ---------------------
c particles going down
c ---------------------
      elseif (zt.lt.edges(1,l)) then
         if (jsl(1,l).lt.nbmax) then
            jsl(1,l) = jsl(1,l) + 1
            if (kb.eq.0) zt = zt + anz
            sbufl(1,jsl(1,l),l) = part(1,j,l)
            sbufr(2,jsl(1,l),l) = part(2,j,l)
            sbufl(3,jsl(1,l),l) = zt

            sbufr(4,jsl(1,l),l) = part(4,j,l)
            sbufr(5,jsl(1,l),l) = part(5,j,l)
            sbufr(6,jsl(1,l),l) = part(6,j,l)

            ihole(jsl(1,l)+jsr(1,l),l) = j
         else
            jss(2,l) = 1
         endif
      endif
   20 continue
      jss(1,l) = jsl(1,l) + jsr(1,l)

      return
      end

C ========================================================================
C ========================================================================

       subroutine extrinsic2_pmove3 (edges,sbufr,sbufl,rbufr,rbufl,
     1   jsr,jsl,jss,nz,kstrt,nvp,idimp,npmax,nblok,idps,nbmax,ntmax,
     1   ierr,sum1)

      dimension edges(idps,nblok),sum1(nblok)
      dimension sbufl(idimp,nbmax,nblok), sbufr(idimp,nbmax,nblok)
      dimension rbufl(idimp,nbmax,nblok), rbufr(idimp,nbmax,nblok)
      dimension jsl(idps,nblok), jsr(idps,nblok), jss(idps,nblok)

      integer myprocnum
      include 'pglocal.f'

      anz = float(nz)
      ks = kstrt - 2
      iter = 4 
 
      myprocnum  = pghpf_myprocnum()
      l = 1

c ------------------------------------------
c check if particles must be passed further
c ------------------------------------------
      nps = 0

c --------------------------------------------------------------
c check if any particles coming from above must go further down
c --------------------------------------------------------------
      jsl(1,l) = 0
      do 90 j = 1, jsr(2,l)
      if (rbufr(2,j,l).lt.edges(1,l)) jsl(1,l) = 1
   90 continue
c ------------------------------------------------------------
c check if any particles coming from below must go further up
c ------------------------------------------------------------
      jsr(1,l) = 0
      do 100 j = 1, jsl(2,l)
      if (rbufl(2,j,l).ge.edges(2,l)) jsr(1,l) = 1
  100 continue
      nps = nps + (jsl(1,l) + jsr(1,l))
     
      sum1(1) = nps

      return
      end

C ========================================================================
C ========================================================================   

      subroutine extrinsic3_pmove3 (edges,npp,sbufr,sbufl,rbufr,rbufl,
     1 jsr,jsl,jss,nz,kstrt,nvp,idimp,npmax,nblok,idps,nbmax,ntmax,ierr)

      dimension edges(idps,nblok), npp(nblok)
      dimension sbufl(idimp,nbmax,nblok), sbufr(idimp,nbmax,nblok)
      dimension rbufl(idimp,nbmax,nblok), rbufr(idimp,nbmax,nblok)
      dimension jsl(idps,nblok), jsr(idps,nblok), jss(idps,nblok)

      integer myprocnum
      include 'pglocal.f' 

      anz = float(nz)
      ks = kstrt - 2
      iter = 4
      myprocnum = pghpf_myprocnum()

c -------------------------------------------
c remove particles which do not belong here
c -------------------------------------------
      l = 1

      kb = myprocnum+1 + ks
c -----------------------------------------
c first check particles coming from above
c -----------------------------------------
      jsl(1,l) = 0
      jss(2,l) = 0
      do 120 j = 1, jsr(2,l)
      zt = rbufr(3,j,l)
c ----------------------
c particles going down
c ----------------------
      if (zt.lt.edges(1,l)) then
         jsl(1,l) = jsl(1,l) + 1
         if (kb.eq.0) zt = zt + anz
         sbufl(1,jsl(1,l),l) = rbufr(1,j,l)
         sbufl(2,jsl(1,l),l) = rbufr(2,j,l)
         sbufl(3,jsl(1,l),l) = zt

         sbufl(4,jsl(1,l),l) = rbufr(4,j,l)
         sbufl(5,jsl(1,l),l) = rbufr(5,j,l)
         sbufl(6,jsl(1,l),l) = rbufr(6,j,l)
c ----------------------
c particles staying here
c ----------------------
      else
         jss(2,l) = jss(2,l) + 1
         rbufr(1,jss(2,l),l) = rbufr(1,j,l) 
         rbufr(2,jss(2,l),l) = rbufr(2,j,l) 
         rbufr(3,jss(2,l),l) = zt

         rbufr(4,jss(2,l),l) = rbufr(4,j,l)
         rbufr(5,jss(2,l),l) = rbufr(5,j,l)
         rbufr(6,jss(2,l),l) = rbufr(6,j,l)
      endif
  120 continue
      jsr(2,l) = jss(2,l)
c ---------------------------------------
c next check particles coming from below
c ---------------------------------------
      jsr(1,l) = 0
      jss(2,l) = 0
      do 130 j = 1, jsl(2,l)
      zt = rbufl(3,j,l)
c -------------------
c particles going up
c -------------------
      if (zt.ge.edges(2,l)) then
         jsr(1,l) = jsr(1,l) + 1
         if ((kb+1).eq.nvp) zt = zt - anz
         sbufr(1,jsr(1,l),l) = rbufl(1,j,l)
         sbufr(2,jsr(1,l),l) = rbufl(2,j,l)
         sbufr(3,jsr(1,l),l) = zt

         sbufr(4,jsr(1,l),l) = rbufl(4,j,l)
         sbufr(5,jsr(1,l),l) = rbufl(5,j,l)
         sbufr(6,jsr(1,l),l) = rbufl(6,j,l)
c ------------------------
c particles staying here
c ------------------------
      else
         jss(2,l) = jss(2,l) + 1
         rbufl(1,jss(2,l),l) = rbufl(1,j,l)
         rbufl(2,jss(2,l),l) = rbufl(2,j,l)
         rbufl(3,jss(2,l),l) = zt
       
         rbufl(4,jss(2,l),l) = rbufl(4,j,l)
         rbufl(5,jss(2,l),l) = rbufl(5,j,l)
         rbufl(6,jss(2,l),l) = rbufl(6,j,l)
      endif
  130 continue
      jsl(2,l) = jss(2,l)

      return
      end

C ========================================================================
C ========================================================================

       subroutine extrinsic4_pmove3 (npp,jsr,jsl,jss,nz,kstrt,nvp,
     1  idimp,npmax,nblok,idps,nbmax,ntmax,ierr,sum1)

       dimension npp(nblok),sum1(nblok)
       dimension jsl(idps,nblok), jsr(idps,nblok), jss(idps,nblok)

c ---------------------------------------------
c check if move would overflow particle array
c ---------------------------------------------
      ierr = 0

      l = 1
      if ((npp(l)+jsl(2,l)+jsr(2,l)-jss(1,l)).gt.npmax) then
         jss(2,l) = 1
      else
         jss(2,l) = 0
      endif
      ierr = ierr + jss(2,l)
      sum1(1) = ierr

      return
      end

C ========================================================================
C ========================================================================
   
      subroutine extrinsic5_pmove3 (part,edges,npp,rbufr,rbufl,ihole,
     1 jsr,jsl,jss,nz,kstrt,nvp,idimp,npmax,nblok,idps,nbmax,ntmax,ierr)
      dimension part(idimp,npmax,nblok)
      dimension edges(idps,nblok), npp(nblok)
      dimension rbufl(idimp,nbmax,nblok), rbufr(idimp,nbmax,nblok)
      dimension jsl(idps,nblok), jsr(idps,nblok), jss(idps,nblok)
      dimension ihole(ntmax,nblok)
      include 'pglocal.f'

      anz = float(nz)
      ks = kstrt - 2

      l = 1 

c ---------------------------------------------------
c distribute incoming particles from buffers
c ---------------------------------------------------
      jss(2,l) = min0(jss(1,l),jsl(2,l))
      do 170 j = 1, jss(2,l)
      part(1,ihole(j,l),l) = rbufl(1,j,l)
      part(2,ihole(j,l),l) = rbufl(2,j,l)
      part(3,ihole(j,l),l) = rbufl(3,j,l)
      part(4,ihole(j,l),l) = rbufl(4,j,l)
      part(5,ihole(j,l),l) = rbufl(5,j,l)
      part(6,ihole(j,l),l) = rbufl(6,j,l)

  170 continue
      if (jss(1,l).gt.jsl(2,l)) then
         jss(2,l) = min0(jss(1,l)-jsl(2,l),jsr(2,l))
      else
         jss(2,l) = jsl(2,l) - jss(1,l)
      endif
      do 180 j = 1, jss(2,l)
      if (jss(1,l).gt.jsl(2,l)) then
         part(1,ihole(j+jsl(2,l),l),l) = rbufr(1,j,l)
         part(2,ihole(j+jsl(2,l),l),l) = rbufr(2,j,l)
         part(3,ihole(j+jsl(2,l),l),l) = rbufr(3,j,l)
         part(4,ihole(j+jsl(2,l),l),l) = rbufr(4,j,l)
         part(5,ihole(j+jsl(2,l),l),l) = rbufr(5,j,l)
         part(6,ihole(j+jsl(2,l),l),l) = rbufr(6,j,l)

      else
c ------------------------------------------------------
c no more holes
c distribute remaining particles from below into bottom
c ------------------------------------------------------
         part(1,j+npp(l),l) = rbufl(1,j+jss(1,l),l)
         part(2,j+npp(l),l) = rbufl(2,j+jss(1,l),l)
         part(3,j+npp(l),l) = rbufl(3,j+jss(1,l),l)
         part(4,j+npp(l),l) = rbufl(4,j+jss(1,l),l)
         part(5,j+npp(l),l) = rbufl(5,j+jss(1,l),l)
         part(6,j+npp(l),l) = rbufl(6,j+jss(1,l),l)

      endif
  180 continue
      if (jss(1,l).le.jsl(2,l)) then
         npp(l) = npp(l) + (jsl(2,l) - jss(1,l))
         jss(1,l) = jsl(2,l)
      endif
      jss(2,l) = jss(1,l) - (jsl(2,l) + jsr(2,l))
      if (jss(2,l).gt.0) then
         jss(1,l) = (jsl(2,l) + jsr(2,l))
         jsr(2,l) = jss(2,l)
      else
         jss(1,l) = jss(1,l) - jsl(2,l)
         jsr(2,l) = -jss(2,l)
      endif
      do 190 j = 1, jsr(2,l)
c ---------------------------------------------------------------------
c holes left over
c fill up remaining holes in particle array with particles from bottom
c ---------------------------------------------------------------------
      if (jss(2,l).gt.0) then
         j1 = npp(l) - j + 1
         j2 = jss(1,l) + jss(2,l) - j + 1
         if (j1.gt.ihole(j2,l)) then
c ------------------------------------------------
c move particle only if it is below current hole
c ------------------------------------------------
            part(1,ihole(j2,l),l) = part(1,j1,l)
            part(2,ihole(j2,l),l) = part(2,j1,l)
            part(3,ihole(j2,l),l) = part(3,j1,l)
            part(4,ihole(j2,l),l) = part(4,j1,l)
            part(5,ihole(j2,l),l) = part(5,j1,l)
            part(6,ihole(j2,l),l) = part(6,j1,l)

         endif
      else
c --------------------------------------------------------
c no more holes
c distribute remaining particles from above into bottom
c --------------------------------------------------------
         part(1,j+npp(l),l) = rbufr(1,j+jss(1,l),l)
         part(2,j+npp(l),l) = rbufr(2,j+jss(1,l),l)
         part(3,j+npp(l),l) = rbufr(3,j+jss(1,l),l)
         part(4,j+npp(l),l) = rbufr(4,j+jss(1,l),l)
         part(5,j+npp(l),l) = rbufr(5,j+jss(1,l),l)
         part(6,j+npp(l),l) = rbufr(6,j+jss(1,l),l)

      endif
  190 continue
      if (jss(2,l).gt.0) then
         npp(l) = npp(l) - jsr(2,l)
      else
         npp(l) = npp(l) + jsr(2,l)
      endif
      jss(1,l) = 0

      return
      end

C ========================================================================
C ========================================================================
      

      subroutine bitrv(mixup,sct,indx,indy,indz,nxhyd,nxyhd)

      complex sct
      dimension mixup(nxhyd), sct(nxyhd)
      include 'pglocal.f'

      indx1 = indx - 1
      indy1 = indy - 1
      indx1y = max0(indx1,indy1)
      indx1y = max0(indx1y,indz)
 
      nx = 2**indx
      ny = 2**indy
      nz  = 2**indz

      nxyz = max0(nx,ny)
      nxyz = max0(nxyz,nz)
      nxyhz = 2**indx1y

      do 30 j = 1, nxyhz
      lb = j - 1
      ll = 0
      do 20 k = 1, indx1y
      jb = lb/2
      it = lb - 2*jb
      lb = jb
      ll = 2*ll + it
   20 continue
      mixup(j) = ll + 1
   30 continue
c ---------------------------------------------
c sine/cosine table for the angles 2*n*pi/nxy
c ---------------------------------------------
      nxyzh = nxyz/2
      dnxyz = 6.28318530717959/float(nxyz)
      do 40 j = 1, nxyzh
      arg = dnxyz*float(j - 1)
      sct(j) = cmplx(cos(arg),-sin(arg))
   40 continue
      return 
      end

C =======================================================================
C =======================================================================

C =======================================================================
C =======================================================================
C In this fft routine, main steps for dfft except scrambling
C are the following :
C We divided 3-d domain only trough z-dimension and dublicated x and y-dim.
C 1. take fft along x axis (there is no communication)
C 2. take fft along y axis ( "                       ) 
C 3. Transpose along x-z plane (it needs all-to-all communication)
C 4. Take fft trough x axis( fft of z-axis)
C For ifft , do the above steps but in reverse order. that is it..
C =======================================================================
c
c This subroutine performs a two dimensional real to complex fast
c fourier transform and its inverse, using complex arithmetic,
c
c   indx/indy = exponent which determines length in x/y direction,
c   where nx=2**indx, ny=2**indy
c   ntpose = (0,1) = (no,yes) input, output data are transposed
c
c   if isign = 0, the fft tables are prepared
c   if isign = -1, an inverse fourier transform is performed
c   if ntpose = 0, f is the input and output array, g is a scratch array
c       f(n,m,l) = (1/nx*ny)*sum(f(j,k,i)*
c       exp(-sqrt(-1)*2pi*n*j/nx)*exp(-sqrt(-1)*2pi*mm*kk/ny))
c       where mm = m + kyp*(l - 1) and kk = k + kyp*(i - 1)
c   if ntpose = 1, f is the input and g is the output
c       g(m,n,l) = (1/nx*ny)*sum(f(j,k,i)*
c       exp(-sqrt(-1)*2pi*nn*j/nx)*exp(-sqrt(-1)*2pi*m*kk/ny))
c       where nn = n + kxp*(l - 1) and kk = k + kyp*(i - 1)
c   if isign = 1, a forward fourier transform is performed
c   if ntpose = 0, f is the input and output array, g is a scratch array
c   f(j,k,i) = sum(f(n,m,l)*exp(sqrt(-1)*2pi*n*j/nx)*
c       exp(sqrt(-1)*2pi*mm*kk/ny))
c   where mm = m + kyp*(l - 1) and kk = k + kyp*(i - 1)
c   if ntpose = 1, g is the input and f is the output
c   f(j,k,i) = sum(g(m,n,l)*exp(sqrt(-1)*2pi*nn*j/nx)*
c       exp(sqrt(-1)*2pi*m*kk/ny))
c   where nn = n + kxp*(l - 1) and kk = k + kyp*(i - 1)
c   g(k,1,1) = mode nx/2,k-1, where ny/2+2 <= k <= ny, and
c   imaginary part of g(1,1,1) = real part of mode nx/2,0 and
c   imaginary part of g(ny/2+1,1,1) = real part of mode nx/2,ny/2
c   written by viktor k. decyk, ucla
C =========================================================================

 
      subroutine pfft3r(f,g,bs,br,isign,ntpose,mixup,sct,indx,indy,indz,
     1 kstrt,nxvh,nyvh,nzv,kxp,kyp,kzp,jblok,kblok,nxhyd,nxyhd)

      complex f, g, bs, br, sct, s, t, t1
      dimension f(nxvh,nyvh,kzp,kblok), g(nzv,nyvh,kxp,jblok)
      dimension bs(kxp,kyp,kzp,kblok),br(kxp,kyp,kzp,kblok)
      dimension mixup(nxhyd), sct(nxyhd)
      integer myprocnum
      include 'pglocal.f'

      indx1 = indx - 1
      indy1 = indy - 1
      indx1y = max0(indx1,indy1)
      indx1y = max0(indx1y,indz)
 
      nx = 2**indx
      nxh = nx/2
      nxhh = nx/4
      nxh2 = nxh + 2

      ny = 2**indy
      nyh = ny/2
      nyhh = ny/4
      nyh2 = nyh + 2

      nz  = 2**indz
      nzh = nz/2
      nz2 = nz + 2

      nxyz = max0(nx,ny)
      nxyz = max0(nxyz,nz)
      nxyhz = 2**indx1y
 
      myprocnum = pghpf_myprocnum()
      ks = kstrt - 2
      if (isign) 50, 10, 300
   10 return
c ----------------------------
c inverse fourier transform
c ----------------------------

   50 nrx = nxyhz/nxh

      l = 1
c --------------------------------
c bit-reverse array elements in x
c --------------------------------
      do 70 j = 1, nxh
      j1 = (mixup(j) - 1)/nrx + 1
      if (j.ge.j1) go to 70
      do 66 iy = 1, nyh
      do 60 k = 1, kzp
      t = f(j1,iy,k,l)
      f(j1,iy,k,l) = f(j,iy,k,l)
      f(j,iy,k,l) = t
   60 continue
   66 continue  
   70 continue

c -----------------------
c first transform in x
c -----------------------
      nrx = nxyz/nxh
      do 130 m = 1, indx1
      ns = 2**(m - 1)
      ns2 = ns + ns
      km = nxhh/ns
      kmr = km*nrx

      l = 1
      do 110 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 100 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = sct(1+kmr*(j-1))
      do 99 iy = 1,nyh
      do 90 i = 1, kzp
      t = s*f(j2,iy,i,l)
      f(j2,iy,i,l) = f(j1,iy,i,l) - t
      f(j1,iy,i,l) = f(j1,iy,i,l) + t
   90 continue
   99 continue  
  100 continue
  110 continue

  130 continue
c---------------------------------------
c unscramble coefficients and normalize
c ---------------------------------------
      kmr = nxyz/nx
      ani = 1./float(2*nx*ny*nz)
      l = 1

      do 150 j = 2, nxhh
      t1 = cmplx(aimag(sct(1+kmr*(j-1))),-real(sct(1+kmr*(j-1))))
      do 144 iy = 1,nyh
      do 140 k  = 1, kzp
      t = conjg(f(nxh2-j,iy,k,l))
      s = f(j,iy,k,l) + t
      t = (f(j,iy,k,l) - t)*t1
      f(j,iy,k,l) = ani*(s + t)
      f(nxh2-j,iy,k,l) = ani*conjg(s - t)
  140 continue
  144 continue 
  150 continue
      do 160 k = 1, kzp
      do 166 iy = 1,nyh
      f(nxhh+1,iy,k,l) = 2.*ani*conjg(f(nxhh+1,iy,k,l))
      f(1,iy,k,l) = 2.*ani*cmplx(real(f(1,iy,k,l)) + aimag(f(1,iy,k,l)),
     1 real(f(1,iy,k,l)) - aimag(f(1,iy,k,l)))
  166 continue 
  160 continue


C ------------------------------------------------------------------
C make transform in y direction with same was as we did x direction
C ------------------------------------------------------------------

      nry = nxyhz/nyh
      l = 1

c --------------------------------
c bit-reverse array elements in y
c --------------------------------
      do 810 j = 1, nyh
      j1 = (mixup(j) - 1)/nry + 1
      if (j.ge.j1) go to 810
      do 820 ix = 1, nxh
      do 830 k = 1, kzp
      t = f(ix,j1,k,l)
      f(ix,j1,k,l) = f(ix,j,k,l)
      f(ix,j,k,l) = t
  830 continue
  820 continue  
  810 continue

c -----------------------
c then transform in y
c -----------------------
      nry = nxyz/nyh
      do 840 m = 1, indy1
      ns = 2**(m - 1)
      ns2 = ns + ns
      km = nyhh/ns
      kmr = km*nry

      l = 1
      do 860 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 870 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = sct(1+kmr*(j-1))
      do 880 ix = 1,nxh
      do 890 i = 1, kzp
      t = s*f(ix,j2,i,l)
      f(ix,j2,i,l) = f(ix,j1,i,l) - t
      f(ix,j1,i,l) = f(ix,j1,i,l) + t
  890 continue
  880 continue  
  870 continue
  860 continue

  840 continue
c---------------------------------------
c unscramble coefficients and normalize
c ---------------------------------------
      kmr = nxyz/ny
      ani = 1./float(2*nx*ny*nz)

      l = 1
      do 910 j = 2, nyhh
      t1 = cmplx(aimag(sct(1+kmr*(j-1))),-real(sct(1+kmr*(j-1))))
      do 920 ix = 1,nxh
      do 930 k = 1, kzp
      t = conjg(f(ix,nyh2-j,k,l))
      s = f(ix,j,k,l) + t
      t = (f(ix,j,k,l) - t)*t1
      f(ix,j,k,l) = ani*(s + t)
      f(ix,nyh2-j,k,l) = ani*conjg(s - t)
  930 continue
  920 continue 
  910 continue
      do 940 k = 1, kzp
      do 950 ix = 1,nxh
      f(ix,nyhh+1,k,l) = 2.*ani*conjg(f(ix,nyhh+1,k,l))
      f(ix,1,k,l) = 2.*ani*cmplx(real(f(ix,1,k,l)) + aimag(f(ix,1,k,l)),
     1 real(f(ix,1,k,l)) - aimag(f(ix,1,k,l)))
  950 continue 
  940 continue


c --------------------------------------
c transpose f array to g along x-z plane
c --------------------------------------
  
  180 call ptpose(f,g,bs,br,nxh,nyh,nz,kstrt,nxvh,nyvh,nzv,
     1 kxp,kyp,kzp,nyvh,jblok,kblok,myprocnum)
  
      nrz = nxyhz/nz

      l = 1
c ---------------------------------
c bit-reverse array elements in z
c ---------------------------------
      do 200 k = 1, nz
      k1 = (mixup(k) - 1)/nrz + 1
      if (k.ge.k1) go to 200
      do 190 j = 1, kxp
      do 199 iy = 1,nyh
      t = g(k1,iy,j,l)
      g(k1,iy,j,l) = g(k,iy,j,l)
      g(k,iy,j,l) = t
  199 continue
  190 continue
  200 continue

c ---------------------
c then transform in z
c ---------------------
      nrz = nxyz/nz
      do 260 m = 1, indz
      ns = 2**(m - 1)
      ns2 = ns + ns
      km = nzh/ns
      kmr = km*nrz

      l = 1
      do 240 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 230 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = sct(1+kmr*(j-1))
      do 222 iy = 1,nyh
      do 220 i = 1, kxp
      t = s*g(j2,iy,i,l)
      g(j2,iy,i,l) = g(j1,iy,i,l) - t
      g(j1,iy,i,l) = g(j1,iy,i,l) + t
  220 continue
  222 continue 
  230 continue
  240 continue

  260 continue
c -------------------------------
c unscramble modes kx = 0, nx/2
c -------------------------------
      l = 1

      if ((myprocnum+1+ks).gt.0) go to 280
      do 277 iy = 1,nyh
      do 270 k = 2, nzh
      s = g(nz2-k,iy,1,l)
      g(nz2-k,iy,1,l) = .5*cmplx(aimag(g(k,iy,1,l) + s),
     1 real(g(k,iy,1,l) - s))
      g(k,iy,1,l) = .5*cmplx(real(g(k,iy,1,l) + s),
     1 aimag(g(k,iy,1,l) - s))
  270 continue
  277 continue

      
c -------------------------------
c unscramble modes ky = 0, ny/2
c -------------------------------
      l = 1

      do 2771 ix = 1, kxp
      do 2701 k = 2, nzh
      s = g(nz2-k,1,ix,l)
      g(nz2-k,1,ix,l) = .5*cmplx(aimag(g(k,1,ix,l) + s),
     1 real(g(k,1,ix,l) - s))
      g(k,1,ix,l) = .5*cmplx(real(g(k,1,ix,l) + s),
     1 aimag(g(k,1,ix,l) - s))
 2701 continue
 2771 continue

C these two calls of transpose are not made in this fft .So skip them.
c ------------------------
c transpose g array to f
c ------------------------
  280 return


c ---------------------------
c forward fourier transform
c transpose f array to g
c ---------------------------
      
  300 nrz = nxyhz/nz
      l = 1

c ---------------------------
c scramble modes ky = 0, ny/2
c ---------------------------

      do 3111 ix = 1,kxp
      do 3101 k = 2, nzh
      s = cmplx(aimag(g(nz2-k,1,ix,l)),real(g(nz2-k,1,ix,l)))
      g(nz2-k,1,ix,l) = conjg(g(k,1,ix,l) - s)
      g(k,1,ix,l) = g(k,1,ix,l) + s
 3101 continue
 3111 continue 


c ---------------------------
c scramble modes kx = 0, nx/2
c ---------------------------
      if ((myprocnum+1+ks).gt.0) go to 320
      do 311 iy = 1,nyh
      do 310 k = 2, nzh
      s = cmplx(aimag(g(nz2-k,iy,1,l)),real(g(nz2-k,iy,1,l)))
      g(nz2-k,iy,1,l) = conjg(g(k,iy,1,l) - s)
      g(k,iy,1,l) = g(k,iy,1,l) + s
  310 continue
  311 continue 

c ----------------------------------
c bit-reverse array elements in z
c ----------------------------------
  320 do 340 k = 1, nz
      k1 = (mixup(k) - 1)/nrz + 1
      if (k.ge.k1) go to 340
      do 330 j = 1, kxp
      do 333 iy = 1,nyh
      t = g(k1,iy,j,l)
      g(k1,iy,j,l) = g(k,iy,j,l)
      g(k,iy,j,l) = t
  333 continue
  330 continue
  340 continue

c ----------------------
c first transform in z
c ----------------------
      nrz = nxyz/nz
      do 400 m = 1, indz
      ns = 2**(m - 1)
      ns2 = ns + ns
      km = nzh/ns
      kmr = km*nrz
      l = 1

      do 380 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 370 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = conjg(sct(1+kmr*(j-1)))
      do 360 i = 1, kxp
      do 366 iy = 1,nyh
      t = s*g(j2,iy,i,l)
      g(j2,iy,i,l) = g(j1,iy,i,l) - t
      g(j1,iy,i,l) = g(j1,iy,i,l) + t
  366 continue
  360 continue
  370 continue
  380 continue

  400 continue
c ------------------------
c transpose g array to f
c ------------------------
 
  
  410 call ptpose(g,f,br,bs,nz,nyh,nxh,kstrt,nzv,nyvh,nxvh,
     1 kzp,kyp,kxp,nyvh,jblok,kblok,myprocnum)

      nrx = nxyhz/nxh
      kmr = nxyz/nx
      l = 1

c ----------------------
c scramble coefficients
c ----------------------
      do 430 j = 2, nxhh
      t1 = cmplx(aimag(sct(1+kmr*(j-1))),real(sct(1+kmr*(j-1))))
      do 420 k = 1, kzp
      do 422 iy = 1,nyh
      t = conjg(f(nxh2-j,iy,k,l))
      s = f(j,iy,k,l) + t
      t = (f(j,iy,k,l) - t)*t1
      f(j,iy,k,l) = s + t
      f(nxh2-j,iy,k,l) = conjg(s - t)
  422 continue 
  420 continue
  430 continue
      do 440 k = 1, kzp
      do 444 iy = 1,nyh
      f(nxhh+1,iy,k,l) = 2.*conjg(f(nxhh+1,iy,k,l))
      f(1,iy,k,l) = cmplx(real(f(1,iy,k,l)) + aimag(f(1,iy,k,l)),
     1 real(f(1,iy,k,l)) - aimag(f(1,iy,k,l))) 
  444 continue 
  440 continue
c ---------------------------------
c bit-reverse array elements in x
c ---------------------------------
      do 460 j = 1, nxh
      j1 = (mixup(j) - 1)/nrx + 1
      if (j.ge.j1) go to 460
      do 450 k = 1, kzp
      do 455 iy = 1,nyh
      t = f(j1,iy,k,l)
      f(j1,iy,k,l) = f(j,iy,k,l)
      f(j,iy,k,l) = t
  455 continue 
  450 continue
  460 continue

c ---------------------
c then transform in x
c ---------------------
      nrx = nxyz/nxh
      do 520 m = 1, indx1
      ns = 2**(m - 1)
      ns2 = ns + ns
      km = nxhh/ns
      kmr = km*nrx
      l = 1

      do 500 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 490 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = conjg(sct(1+kmr*(j-1)))
      do 480 i = 1, kzp
      do 488 iy = 1,nyh
      t = s*f(j2,iy,i,l)
      f(j2,iy,i,l) = f(j1,iy,i,l) - t
      f(j1,iy,i,l) = f(j1,iy,i,l) + t
  488 continue 
  480 continue
  490 continue
  500 continue

  520 continue

C take fft along y direction 

      nry = nxyhz/nyh
      kmr = nxyz/ny
      l = 1

c ----------------------
c scramble coefficients
c ----------------------
      do 610 j = 2, nyhh
      t1 = cmplx(aimag(sct(1+kmr*(j-1))),real(sct(1+kmr*(j-1))))
      do 620 k = 1, kzp
      do 630 ix = 1,nxh
      t = conjg(f(ix,nyh2-j,k,l))
      s = f(ix,j,k,l) + t
      t = (f(ix,j,k,l) - t)*t1
      f(ix,j,k,l) = s + t
      f(ix,nyh2-j,k,l) = conjg(s - t)
  630 continue 
  620 continue
  610 continue
      do 640 k = 1, kzp
      do 650 ix = 1,nxh
      f(ix,nyhh+1,k,l) = 2.*conjg(f(ix,nyhh+1,k,l))
      f(ix,1,k,l) = cmplx(real(f(ix,1,k,l)) + aimag(f(ix,1,k,l)),
     1 real(f(ix,1,k,l)) - aimag(f(ix,1,k,l))) 
  650 continue 
  640 continue
c ---------------------------------
c bit-reverse array elements in y
c ---------------------------------
      do 660 j = 1, nyh
      j1 = (mixup(j) - 1)/nry + 1
      if (j.ge.j1) go to 660
      do 670 k = 1, kzp
      do 680 ix = 1,nxh
      t = f(ix,j1,k,l)
      f(ix,j1,k,l) = f(ix,j,k,l)
      f(ix,j,k,l) = t
  680 continue 
  670 continue
  660 continue

c ---------------------
c then transform in y
c ---------------------
      nry = nxyz/nyh
      do 690 m = 1, indy1
      ns = 2**(m - 1)
      ns2 = ns + ns
      km = nyhh/ns
      kmr = km*nry
      l = 1

      do 710 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 720 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = conjg(sct(1+kmr*(j-1)))
      do 730 i = 1, kzp
      do 740 ix = 1,nxh
      t = s*f(ix,j2,i,l)
      f(ix,j2,i,l) = f(ix,j1,i,l) - t
      f(ix,j1,i,l) = f(ix,j1,i,l) + t
  740 continue 
  730 continue
  720 continue
  710 continue

  690 continue


      return
      end


C ========================================================================
c
c  This subroutine performs a transpose of a matrix f, distributed in y,
c  to a matrix g, distributed in x, that is,
c    g(k+kyp*(m-1),j,l) = f(j+kxp*(l-1),k,m), where
c    1 <= j <= kxp, 1 <= k <= kyp, 1 <= l <= nx/kxp, 1 <= m <= ny/kyp
c  and where indices l and m can be distributed across processors.
c    f = complex input array
c    g = complex output array
c    s, t = complex scratch arrays
C ========================================================================
C     this is not extrinsic function which is called by pfft2r

      subroutine ptpose(f,g,s,t,nx,ny,nz,kstrt,nxv,nyv,nzv,
     1 kxp,kyp,kzp,nyvh,jblok,kblok,myprocnum)
      complex f, g, s, t
      dimension f(nxv,nyvh,kzp,kblok), g(nzv,nyvh,kxp,jblok)
      dimension s(kxp,kyp,kzp,kblok), t(kxp,kyp,kzp,jblok)
      integer myprocnum
      include 'pglocal.f'
      integer p


      ks = kstrt - 2
      kxb = nx/kxp
      kzb = nz/kzp

      do l=0,jblok-1
        if (myprocnum .eq. l) then
          do p=0,kblok-1
            joff = kxp*p

            do iy = 1,nyvh
            do k=1,kzp
             do j=1,kxp
               t(j,iy,k,1) = f(j+joff,iy,k,1)
             enddo
            enddo
            enddo

           if (p .ne. myprocnum) then
              call pghpf_csend(p,t(1,1,1,1),kxp*kzp*nyvh,1,PGLCL_CPLX8)
           endif
          enddo

          koff = kzp*myprocnum
          joff = kxp*myprocnum

          do iy = 1,nyvh
          do k=1,kzp
             do j=1,kxp
               g(k+koff,iy,j,1) = f(j+joff,iy,k,1)
             enddo
          enddo
          enddo

        else
          call pghpf_crecv(l,t(1,1,1,1),kxp*kzp*nyvh,1,PGLCL_CPLX8) 
          koff = kzp*l

          do iy = 1,nyvh
          do  k = 1, kyp
           do  j = 1, kxp
             g(koff+k,iy,j,1) = t(j,iy,k,1)
           enddo
          enddo
          enddo  

        endif        
      enddo
      return
      end

C ========================================================================
C ========================================================================
C ========================================================================

      function ranorm(d)
c this program calculates a random number y from a gaussian distribution
c with zero mean and unit variance, according to the method of
c mueller and box:
c    y(k) = (-2*ln(x(k)))**1/2*sin(2*pi*x(k+1))
c    y(k+1) = (-2*ln(x(k)))**1/2*cos(2*pi*x(k+1)),
c where x is a random number uniformly distributed on (0,1).
c written for the ibm by viktor k. decyk, ucla
      integer r1,r2,r4,r5
      double precision ranorm,h1l,h1u,h2l,r0,r3,asc,bsc,temp
      save iflg,r1,r2,r4,r5,h1l,h1u,h2l,r0
      data r1,r2,r4,r5 /885098780,1824280461,1396483093,55318673/
      data h1l,h1u,h2l /65531.0d0,32767.0d0,65525.0d0/
      data iflg,r0 /0,0.0d0/
      if (iflg.eq.0) go to 10
      ranorm = r0
      r0 = 0.0d0
      iflg = 0
      return
   10 isc = 65536
      asc = dfloat(isc)
      bsc = asc*asc
      i1 = r1 - (r1/isc)*isc
      r3 = h1l*dfloat(r1) + asc*h1u*dfloat(i1)
      i1 = r3/bsc
      r3 = r3 - dfloat(i1)*bsc
      bsc = 0.5d0*bsc
      i1 = r2/isc
      isc = r2 - i1*isc
      r0 = h1l*dfloat(r2) + asc*h1u*dfloat(isc)
      asc = 1.0d0/bsc
      isc = r0*asc
      r2 = r0 - dfloat(isc)*bsc
      r3 = r3 + (dfloat(isc) + 2.0d0*h1u*dfloat(i1))
      isc = r3*asc
      r1 = r3 - dfloat(isc)*bsc
      temp = dsqrt(-2.0d0*dlog((dfloat(r1) + dfloat(r2)*asc)*asc))
      isc = 65536
      asc = dfloat(isc)
      bsc = asc*asc
      i1 = r4 - (r4/isc)*isc
      r3 = h2l*dfloat(r4) + asc*h1u*dfloat(i1)
      i1 = r3/bsc
      r3 = r3 - dfloat(i1)*bsc
      bsc = 0.5d0*bsc
      i1 = r5/isc
      isc = r5 - i1*isc
      r0 = h2l*dfloat(r5) + asc*h1u*dfloat(isc)
      asc = 1.0d0/bsc
      isc = r0*asc
      r5 = r0 - dfloat(isc)*bsc
      r3 = r3 + (dfloat(isc) + 2.0d0*h1u*dfloat(i1))
      isc = r3*asc
      r4 = r3 - dfloat(isc)*bsc
      r0 = 6.28318530717959d0*((dfloat(r4) + dfloat(r5)*asc)*asc)
      ranorm = temp*dsin(r0)
      r0 = temp*dcos(r0)
      iflg = 1
      return
      end

      SUBROUTINE TIMERA(T)
C   ==================================================================
C   ===                                                            ===
C   ===    This subroutine returns the elapsed time in seconds     ===
C   ===                    since its first CALL                    ===
C   ===                                                            ===
C   ==================================================================
 
      real T
      real secs, tarray(2)
c      secs = etime(tarray)
      t = secs
      return

      END
