c  
C* -------------------------------------------------------------------
C* The following code is Dr. Decyk's 2-D PIC code for single-processor
C* computers. [KD]
C*
C* -------------------------------------------------------------------

c * * * periodic 2d electrostatic particle simulation kernel code * * *
c This is a simple 2d skeleton particle-in-cell code designed for
c exploring new computer architectures.  It contains the critical pieces
c needed for depositing charge, advancing particles, and solving the
c field.  The code moves only electrons, with periodic electrostatic
c forces obtained by solving poisson's equation with fast fourier
c transforms.  The only diagnostic is particle and field energy.
c portable gcpic kernel code, using algorithm described in:
c p. c. liewer and v. k. decyk, j. computational phys. 85, 302 (1989).
c written by viktor k. decyk, ucla



      program beps2k

c      parameter(indx=6,indy=5,indz=7,npx=192,npy=96,npz=384)
       parameter(indx=3,indy=4,indz=5,npx=24,npy=48,npz=96) 
c       parameter(indx=4,indy=5,indz=6,npx=48,npy=96,npz=192) 
c      parameter(indx=4,indy=5,indz=6,npx=32,npy=64,npz=128) running
cbenc parameter( indx =   6, indy =   7, npx =     384, npy =     768)
cbenc parameter( indx =   7, indy =   8, npx =    1280, npy =    2560)

c      parameter( npxb =  96, npyb =  48, npzb = 192)
       parameter( npxb =  12, npyb =  24, npzb = 48) 
c      parameter( npxb =  24, npyb =  48, npzb = 96) 
c     parameter( npxb =  16, npyb =  16, npzb = 16) running
cbenc parameter( npxb = 128, npyb = 256)
cbenc parameter( npxb = 384, npyb = 768)

      parameter( tend =  15.000, dt = 0.2000000e+00)

      parameter( vtz =   1.000, vdz =   5.000, vtdz =   0.500)

      parameter( avdz =   5.000)

      parameter( indnvp =   2, idps =    2, idimp =   6)

      parameter(npxyz=npx*npy*npz,npxyzb=npxb*npyb*npzb)
      parameter(np=npxyz+npxyzb)
      parameter(nx=2**indx,ny=2**indy,nz=2**indz)
      parameter(nxh=nx/2,nyh=ny/2,nzh=nz/2)
      parameter(nxv=nx+2,nyv=ny+2,nzv=nz+2)
      parameter(nxvh=nxv/2,nyvh=nyv/2)

      parameter(nloop=tend/dt+.0001)

      parameter(nvp=2**indnvp,nblok=1+(nvp-1))

      parameter(npmax=(np/nvp)*1.01+50000,nzpmx=(nz-1)/nvp+4)

      parameter(kzp=(nz-1)/nvp+1,kyp=(nyh-1)/nvp+1,kxp=(nxh-1)/nvp+1)

      parameter(iblok=1+(nxh/kxp-1),jblok=1+(nyh/kyp-1) )

      parameter(kblok=1+(nz/kzp-1))

      parameter(nxyzh=max(nx,ny,nz)/2 )
      parameter(nxhyhz=max(nx/2,ny/2,nz) ) 

      parameter(nbmax_temp=1+2*(npxyz*vtz+npxyzb*vtdz))
      parameter(nbmax = (nbmax_temp + 1.4*npxyzb*avdz)*dt/nz)

      parameter(ntmax=2*nbmax)

      parameter(npd=128,nine=9)
      complex qc,fxc,fyc,fzc,qt,fxt,fyt,fzt,ffc,bs,br,scr,sct,tf
      common /large/ part

C kblok=jblok=nvp dont forget this equality

      dimension part(idimp,npmax,nblok)

      dimension q(nxv,nyv,nzpmx,nblok)
      dimension fx(nxv,nyv,nzpmx,nblok), fy(nxv,nyv,nzpmx,nblok)
      dimension fz(nxv,nyv,nzpmx,nblok )

      dimension qc(nxvh,nyvh,kzp,kblok),tf(nxvh,nyvh,kzp,kblok)
      dimension fxc(nxvh,nyvh,kzp,kblok), fyc(nxvh,nyvh,kzp,kblok)
      dimension fzc(nxvh,nyvh,kzp,kblok)

      dimension qt(nzv,nyvh,kxp,jblok)
      dimension fxt(nzv,nyvh,kxp,jblok),fyt(nzv,nyvh,kxp,jblok)
      dimension fzt(nzv,nyvh,kxp,jblok)

      dimension ffc(nzh,nyvh,kxp,jblok)

c      dimension ffc(nzh,kyp,kxp,jblok)

      dimension mixup(nxhyhz), sct(nxyzh)

      dimension bs(kxp,kyp,kzp,kblok), br(kxp,kyp,kzp,jblok)

      dimension edges(idps,nblok)

      dimension nzp(nblok), noff(nblok)

      dimension npp(nblok), nps(nblok)

      dimension sbufl(idimp,nbmax,nblok), sbufr(idimp,nbmax,nblok)

      dimension rbufl(idimp,nbmax,nblok), rbufr(idimp,nbmax,nblok)

      dimension ihole(ntmax,nblok)

      dimension jsl(idps,nblok), jsr(idps,nblok)

      dimension jss(idps,nblok), scr(nxvh,nyvh,idps,nblok)

      real TIME_INIT, TIME_TOTAL
      real tbegin,timeend

  991 format (5h t = ,i7)
  992 format (19h * * * q.e.d. * * *)
  993 format (34h field, kinetic, total energies = ,3e14.7)

      data vtx,qme,vdx,vtdx /1.,-1.,0.,.5/

      data vty,vdy,vtdy /1., 0., .5/


      data ax,ay,az /.866667,.866667,.866667/

c ntpose = (0,1) = (no,yes) input, output data are transposed in pfft2r
      data ntpose /1/

c -----------
c initialize
c -----------
      kstrt = 1
c      open(unit=6,file='output2',form='formatted',status='unknown')
c -------------------------------------
c initialize timer
c -------------------------------------
      call TIMERA(TIME_INIT)
c -------------------------------------
c initialize constants
c -------------------------------------
      itime = 0
      qtme = qme*dt
      affp = float(nx*ny*nz)/float(np)
      zero = 0.
c -------------------------------------
c calculate partition variables
c -------------------------------------
      print *,'begin..',elapsed_time()
      call dcomp2(edges,nzp,noff,nz,kstrt,nvp,idps,nblok)
      print *,'after dcomp2',elapsed_time()
    
c -------------------------------------
c prepare fft tables
c -------------------------------------
      isign = 0
      call  bitrv(mixup,sct,indx,indy,indz,nxhyhz,nxyzh)

      print *,'after bitrv',elapsed_time()
    
c -------------------------------------
c calculate form factors
c -------------------------------------
      call ppois2(qt,fxt,fyt,fzt,isign,ffc,ax,ay,az,affp,we,nx,ny,
     1 nz,kstrt,kxp,kyp,jblok,nzv,nzh,nyvh,nzp)

      print *,'after ppois2',elapsed_time() 
    

c ------------------------------------------------------
c initialize density profile and velocity distribution
c background electrons
c ------------------------------------------------------

      do 130 l = 1, nblok
      nps(l) = 1
      npp(l) = 0
      nzp3 = nzp(l) + 3
      do 120 k = 1, nzp3
       do 110 j = 1, nx
       do 111 iy = 1,ny
        q(j,iy,k,l) = 0.
  111 continue  
  110 continue
  120 continue
  130 continue
      if (npxyz.gt.0) call pistr2 (part,edges,npp,nps,vtx,vty,vtz,
     1  zero,zero,zero,npx,npy,npz,nx,ny,nz,idimp,npmax,nblok,idps)
      
       print *,'after pistr2',elapsed_time() 
       print *,'npp after initialization..',npp 
c ----------------
c beam electrons
c ----------------
      do 140 l = 1, nblok
      nps(l) = npp(l) + 1
  140 continue
      if (npxyzb.gt.0) call pistr2 (part,edges,npp,nps,vtdx,vtdy,vtdz,
     1 vdx,vdy,vdz,npxb,npyb,npzb,nx,ny,nz,idimp,npmax,nblok,idps)
       print *,'after pistr2',elapsed_time()
        
c ----------------------------------------
c deposit charge for initial distribution
c ----------------------------------------
      call pdost2(part,q,npp,noff,qme,nx,ny,idimp,npmax,nblok,
     1 nxv,nyv,nzpmx)
        print *,'after pdost2',elapsed_time()
   
c       do i=1,nxv
c        do j=1,nyv
c          do k=1,nzpmx
c             do l=1,nblok
c               print *,i,j,k,l,q(i,j,k,l)
c            enddo
c            enddo
c            enddo
c            enddo
c       stop           
c -----------------------------
c add background ion density
c -----------------------------
      qi0 = -qme/affp
      do 240 l = 1, nblok
      do 230 k = 1, nzp(l)
      do 220 j = 1, nx
      do 222 iy = 1,ny
      q(j,iy,k+1,l) = q(j,iy,k+1,l) + qi0
  222 continue
  220 continue
  230 continue
  240 continue
    
c =======================================
c
c * * * start main iteration loop * * *
c
c =======================================
        call timer(timebegin)
  500 if (nloop.le.itime) go to 2000
      write (6,991) itime
      isign = -1
c -------------------------------------------
c copy data from particle to field partition
c -------------------------------------------
      call cppfp2 (q,qc,isign,scr,kstrt,nvp,nx,ny,nxv,
     1 nyv,nzpmx,nblok,nxvh,nyvh,kzp,kblok,idps)
   
      print *,'after cppfp2',elapsed_time()

c ----------------------------------
c transform charge to fourier space
c ----------------------------------

      call pfft2r(qc,qt,bs,br,isign,ntpose,mixup,sct,indx,indy,
     1 indz,kstrt,nxvh,nyvh,nzv,kxp,kyp,kzp,jblok,kblok,nxhyhz,nxyzh)

      print *,'after pfft2r',elapsed_time()
c -----------------------------------------
c calculate force/charge in fourier space
c -----------------------------------------
     
        call ppois2(qt,fxt,fyt,fzt,isign,ffc,ax,ay,az,affp,we,nx,ny,
     1 nz,kstrt,kxp,kyp,jblok,nzv,nzh,nyvh,nzp)
        
         print *,'after ppois2',elapsed_time()

c -------------------------------------
c transform force/charge to real space
c -------------------------------------
      isign = 1
      call pfft2r(fxc,fxt,bs,br,isign,ntpose,mixup,sct,indx,indy,indz,
     1 kstrt,nxvh,nyvh,nzv,kxp,kyp,kzp,jblok,kblok,nxhyhz,nxyzh)
    
      print *,'after pfft2r',elapsed_time()

      call pfft2r(fyc,fyt,bs,br,isign,ntpose,mixup,sct,indx,indy,indz, 
     1 kstrt,nxvh,nyvh,nzv,kxp,kyp,kzp,jblok,kblok,nxhyhz,nxyzh)

      
      print *,'after pfft2r',elapsed_time()

      call pfft2r(fzc,fzt,bs,br,isign,ntpose,mixup,sct,indx,indy,indz,
     1 kstrt,nxvh,nyvh,nzv,kxp,kyp,kzp,jblok,kblok,nxhyhz,nxyzh)

c -------------------------------------------
c copy data from field to particle partition
c -------------------------------------------
      call cppfp2 (fx,fxc,isign,scr,kstrt,nvp,nx,ny,nxv,
     1 nyv,nzpmx,nblok,nxvh,nyvh,kzp,kblok,idps)
   
      call cppfp2 (fy,fyc,isign,scr,kstrt,nvp,nx,ny,nxv,
     1 nyv,nzpmx,nblok,nxvh,nyvh,kzp,kblok,idps)
    
      call cppfp2 (fz,fzc,isign,scr,kstrt,nvp,nx,ny,nxv,
     1 nyv,nzpmx,nblok,nxvh,nyvh,kzp,kblok,idps)
    
      print *,'after cppfp2',elapsed_time()

c -----------------------------------------
c particle push and charge density update
c -----------------------------------------
c     call timera(TIME_PUSH)
c -----------------------------------
c initialize charge density to zero
c -----------------------------------
      do 1250 l = 1, nblok
      nzp3 = nzp(l) + 3
      do 1240 k = 1, nzp3
      do 1230 j = 1, nx
      do 1233 iy = 1,ny
      q(j,iy,k,l) = 0.
 1233 continue   
 1230 continue
 1240 continue
 1250 continue
      wke = 0.
c ----------------
c push particles
c ----------------
      call ppush2 (part,fx,fy,fz,npp,noff,qtme,dt,wke,nx,ny,idimp,
     1 npmax,nblok,nxv,nyv,nzpmx)

      print *,'after ppush2',elapsed_time()
c -------------------------------------------------
c move particles into appropriate spatial regions
c -------------------------------------------------
      call pmove2 (part,edges,npp,sbufr,sbufl,rbufr,rbufl,ihole,jsr,jsl,
     1jss,nz,kstrt,nvp,idimp,npmax,nblok,idps,nbmax,ntmax,ierr)
       print *,'after pmove2',elapsed_time() 

c ----------------
c deposit charge
c ----------------
      call pdost2(part,q,npp,noff,qme,nx,ny,idimp,npmax,nblok,
     1   nxv,nyv,nzpmx)
        print *,'after pdost2',elapsed_time()
c ----------------------
c add background charge
c ----------------------
      do 1280 l = 1, nblok
      do 1270 k = 1, nzp(l)
      do 1260 j = 1, nx
      do 1266 iy = 1,ny
      q(j,iy,k+1,l) = q(j,iy,k+1,l) + qi0
 1266 continue
 1260 continue
 1270 continue
 1280 continue

      print *,'we-wke',we,wke
 
c --------------------
c energy diagnostic
c --------------------
      wt = we + wke
      write (6,993) we, wke, wt
      itime = itime + 1
      go to 500
 2000 continue
c =====================================
c
c * * * end main iteration loop * * *
c
c =====================================
      write (6,992)
      call timer(timeend)

      print *,'total time = ',timeend-timebegin, 'seconds'
      end

C--------------------------------------------------------------------------
C--------------------------------------------------------------------------
  
      real function elapsed_time()
      integer :: clock_count, clock_rate, clock_max
      integer, save :: last_count
      logical, save :: system_clock_not_yet_called = .true.
 
      call system_clock(count=clock_count,
     1                    count_rate=clock_rate,
     1                  count_max=clock_max)
      if (system_clock_not_yet_called) then
         system_clock_not_yet_called = .false.
         elapsed_time = 0.0
      else if (clock_count > last_count) then
         elapsed_time = real(clock_count-last_count)/real(clock_rate)
      else
         elapsed_time = real(clock_max-last_count+clock_count)
     1       /real(clock_rate)
      end if
      last_count = clock_count
      return
      end
C--------------------------------------------------------------------------
C--------------------------------------------------------------------------
  
      subroutine timer(t)
 
      integer val(8)
      double precision t
 
      call date_and_time(values=val)
      t = dble(val(8))*1d-3 + dble(val(7)) +
     1   dble(val(6))*60d0 + dble(val(5))*3600d0
 
      end subroutine timer


C =======================================================================
c
c  This subroutine determines spatial boundaries for particle
c  decomposition, calculates number of grid points in each spatial
c  region, and the offset of these grid points from the global address
c
C =======================================================================
      subroutine dcomp2(edges,nzp,noff,nz,kstrt,nvp,idps,nblok)
      dimension edges(idps,nblok)
      dimension nzp(nblok), noff(nblok)
      ks = kstrt - 2
      at1 = float(nz)/float(nvp)
      do 10 l = 1, nblok
      kb = l + ks
      edges(1,l) = at1*float(kb)
      noff(l) = edges(1,l) + .5
      edges(2,l) = at1*float(kb + 1)
      kr = edges(2,l) + .5
      nzp(l) = kr - noff(l)
   10 continue
      return
      end
C =======================================================================
c
c  This subroutine calculates initial particle co-ordinates
c  and velocities with uniform density and maxwellian velocity with drift.
c
C =======================================================================
      subroutine pistr2 (part,edges,npp,nps,vtx,vty,vtz,vdx,vdy,vdz,
     1 npx,npy,npz,nx,ny,nz,idimp,npmax,nblok,idps)
 
      double precision ranorm
      dimension part(idimp,npmax,nblok)
      dimension edges(idps,nblok), npp(nblok), nps(nblok)
      dimension sum3(3)
   

      npxyz = npx*npy*npz
      at1 = float(nx)/float(npx)
      at2 = float(ny)/float(npy)
      at3 = float(nz)/float(npz)
     
      do 33 i = 1, npz
      zt = at3*(float(i) - .5)
      do 30 j = 1, npy
      yt = at2*(float(j) - .5)
      do 20 k = 1, npx
c ---------------------------------
c uniform density profile
c ---------------------------------
      xt = at1*(float(k) - .5)
c ---------------------------------
c maxwellian velocity distribution
c ---------------------------------
      vxt = vtx*ranorm(d)
      vyt = vty*ranorm(d)
      vzt = vtz*ranorm(d)
      do 10 l = 1, nblok
      if ((zt.ge.edges(1,l)).and.(zt.lt.edges(2,l))) then
         npt = npp(l) + 1
         part(1,npt,l) = xt
         part(2,npt,l) = yt
         part(3,npt,l) = zt
         part(4,npt,l) = vxt
         part(5,npt,l) = vyt
         part(6,npt,l) = vzt
         npp(l) = npt
      endif
   10 continue
   20 continue
   30 continue
   33 continue  
c -------------------
c add correct drift
c -------------------
      sum3(1) = 0.
      sum3(2) = 0.
      sum3(3) = 0.
      do 50 l = 1, nblok
      sum0 = 0.
      sum1 = 0.
      sum2 = 0.
      do 40 j = nps(l), npp(l)
      sum0 = sum0 + part(4,j,l)
      sum1 = sum1 + part(5,j,l)
      sum2 = sum2 + part(6,j,l)
   40 continue
      sum3(1) = sum3(1) + sum0
      sum3(2) = sum3(2) + sum1
      sum3(3) = sum3(3) + sum2
   50 continue

      at1 = 1./float(npxyz)
      sum3(1) = at1*sum3(1) - vdx
      sum3(2) = at1*sum3(2) - vdy
      sum3(3) = at1*sum3(3) - vdz
      do 70 l = 1, nblok
      do 60 j = nps(l), npp(l)
      part(4,j,l) = part(4,j,l) - sum3(1)
      part(5,j,l) = part(5,j,l) - sum3(2)
      part(6,j,l) = part(6,j,l) - sum3(3)
   60 continue
   70 continue
      return
      end
C =========================================================================
c
c  This subroutine calculates particle charge density
c  using second-order spline interpolation, periodic boundaries.
c
c charge density is approximated by values at the nearest grid points
c
C =========================================================================
  
      subroutine pdost2(part,q,npp,noff,qm,nx,ny,idimp,npmax,nblok,nxv,
     1   nyv,nzpmx)

      dimension part(idimp,npmax,nblok), q(nxv,nyv,nzpmx,nblok)
      dimension npp(nblok), noff(nblok)

      do 20 l = 1, nblok
      mnoff = noff(l) - 2
c -------------------------------
c find interpolation weights
c -------------------------------
      print *,'npp',l,npp(l)
      do 10 j = 1, npp(l)
    

      nn = part(1,j,l) + .5
      dxl = part(1,j,l) - float(nn)
      nn = nn + 1
      if (nn.gt.nx) nn = nn - nx
      amx = qm*(.75 - dxl*dxl)
      np = nn + 1
      if (np.gt.nx) np = np - nx
      dxp = .5*qm*(.5 + dxl)**2
      nl = nn - 1
      if (nl.lt.1) nl = nl + nx
      dxl = .5*qm*(.5 - dxl)**2
      
      kk = part(2,j,l) + .5
      dyl = part(2,j,l) - float(kk)
      kk = kk + 1
      if (kk.gt.ny) kk = kk - ny
      amy = qm*(.75 - dyl*dyl)
      kp = kk + 1
      if (kp.gt.ny) kp = kp - ny
      dyp = .5*qm*(.5 + dyl)**2
      kl = kk - 1
      if (kl.lt.1) kl = kl + ny
      dyl = .5*qm*(.5 - dyl)**2


      mm = part(3,j,l) + .5
      dzl = part(3,j,l) - float(mm)
      mm = mm - mnoff
      amz = .75 - dzl*dzl
      mp = mm + 1
      dzp = .5*(.5 + dzl)**2
      ml = mm - 1
      dzl = .5*(.5 - dzl)**2
c ----------------
c deposit charge
c ----------------
      q(nn,kk,mm,l) = q(nn,kk,mm,l) + amx*amy*amz
      q(np,kk,mm,l) = q(np,kk,mm,l) + dxp*amy*amz
      q(nl,kk,mm,l) = q(nl,kk,mm,l) + dxl*amy*amz
      q(nn,kp,mm,l) = q(nn,kp,mm,l) + amx*dyp*amz
      q(np,kp,mm,l) = q(np,kp,mm,l) + dxp*dyp*amz
      q(nl,kp,mm,l) = q(nl,kp,mm,l) + dxl*dyp*amz
      q(nn,kl,mm,l) = q(nn,kl,mm,l) + amx*dyl*amz
      q(np,kl,mm,l) = q(np,kl,mm,l) + dxp*dyl*amz
      q(nl,kl,mm,l) = q(nl,kl,mm,l) + dxl*dyl*amz

      q(nn,kk,mp,l) = q(nn,kk,mp,l) + amx*amy*dzp
      q(np,kk,mp,l) = q(np,kk,mp,l) + dxp*amy*dzp
      q(nl,kk,mp,l) = q(nl,kk,mp,l) + dxl*amy*dzp
      q(nn,kp,mp,l) = q(nn,kp,mp,l) + amx*dyp*dzp
      q(np,kp,mp,l) = q(np,kp,mp,l) + dxp*dyp*dzp
      q(nl,kp,mp,l) = q(nl,kp,mp,l) + dxl*dyp*dzp
      q(nn,kl,mp,l) = q(nn,kl,mp,l) + amx*dyl*dzp
      q(np,kl,mp,l) = q(np,kl,mp,l) + dxp*dyl*dzp
      q(nl,kl,mp,l) = q(nl,kl,mp,l) + dxl*dyl*dzp

      q(nn,kk,ml,l) = q(nn,kk,ml,l) + amx*amy*dzl
      q(np,kk,ml,l) = q(np,kk,ml,l) + dxp*amy*dzl
      q(nl,kk,ml,l) = q(nl,kk,ml,l) + dxl*amy*dzl
      q(nn,kp,ml,l) = q(nn,kp,ml,l) + amx*dyp*dzl
      q(np,kp,ml,l) = q(np,kp,ml,l) + dxp*dyp*dzl
      q(nl,kp,ml,l) = q(nl,kp,ml,l) + dxl*dyp*dzl
      q(nn,kl,ml,l) = q(nn,kl,ml,l) + amx*dyl*dzl
      q(np,kl,ml,l) = q(np,kl,ml,l) + dxp*dyl*dzl
      q(nl,kl,ml,l) = q(nl,kl,ml,l) + dxl*dyl*dzl
  

   10 continue
   20 continue
      return
      end

C =========================================================================
c
c  This subroutine copies data between particle and field partitions,
c  where the field and particle partitions are assumed to be the same.
c
c    f(j,k,l) = real data for grid j,k in particle partition l.  
c    fc(j,k,l) = complex data for grid j,k in field partition l.  

c    if isign = -1, data is copied from particle to field partition.
c    if isign = 1, data is copied from field to particle partition.
C =========================================================================
      subroutine cppfp2 (f,fc,isign,scr,kstrt,nvp,nx,ny,nxv,
     1 nyv,nzpmx,nblok,nxvh,nyvh,kzp,kblok,idps)
  
      complex fc, scr
      dimension f(nxv,nyv,nzpmx,nblok), fc(nxvh,nyvh,kzp,kblok)
      dimension scr(nxvh,nyvh,idps,nblok)
      nxh = nx/2
      nyh = ny/2
      ks = kstrt - 2
      moff = nzpmx*nvp
      if (isign) 10, 100, 110
c --------------------------------------
c copy from particle to field partition
c --------------------------------------
   10 do 40 l = 1, kblok
      do 30 k = 1, kzp
      do 33 j = l, nyh
      do 20 i = 1, nxh
      fc(i,j,k,l) = cmplx(f(2*i-1,2*j-1,k+1,l),f(2*i,2*j,k+1,l))
   20 continue
   33 continue  
   30 continue
   40 continue


      do 90 l = 1, kblok
      kr = l + ks + 2
      if (kr.gt.nvp) then
         kr = kr - nvp
      endif
      krr = kr
      kl = l + ks
      if (kl.lt.1) then
         kl = kl + nvp
      endif
      kll = kl
      ngc = 2
c ---------------------------------------------
c special case of only one grid per processor
c ---------------------------------------------
      if (kzp.eq.1) then
         krr = krr + 1
         if (krr.gt.nvp) then
            krr = krr - nvp
         endif
         kll = kll - 1
         if (kll.lt.1) then
            kll = kll + nvp
         endif
         ngc = 1
      endif

    
      do 50 j = 1, nyh
      do 55 i = 1, nxh
      scr(i,j,1,l) = cmplx(f(2*i-1,2*j-1,kzp+2,kl),
     1                     f(2*i,2*j,kzp+2,kl))
      scr(i,j,2,l) = cmplx(f(2*i-1,2*j-1,kzp+3,kll),
     1                     f(2*i,2*j,kzp+3,kll))      
   55 continue  
   50 continue

    
      do 60 j = 1, nyh
      do 66 i = 1, nxh
      fc(i,j,1,l) = fc(i,j,1,l) + scr(i,j,1,l)
      fc(i,j,ngc,l) = fc(i,j,ngc,l) + scr(i,j,2,l)
   66 continue  
   60 continue

   
      do 70 j = 1, nyh
      do 77 i = 1, nxh
      scr(i,j,1,l) = cmplx(f(2*i-1,2*j-1,1,kr),f(2*i,2*j,1,kr))
   77 continue  
   70 continue

   
      do 80 j = 1, nyh
      do 88 i = 1, nxh
      fc(i,j,kzp,l) = fc(i,j,kzp,l) + scr(i,j,1,l)
   88 continue  
   80 continue
 
   90 continue
  100 return
c ---------------------------------------
c copy from field to particle partition
c ---------------------------------------
  110 do 140 l = 1, kblok
      do 130 k = 1, kzp
      do 120 j = 1, nyh
      do 122 i = 1, nxh
      f(2*i-1,2*j-1,k+1,l) = real(fc(i,j,k,l))
      f(2*i,2*j,k+1,l)     = aimag(fc(i,j,k,l))
  122 continue 
  120 continue
  130 continue
  140 continue

      do 170 l = 1, kblok
      kr = l + ks + 2
      if (kr.gt.nvp) then
         kr = kr - nvp
      endif
      krr = kr
      kl = l + ks
      if (kl.lt.1) then
         kl = kl + nvp
      endif
      kll = kl
      ngc = 2
c --------------------------------------------
c special case of only one grid per processor
c --------------------------------------------
      if (kzp.eq.1) then
         krr = krr + 1
         if (krr.gt.nvp) then
            krr = krr - nvp
         endif
         kll = kll - 1
         if (kll.lt.1) then
            kll = kll + nvp
         endif
         ngc = 1
      endif

      do 150 j = 1, nyh
      do 155 i = 1, nxh
      f(2*i-1,2*j-1,1,l) = real(fc(i,j,kzp,kl))
      f(2*i,2*j,1,l) = aimag(fc(i,j,kzp,kl))
  155 continue
  150 continue

      do 160 j = 1, nyh
      do 166 i = 1, nxh
      f(2*i-1,2*j-1,kzp+2,l) = real(fc(i,j,1,kr))
      f(2*i,2*j,kzp+2,l) = aimag(fc(i,j,1,kr))
      f(2*i-1,2*j-1,kzp+3,l) = real(fc(i,j,ngc,krr))
      f(2*i,2*j,kzp+3,l) = aimag(fc(i,j,ngc,krr))
  166 continue
  160 continue

  170 continue
      return
      end

C ========================================================================
c
c  This subroutine solves 2d poisson's equation in fourier space for
c  force/charge (or convolution of electric field over particle shape)
c  with periodic boundary conditions.
c
c  The equation used is:
c    fx(kx,ky) = -sqrt(-1)*kx*g(kx,ky)*s(kx,ky),
c    fy(kx,ky) = -sqrt(-1)*ky*g(kx,ky)*s(kx,ky),
c    where kx = 2pi*j/nx, ky = 2pi*k/ny, and j,k = fourier mode numbers,
c      g(kx,ky) = (affp/(kx**2+ky**2))*s(kx,ky),
c      s(kx,ky) = exp(-((kx*ax)**2+(ky*ay)**2)/2), except for
c    for fourier mode (jj-1,k-1), where jj = j + kxp*(l - 1)
c
c    if isign = 0, form factor array is prepared
c    if isign is not equal to 0, force/charge is calculated.
c
C ========================================================================
      subroutine ppois2(q,fx,fy,fz,isign,ffc,ax,ay,az,affp,wp,nx,ny,nz,
     1 kstrt,kxp,kyp,jblok,nzv,nzh,nyvh,nzp)       

      complex q, fx, fy, fz,ffc, zero
      dimension q(nzv,nyvh,kxp,jblok)
      dimension fx(nzv,nyvh,kxp,jblok),fy(nzv,nyvh,kxp,jblok)
      dimension fz(nzv,nyvh,kxp,jblok)
      dimension ffc(nzh,nyvh,kxp,jblok)
      dimension nzp(jblok)

       
      nxh = nx/2
      nyh = ny/2
C we have nzh as parameter in subroutine
      nz2 = nz  + 2
      ks = kstrt - 2
      dnx = 6.28318530717959/float(nx)
      dny = 6.28318530717959/float(ny)
      dnz = 6.28318530717959/float(nz)
      zero = cmplx(0.,0.)


      if (isign.ne.0) go to 40
c ---------------------------
c prepare form factor array
c ---------------------------
      do 30 l = 1, jblok
      joff = kxp*(l + ks) - 1
      do 20 j = 1, kxp
        dkx = dnx*float(j + joff)
        xat1 = dkx*dkx
        xat2 = (dkx*ax)**2

      do 10 k = 1, nyvh
        dky = dny*float(k - 1)
        yat1 = dky*dky
        yat2 = (dky*ay)**2   
  
      do 11 i = 1, nzh
        dkz = dnz*float(i - 1)
        zat1 = dkz*dkz
        zat2 = (dkz*az)**2   

      at3 = xat1+yat1+zat1 
      at4 = exp(-.5*(xat2+yat2+zat2))
    
      if (at3.eq.0.) then
         ffc(i,k,j,l) = cmplx(0.,1.)
      else
         ffc(i,k,j,l) = cmplx(affp*at4/at3,at4)
      endif
   11 continue  
   10 continue
   20 continue
   30 continue
    
      return
c ---------------------------------------------
c calculate force/charge and sum field energy
c ---------------------------------------------
   40 wp = 0.
      do 80 l = 1, jblok
c mode numbers 0 < kx < nx/2 and 0 < kz < nz/2 and 0 < ky < ny/2
      joff = kxp*(l + ks) - 1
      do 60 j = 2, kxp
      dkx = dnx*float(j + joff)
         do 50 k = 2, nyh
         dky = dny*float(k - 1)
         do 55 i = 2, nzh
         i1 = nz2 - i
         dkz = dnz*float(i - 1)
  
         at1 = real(ffc(i,k,j,l))*aimag(ffc(i,k,j,l))
         at2 = dkx*at1
         at3 = dky*at1
         at4 = dkz*at1

         fx(i,k,j,l) = at2*cmplx(aimag(q(i,k,j,l)),-real(q(i,k,j,l)))
         fx(i1,k,j,l) = at2*cmplx(aimag(q(i1,k,j,l)),-real(q(i1,k,j,l)))

         fy(i,k,j,l) = at3*cmplx(aimag(q(i,k,j,l)),-real(q(i,k,j,l)))
         fy(i1,k,j,l) = at3*cmplx(aimag(q(i1,k,j,l)),-real(q(i1,k,j,l)))

         fz(i,k,j,l) = at4*cmplx(aimag(q(i,k,j,l)),-real(q(i,k,j,l)))
         fz(i1,k,j,l) = at4*cmplx(aimag(q(i1,k,j,l)),-real(q(i1,k,j,l)))
               
         wp = wp + at1*(q(i,k,j,l)*conjg(q(i,k,j,l)) + 
     1    q(i1,k,j,l)*conjg(q(i1,k,j,l)))

   55    continue
          
c ---------------------------
c mode numbers kz = 0, nz/2
c ---------------------------
      i1 = nzh + 1

      fx(1,k,j,l) = zero 
      fx(i1,k,j,l) = zero
      fy(1,k,j,l) = zero
      fy(i1,k,j,l) = zero
      fz(1,k,j,l) = zero
      fz(i1,k,j,l) = zero
     
      wp = wp + at1*(q(1,k,j,l)*conjg(q(1,k,j,l)))
   50 continue 

c -------------------------------------------
c mode numbers ky = 0, ny/2 
c -------------------------------------------
      do 90 i = 2, nzh  
         i1 = nz2 - i
       
         fx(i,1,j,l) = zero
         fx(i1,1,j,l) = zero
         fy(i,1,j,l) = zero
         fy(i1,1,j,l) = zero
         fz(i,1,j,l) = zero
         fz(i1,1,j,l) = zero
     
         wp = wp + at1*(q(i,1,j,l)*conjg(q(i,1,j,l)))
   90  continue  

   60 continue
c -------------------------------------------
c mode numbers kx = 0, nx/2 
c -------------------------------------------
         if ((l+ks) .eq. 0) then
         do 70 k = 1,nyh
         do 77 i = 1, nzh
         i1 = nz2 - i
    
         fx(i,k,1,l) = zero
         fx(i1,k,1,l) = zero
         fy(i,k,1,l) = zero
         fy(i1,k,1,l) = zero
         fz(i,k,1,l) = zero
         fz(i1,k,1,l) = zero
     
         wp = wp + at1*(q(i,k,1,l)*conjg(q(i,k,1,l)))
   77    continue  
   70    continue
         endif
      
   80 continue

      wp = float(nx*ny*nz)*wp
      return
      end         

C =========================================================================
c
c  This subroutine updates particle co-ordinates and
c  velocities using leap-frog scheme in time and second-order spline
c  interpolation in space, with periodic boundary conditions,
c  The equations used are:
c      vx(t+dt/2) = vx(t-dt/2) + (q/m)*fx(x(t),y(t))*dt,
c      vy(t+dt/2) = vy(t-dt/2) + (q/m)*fy(x(t),y(t))*dt,
c  where q/m is charge/mass, and
c      x(t+dt) = x(t) + vx(t+dt/2)*dt, y(t+dt) = y(t) + vy(t+dt/2)*dt
c      fx(x(t),y(t)) and fy(x(t),y(t)) are approximated by interpolation from
c the nearest grid points.
c 
C =========================================================================
      subroutine ppush2 (part,fx,fy,fz,npp,noff,qtm,dt,ek,nx,ny,
     1 idimp, npmax,nblok,nxv,nyv,nzpmx)
                                   
      dimension part(idimp,npmax,nblok)
      dimension fx(nxv,nyv,nzpmx,nblok), fy(nxv,nyv,nzpmx,nblok)
      dimension fz(nxv,nyv,nzpmx,nblok)
      dimension npp(nblok), noff(nblok)

      zero = 0.
      anx = float(nx)
      any = float(ny)
      anz = float(nz)
      sum1 = 0.
 
      do 20 l = 1, nblok
      mnoff = noff(l) - 2
c ---------------------------
c find interpolation weights
c ---------------------------
      do 10 j = 1, npp(l)

      nn = part(1,j,l) + .5
      dx = part(1,j,l) - float(nn)
      nn = nn + 1
      if (nn.gt.nx) nn = nn - nx
      amx = .75 - dx*dx
      np = nn + 1
      if (np.gt.nx) np = np - nx
      dxp = .5*(.5 + dx)**2
      nl = nn - 1
      if (nl.lt.1) nl = nl + nx
      dxl = .5*(.5 - dx)**2


      kk = part(2,j,l) + .5
      dy = part(2,j,l) - float(kk)
      kk = kk + 1
      if (kk.gt.ny) kk = kk - ny
      amy = .75 - dy*dy
      kp = kk + 1
      if (kp.gt.ny) kp = kp - ny
      dyp = .5*(.5 + dy)**2
      kl = kk - 1
      if (kl.lt.1) kl = kl + ny
      dyl = .5*(.5 - dy)**2


      mm = part(3,j,l) + .5
      dz = part(3,j,l) - float(mm)
      mm = mm - mnoff
      amz = .75 - dz*dz
      mp = mm + 1
      dzp = .5*(.5 + dz)**2
      ml = mm - 1
      dzl = .5*(.5 - dz)**2


c ------------------
c find acceleration
c ------------------
      dx1 = amx*(amy*(amz*fx(nn,kk,mm,l) + dzp*fx(nn,kk,mp,l) + 
     *     dzl*fx(nn,kk,ml,l)) + dyp*(amz*fx(nn,kp,mm,l) +
     *     dzp*fx(nn,kp,mp,l) + dzl*fx(nn,kp,ml,l)) + 
     *     dyl*(amz*fx(nn,kl,mm,l)+ dzp*fx(nn,kl,mp,l) +
     *     dzl*fx(nn,kl,ml,l)) )

      dx2 = dxp*(amy*(amz*fx(np,kk,mm,l) + dzp*fx(np,kk,mp,l) + 
     *     dzl*fx(np,kk,ml,l)) + dyp*(amz*fx(np,kp,mm,l) +
     *     dzp*fx(np,kp,mp,l) + dzl*fx(np,kp,ml,l)) + 
     *     dyl*(amz*fx(np,kl,mm,l)+ dzp*fx(np,kl,mp,l) +
     *     dzl*fx(np,kl,ml,l)) )

      dx3 = dxl*(amy*(amz*fx(nl,kk,mm,l) + dzp*fx(nl,kk,mp,l) + 
     *     dzl*fx(nl,kk,ml,l)) + dyp*(amz*fx(nl,kp,mm,l) +
     *     dzp*fx(nl,kp,mp,l) + dzl*fx(nl,kp,ml,l)) + 
     *     dyl*(amz*fx(nl,kl,mm,l)+ dzp*fx(nl,kl,mp,l) +
     *     dzl*fx(nl,kl,ml,l)) )

      dx = dx1+dx2+dx3


       dy1 = amx*(amy*(amz*fy(nn,kk,mm,l) + dzp*fy(nn,kk,mp,l) + 
     *     dzl*fy(nn,kk,ml,l)) + dyp*(amz*fy(nn,kp,mm,l) +
     *     dzp*fy(nn,kp,mp,l) + dzl*fy(nn,kp,ml,l)) + 
     *     dyl*(amz*fy(nn,kl,mm,l)+ dzp*fy(nn,kl,mp,l) +
     *     dzl*fy(nn,kl,ml,l)) )

      dy2 = dxp*(amy*(amz*fy(np,kk,mm,l) + dzp*fy(np,kk,mp,l) + 
     *     dzl*fy(np,kk,ml,l)) + dyp*(amz*fy(np,kp,mm,l) +
     *     dzp*fy(np,kp,mp,l) + dzl*fy(np,kp,ml,l)) + 
     *     dyl*(amz*fy(np,kl,mm,l)+ dzp*fy(np,kl,mp,l) +
     *     dzl*fy(np,kl,ml,l)) )

      dy3 = dxl*(amy*(amz*fy(nl,kk,mm,l) + dzp*fy(nl,kk,mp,l) + 
     *     dzl*fy(nl,kk,ml,l)) + dyp*(amz*fy(nl,kp,mm,l) +
     *     dzp*fy(nl,kp,mp,l) + dzl*fy(nl,kp,ml,l)) + 
     *     dyl*(amz*fy(nl,kl,mm,l)+ dzp*fy(nl,kl,mp,l) +
     *     dzl*fy(nl,kl,ml,l)) )

      dy = dy1+dy2+dy3


      dz1 = amx*(amy*(amz*fz(nn,kk,mm,l) + dzp*fz(nn,kk,mp,l) + 
     *     dzl*fz(nn,kk,ml,l)) + dyp*(amz*fz(nn,kp,mm,l) +
     *     dzp*fz(nn,kp,mp,l) + dzl*fz(nn,kp,ml,l)) + 
     *     dyl*(amz*fz(nn,kl,mm,l)+ dzp*fz(nn,kl,mp,l) +
     *     dzl*fz(nn,kl,ml,l)) )

      dz2 = dxp*(amy*(amz*fz(np,kk,mm,l) + dzp*fz(np,kk,mp,l) + 
     *     dzl*fz(np,kk,ml,l)) + dyp*(amz*fz(np,kp,mm,l) +
     *     dzp*fz(np,kp,mp,l) + dzl*fz(np,kp,ml,l)) + 
     *     dyl*(amz*fz(np,kl,mm,l)+ dzp*fz(np,kl,mp,l) +
     *     dzl*fz(np,kl,ml,l)) )

      dz3 = dxl*(amy*(amz*fz(nl,kk,mm,l) + dzp*fz(nl,kk,mp,l) + 
     *     dzl*fz(nl,kk,ml,l)) + dyp*(amz*fz(nl,kp,mm,l) +
     *     dzp*fz(nl,kp,mp,l) + dzl*fz(nl,kp,ml,l)) + 
     *     dyl*(amz*fz(nl,kl,mm,l)+ dzp*fz(nl,kl,mp,l) +
     *     dzl*fz(nl,kl,ml,l)) )

      dz = dz1+dz2+dz3


c --------------
c new velocity
c --------------
      dx = part(4,j,l) + qtm*dx
      dy = part(5,j,l) + qtm*dy
      dz = part(6,j,l) + qtm*dz
c -----------------------
c average kinetic energy
c -----------------------
      sum1 = sum1 + (dx + part(4,j,l))**2 + (dy + part(5,j,l))**2 +
     *              (dz + part(6,j,l))**2

      part(4,j,l) = dx
      part(5,j,l) = dy
      part(6,j,l) = dz
c -------------
c new position
c -------------
      dx = part(1,j,l) + dx*dt
      dy = part(2,j,l) + dy*dt
      dz = part(3,j,l) + dz*dt
c ------------------------------
c periodic boundary conditions
c ------------------------------
      if (dx.lt.zero) dx = dx + anx
      if (dx.ge.anx) dx = dx - anx

      if (dy.lt.zero) dy = dy + any
      if (dy.ge.any) dy = dy - any
      
      if ((dx .gt. anx) .or. (dx .lt. 0)) then
        print *,'dx out of bound',dx,'anx',anx
       endif 
      if ((dy .gt. any) .or. (dy .lt. 0)) then
        print *,'dy out of bound',dy,'any',any
       endif 

      part(1,j,l) = dx
      part(2,j,l) = dy
      part(3,j,l) = dz

   10 continue
   20 continue
c --------------------------
c normalize kinetic energy
c --------------------------
      ek = ek + .125*sum1
      return
      end


C ========================================================================
c
c  This subroutine moves particles into appropriate spatial regions
c   ierr = (0,1) = (no,yes) error condition exists
c  
C ========================================================================
      subroutine pmove2 (part,edges,npp,sbufr,sbufl,rbufr,rbufl,ihole,
     1 jsr,jsl,jss,nz,kstrt,nvp,idimp,npmax,nblok,idps,nbmax,ntmax,ierr)
      dimension part(idimp,npmax,nblok)
      dimension edges(idps,nblok), npp(nblok)
      dimension sbufl(idimp,nbmax,nblok), sbufr(idimp,nbmax,nblok)
      dimension rbufl(idimp,nbmax,nblok), rbufr(idimp,nbmax,nblok)
      dimension jsl(idps,nblok), jsr(idps,nblok), jss(idps,nblok)
      dimension ihole(ntmax,nblok)

      anz = float(nz)
      ks = kstrt - 2
      iter = 4
c debugging section: count total number of particles before move
      npr = 0
      do 5 l = 1, nblok
      npr = npr + npp(l)
    5 continue
c --------------------------
c buffer outgoing particles
c --------------------------
   10 do 30 l = 1, nblok
      kb = l + ks
      jsl(1,l) = 0
      jsr(1,l) = 0
      jss(2,l) = 0
      do 20 j = 1, npp(l)
      zt = part(3,j,l)
c --------------------
c particles going up
c --------------------
      if (zt.ge.edges(2,l)) then
         if (jsr(1,l).lt.nbmax) then
            jsr(1,l) = jsr(1,l) + 1
            if ((kb+1).eq.nvp) zt = zt - anz
            if ((zt .gt. anz) .or. (zt .lt. 0)) then
                 print *,'zt out of bound',zt,'anz',anz
             endif 

            sbufr(1,jsr(1,l),l) = part(1,j,l)
            sbufr(2,jsr(1,l),l) = part(2,j,l)
            sbufr(3,jsr(1,l),l) = zt
            
            sbufr(4,jsr(1,l),l) = part(4,j,l)
            sbufr(5,jsr(1,l),l) = part(5,j,l)
            sbufr(6,jsr(1,l),l) = part(6,j,l)

            ihole(jsl(1,l)+jsr(1,l),l) = j
         else
            jss(2,l) = 1
         endif
c ---------------------
c particles going down
c ---------------------
      elseif (zt.lt.edges(1,l)) then
         if (jsl(1,l).lt.nbmax) then
            jsl(1,l) = jsl(1,l) + 1
            if (kb.eq.0) zt = zt + anz
            if ((zt .gt. anz) .or. (zt .lt. 0)) then
                 print *,'zt out of bound',zt,'anz',anz
             endif 
            sbufl(1,jsl(1,l),l) = part(1,j,l)
            sbufr(2,jsl(1,l),l) = part(2,j,l)
            sbufl(3,jsl(1,l),l) = zt

            sbufr(4,jsl(1,l),l) = part(4,j,l)
            sbufr(5,jsl(1,l),l) = part(5,j,l)
            sbufr(6,jsl(1,l),l) = part(6,j,l)

            ihole(jsl(1,l)+jsr(1,l),l) = j
         else
            jss(2,l) = 1
         endif
      endif
   20 continue
      jss(1,l) = jsl(1,l) + jsr(1,l)
   30 continue
c ---------------------------------
c check for full buffer condition
c ---------------------------------
      ibflg = 0
      do 40 l = 1, nblok
      ibflg = ibflg + jss(2,l)
   40 continue
c -----------------------
c copy particle buffers
c -----------------------
   50 iter = iter + 4
      do 80 l = 1, nblok
c --------------------------
c get particles from below
c --------------------------
      kl = l + ks
      if (kl.lt.1) then
         jsl(2,l) = jsr(1,kl+nvp)
      else
         jsl(2,l) = jsr(1,kl)
      endif

      do 60 j = 1, jsl(2,l)
      if (kl.lt.1) then
         rbufl(1,j,l) = sbufr(1,j,kl+nvp)
         rbufl(2,j,l) = sbufr(2,j,kl+nvp)
         rbufl(3,j,l) = sbufr(3,j,kl+nvp)
         rbufl(4,j,l) = sbufr(4,j,kl+nvp)
         rbufl(5,j,l) = sbufr(5,j,kl+nvp)
         rbufl(6,j,l) = sbufr(6,j,kl+nvp)

      else
         rbufl(1,j,l) = sbufr(1,j,kl)
         rbufl(2,j,l) = sbufr(2,j,kl)
         rbufl(3,j,l) = sbufr(3,j,kl)
         rbufl(4,j,l) = sbufr(4,j,kl)
         rbufl(5,j,l) = sbufr(5,j,kl)
         rbufl(6,j,l) = sbufr(6,j,kl)

      endif
   60 continue
c ---------------------------
c get particles from above
c ---------------------------
      kr = l + ks + 2
      if (kr.gt.nvp) then
         jsr(2,l) = jsl(1,kr-nvp)
      else
         jsr(2,l) = jsl(1,kr)
      endif
      do 70 j = 1, jsr(2,l)
      if (kr.gt.nvp) then
         rbufr(1,j,l) = sbufl(1,j,kr-nvp)
         rbufr(2,j,l) = sbufl(2,j,kr-nvp)
         rbufr(3,j,l) = sbufl(3,j,kr-nvp)
         rbufr(4,j,l) = sbufl(4,j,kr-nvp)
         rbufr(5,j,l) = sbufl(5,j,kr-nvp)
         rbufr(6,j,l) = sbufl(6,j,kr-nvp)

      else
         rbufr(1,j,l) = sbufl(1,j,kr)
         rbufr(2,j,l) = sbufl(2,j,kr)
         rbufr(3,j,l) = sbufl(3,j,kr)
         rbufr(4,j,l) = sbufl(4,j,kr)
         rbufr(5,j,l) = sbufl(5,j,kr)
         rbufr(6,j,l) = sbufl(6,j,kr)

      endif
   70 continue

   80 continue
c ------------------------------------------
c check if particles must be passed further
c ------------------------------------------
      nps = 0
      do 110 l = 1, nblok
c --------------------------------------------------------------
c check if any particles coming from above must go further down
c --------------------------------------------------------------
      jsl(1,l) = 0
      do 90 j = 1, jsr(2,l)
      if (rbufr(2,j,l).lt.edges(1,l)) jsl(1,l) = 1
   90 continue
c ------------------------------------------------------------
c check if any particles coming from below must go further up
c ------------------------------------------------------------
      jsr(1,l) = 0
      do 100 j = 1, jsl(2,l)
      if (rbufl(2,j,l).ge.edges(2,l)) jsr(1,l) = 1
  100 continue
      nps = nps + (jsl(1,l) + jsr(1,l))
  110 continue

      if (nps.eq.0) go to 150
c -------------------------------------------
c remove particles which do not belong here
c -------------------------------------------
      do 140 l = 1, nblok
      kb = l + ks
c -----------------------------------------
c first check particles coming from above
c -----------------------------------------
      jsl(1,l) = 0
      jss(2,l) = 0
      do 120 j = 1, jsr(2,l)
      zt = rbufr(3,j,l)
c ----------------------
c particles going down
c ----------------------
      if (zt.lt.edges(1,l)) then
         jsl(1,l) = jsl(1,l) + 1
         if (kb.eq.0) zt = zt + anz
         sbufl(1,jsl(1,l),l) = rbufr(1,j,l)
         sbufl(2,jsl(1,l),l) = rbufr(2,j,l)
         sbufl(3,jsl(1,l),l) = zt

         sbufl(4,jsl(1,l),l) = rbufr(4,j,l)
         sbufl(5,jsl(1,l),l) = rbufr(5,j,l)
         sbufl(6,jsl(1,l),l) = rbufr(6,j,l)
c ----------------------
c particles staying here
c ----------------------
      else
         jss(2,l) = jss(2,l) + 1
         rbufr(1,jss(2,l),l) = rbufr(1,j,l) 
         rbufr(2,jss(2,l),l) = rbufr(2,j,l) 
         rbufr(3,jss(2,l),l) = zt

         rbufr(4,jss(2,l),l) = rbufr(4,j,l)
         rbufr(5,jss(2,l),l) = rbufr(5,j,l)
         rbufr(6,jss(2,l),l) = rbufr(6,j,l)
      endif
  120 continue
      jsr(2,l) = jss(2,l)
c ---------------------------------------
c next check particles coming from below
c ---------------------------------------
      jsr(1,l) = 0
      jss(2,l) = 0
      do 130 j = 1, jsl(2,l)
      zt = rbufl(3,j,l)
c -------------------
c particles going up
c -------------------
      if (zt.ge.edges(2,l)) then
         jsr(1,l) = jsr(1,l) + 1
         if ((kb+1).eq.nvp) zt = zt - anz
         sbufr(1,jsr(1,l),l) = rbufl(1,j,l)
         sbufr(2,jsr(1,l),l) = rbufl(2,j,l)
         sbufr(3,jsr(1,l),l) = zt

         sbufr(4,jsr(1,l),l) = rbufl(4,j,l)
         sbufr(5,jsr(1,l),l) = rbufl(5,j,l)
         sbufr(6,jsr(1,l),l) = rbufl(6,j,l)
c ------------------------
c particles staying here
c ------------------------
      else
         jss(2,l) = jss(2,l) + 1
         rbufl(1,jss(2,l),l) = rbufl(1,j,l)
         rbufl(2,jss(2,l),l) = rbufl(2,j,l)
         rbufl(3,jss(2,l),l) = zt
       
         rbufl(4,jss(2,l),l) = rbufl(4,j,l)
         rbufl(5,jss(2,l),l) = rbufl(5,j,l)
         rbufl(6,jss(2,l),l) = rbufl(6,j,l)
      endif
  130 continue
      jsl(2,l) = jss(2,l)
  140 continue
c ---------------------------------------------
c check if move would overflow particle array
c ---------------------------------------------
  150 ierr = 0
      do 160 l = 1, nblok
      if ((npp(l)+jsl(2,l)+jsr(2,l)-jss(1,l)).gt.npmax) then
         jss(2,l) = 1
      else
         jss(2,l) = 0
      endif
      ierr = ierr + jss(2,l)
  160 continue

      if (ierr.gt.0) then
         write (6,*) 'particle overflow error, ierr =', ierr
         return
      endif
      do 200 l = 1, nblok
c ---------------------------------------------------
c distribute incoming particles from buffers
c ---------------------------------------------------
      jss(2,l) = min0(jss(1,l),jsl(2,l))
      do 170 j = 1, jss(2,l)
      part(1,ihole(j,l),l) = rbufl(1,j,l)
      part(2,ihole(j,l),l) = rbufl(2,j,l)
      part(3,ihole(j,l),l) = rbufl(3,j,l)
      part(4,ihole(j,l),l) = rbufl(4,j,l)
      part(5,ihole(j,l),l) = rbufl(5,j,l)
      part(6,ihole(j,l),l) = rbufl(6,j,l)


  170 continue
      if (jss(1,l).gt.jsl(2,l)) then
         jss(2,l) = min0(jss(1,l)-jsl(2,l),jsr(2,l))
      else
         jss(2,l) = jsl(2,l) - jss(1,l)
      endif
      do 180 j = 1, jss(2,l)
      if (jss(1,l).gt.jsl(2,l)) then
         part(1,ihole(j+jsl(2,l),l),l) = rbufr(1,j,l)
         part(2,ihole(j+jsl(2,l),l),l) = rbufr(2,j,l)
         part(3,ihole(j+jsl(2,l),l),l) = rbufr(3,j,l)
         part(4,ihole(j+jsl(2,l),l),l) = rbufr(4,j,l)
         part(5,ihole(j+jsl(2,l),l),l) = rbufr(5,j,l)
         part(6,ihole(j+jsl(2,l),l),l) = rbufr(6,j,l)

      else
c ------------------------------------------------------
c no more holes
c distribute remaining particles from below into bottom
c ------------------------------------------------------
         part(1,j+npp(l),l) = rbufl(1,j+jss(1,l),l)
         part(2,j+npp(l),l) = rbufl(2,j+jss(1,l),l)
         part(3,j+npp(l),l) = rbufl(3,j+jss(1,l),l)
         part(4,j+npp(l),l) = rbufl(4,j+jss(1,l),l)
         part(5,j+npp(l),l) = rbufl(5,j+jss(1,l),l)
         part(6,j+npp(l),l) = rbufl(6,j+jss(1,l),l)

      endif
  180 continue
      if (jss(1,l).le.jsl(2,l)) then
         npp(l) = npp(l) + (jsl(2,l) - jss(1,l))
         jss(1,l) = jsl(2,l)
      endif
      jss(2,l) = jss(1,l) - (jsl(2,l) + jsr(2,l))
      if (jss(2,l).gt.0) then
         jss(1,l) = (jsl(2,l) + jsr(2,l))
         jsr(2,l) = jss(2,l)
      else
         jss(1,l) = jss(1,l) - jsl(2,l)
         jsr(2,l) = -jss(2,l)
      endif
      do 190 j = 1, jsr(2,l)
c ---------------------------------------------------------------------
c holes left over
c fill up remaining holes in particle array with particles from bottom
c ---------------------------------------------------------------------
      if (jss(2,l).gt.0) then
         j1 = npp(l) - j + 1
         j2 = jss(1,l) + jss(2,l) - j + 1
         if (j1.gt.ihole(j2,l)) then
c ------------------------------------------------
c move particle only if it is below current hole
c ------------------------------------------------
            part(1,ihole(j2,l),l) = part(1,j1,l)
            part(2,ihole(j2,l),l) = part(2,j1,l)
            part(3,ihole(j2,l),l) = part(3,j1,l)
            part(4,ihole(j2,l),l) = part(4,j1,l)
            part(5,ihole(j2,l),l) = part(5,j1,l)
            part(6,ihole(j2,l),l) = part(6,j1,l)

         endif
      else
c --------------------------------------------------------
c no more holes
c distribute remaining particles from above into bottom
c --------------------------------------------------------
         part(1,j+npp(l),l) = rbufr(1,j+jss(1,l),l)
         part(2,j+npp(l),l) = rbufr(2,j+jss(1,l),l)
         part(3,j+npp(l),l) = rbufr(3,j+jss(1,l),l)
         part(4,j+npp(l),l) = rbufr(4,j+jss(1,l),l)
         part(5,j+npp(l),l) = rbufr(5,j+jss(1,l),l)
         part(6,j+npp(l),l) = rbufr(6,j+jss(1,l),l)

      endif
  190 continue
      if (jss(2,l).gt.0) then
         npp(l) = npp(l) - jsr(2,l)
      else
         npp(l) = npp(l) + jsr(2,l)
      endif
      jss(1,l) = 0
  200 continue
c -------------------------------------------------
c check if any particles have to be passed further
c -------------------------------------------------
      if (nps.gt.0) go to 50
c -------------------------------------------------------------------
c check if buffer overflowed and more particles remain to be checked
c -------------------------------------------------------------------
      if (ibflg.gt.0) go to 10
c ---------------------------------------------------------------
c debugging section: count total number of particles after move
c ---------------------------------------------------------------
      nps = 0
      do 205 l = 1, nblok
      nps = nps + npp(l)
  205 continue
      if (npr.ne.nps) write (6,*) 'number error, np-old,np-new=',npr,nps
      return
      end

      subroutine bitrv(mixup,sct,indx,indy,indz,nxhyd,nxyhd)

      complex sct
      dimension mixup(nxhyd), sct(nxyhd)


      indx1 = indx - 1
      indy1 = indy - 1
      indx1y = max0(indx1,indy1)
      indx1y = max0(indx1y,indz)
 
      nx = 2**indx
      ny = 2**indy
      nz  = 2**indz

      nxyz = max0(nx,ny)
      nxyz = max0(nxyz,nz)
      nxyhz = 2**indx1y

      do 30 j = 1, nxyhz
      lb = j - 1
      ll = 0
      do 20 k = 1, indx1y
      jb = lb/2
      it = lb - 2*jb
      lb = jb
      ll = 2*ll + it
   20 continue
      mixup(j) = ll + 1
   30 continue
c ---------------------------------------------
c sine/cosine table for the angles 2*n*pi/nxy
c ---------------------------------------------
      nxyzh = nxyz/2
      dnxyz = 6.28318530717959/float(nxyz)
      do 40 j = 1, nxyzh
      arg = dnxyz*float(j - 1)
      sct(j) = cmplx(cos(arg),-sin(arg))
   40 continue
      return 
      end

C =======================================================================
C =======================================================================
C In this fft routine, main steps for dfft except scrambling
C are the following :
C We divided 3-d domain only trough z-dimension and dublicated x and y-dim.
C 1. take fft along x axis (there is no communication)
C 2. take fft along y axis ( "                       ) 
C 3. Transpose along x-z plane (it needs all-to-all communication)
C 4. Take fft trough x axis( fft of z-axis)
C For ifft , do the above steps but in reverse order. that is it..
C =======================================================================
c
c This subroutine performs a two dimensional real to complex fast
c fourier transform and its inverse, using complex arithmetic,
c
c   indx/indy = exponent which determines length in x/y direction,
c   where nx=2**indx, ny=2**indy
c   ntpose = (0,1) = (no,yes) input, output data are transposed
c
c   if isign = 0, the fft tables are prepared
c   if isign = -1, an inverse fourier transform is performed
c   if ntpose = 0, f is the input and output array, g is a scratch array
c       f(n,m,l) = (1/nx*ny)*sum(f(j,k,i)*
c       exp(-sqrt(-1)*2pi*n*j/nx)*exp(-sqrt(-1)*2pi*mm*kk/ny))
c       where mm = m + kyp*(l - 1) and kk = k + kyp*(i - 1)
c   if ntpose = 1, f is the input and g is the output
c       g(m,n,l) = (1/nx*ny)*sum(f(j,k,i)*
c       exp(-sqrt(-1)*2pi*nn*j/nx)*exp(-sqrt(-1)*2pi*m*kk/ny))
c       where nn = n + kxp*(l - 1) and kk = k + kyp*(i - 1)
c   if isign = 1, a forward fourier transform is performed
c   if ntpose = 0, f is the input and output array, g is a scratch array
c   f(j,k,i) = sum(f(n,m,l)*exp(sqrt(-1)*2pi*n*j/nx)*
c       exp(sqrt(-1)*2pi*mm*kk/ny))
c   where mm = m + kyp*(l - 1) and kk = k + kyp*(i - 1)
c   if ntpose = 1, g is the input and f is the output
c   f(j,k,i) = sum(g(m,n,l)*exp(sqrt(-1)*2pi*nn*j/nx)*
c       exp(sqrt(-1)*2pi*m*kk/ny))
c   where nn = n + kxp*(l - 1) and kk = k + kyp*(i - 1)
c   g(k,1,1) = mode nx/2,k-1, where ny/2+2 <= k <= ny, and
c   imaginary part of g(1,1,1) = real part of mode nx/2,0 and
c   imaginary part of g(ny/2+1,1,1) = real part of mode nx/2,ny/2
c   written by viktor k. decyk, ucla
C =========================================================================

 
      subroutine pfft2r(f,g,bs,br,isign,ntpose,mixup,sct,indx,indy,indz,
     1 kstrt,nxvh,nyvh,nzv,kxp,kyp,kzp,jblok,kblok,nxhyd,nxyhd)

      complex f, g, bs, br, sct, s, t, t1,tf
      dimension f(nxvh,nyvh,kzp,kblok), g(nzv,nyvh,kxp,jblok)
      dimension mixup(nxhyd), sct(nxyhd)
      dimension tf(nxvh,nyvh,kzp,kblok)

      indx1 = indx - 1
      indy1 = indy - 1
      indx1y = max0(indx1,indy1)
      indx1y = max0(indx1y,indz)
 
      nx = 2**indx
      nxh = nx/2
      nxhh = nx/4
      nxh2 = nxh + 2

      ny = 2**indy
      nyh = ny/2
      nyhh = ny/4
      nyh2 = nyh + 2

      nz  = 2**indz
      nzh = nz/2
      nz2 = nz + 2

      nxyz = max0(nx,ny)
      nxyz = max0(nxyz,nz)
      nxyhz = 2**indx1y

      ks = kstrt - 2
      if (isign) 50, 10, 300
   10 return
c ----------------------------
c inverse fourier transform
c ----------------------------

   50 nrx = nxyhz/nxh
      do 80 l = 1, kblok
c --------------------------------
c bit-reverse array elements in x
c --------------------------------
      do 70 j = 1, nxh
      j1 = (mixup(j) - 1)/nrx + 1
      if (j.ge.j1) go to 70
      do 66 iy = 1, nyh
      do 60 k = 1, kzp
      t = f(j1,iy,k,l)
      f(j1,iy,k,l) = f(j,iy,k,l)
      f(j,iy,k,l) = t
   60 continue
   66 continue  
   70 continue
   80 continue
c -----------------------
c first transform in x
c -----------------------
      nrx = nxyz/nxh
      do 130 m = 1, indx1
      ns = 2**(m - 1)
      ns2 = ns + ns
      km = nxhh/ns
      kmr = km*nrx
      do 120 l = 1, kblok
      do 110 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 100 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = sct(1+kmr*(j-1))
      do 99 iy = 1,nyh
      do 90 i = 1, kzp
      t = s*f(j2,iy,i,l)
      f(j2,iy,i,l) = f(j1,iy,i,l) - t
      f(j1,iy,i,l) = f(j1,iy,i,l) + t
   90 continue
   99 continue  
  100 continue
  110 continue
  120 continue
  130 continue
c---------------------------------------
c unscramble coefficients and normalize
c ---------------------------------------
      kmr = nxyz/nx
      ani = 1./float(2*nx*ny*nz)
      do 170 l = 1, kblok
      do 150 j = 2, nxhh
      t1 = cmplx(aimag(sct(1+kmr*(j-1))),-real(sct(1+kmr*(j-1))))
      do 144 iy = 1,nyh
      do 140 k  = 1, kzp
      t = conjg(f(nxh2-j,iy,k,l))
      s = f(j,iy,k,l) + t
      t = (f(j,iy,k,l) - t)*t1
      f(j,iy,k,l) = ani*(s + t)
      f(nxh2-j,iy,k,l) = ani*conjg(s - t)
  140 continue
  144 continue 
  150 continue
      do 160 k = 1, kzp
      do 166 iy = 1,nyh
      f(nxhh+1,iy,k,l) = 2.*ani*conjg(f(nxhh+1,iy,k,l))
      f(1,iy,k,l) = 2.*ani*cmplx(real(f(1,iy,k,l)) + aimag(f(1,iy,k,l)),
     1 real(f(1,iy,k,l)) - aimag(f(1,iy,k,l)))
  166 continue 
  160 continue
  170 continue


C ------------------------------------------------------------------
C make transform in y direction with same was as we did x direction
C ------------------------------------------------------------------

      nry = nxyhz/nyh
      do 800 l = 1, kblok
c --------------------------------
c bit-reverse array elements in y
c --------------------------------
      do 810 j = 1, nyh
      j1 = (mixup(j) - 1)/nry + 1
      if (j.ge.j1) go to 810
      do 820 ix = 1, nxh
      do 830 k = 1, kzp
      t = f(ix,j1,k,l)
      f(ix,j1,k,l) = f(ix,j,k,l)
      f(ix,j,k,l) = t
  830 continue
  820 continue  
  810 continue
  800 continue
c -----------------------
c then transform in y
c -----------------------
      nry = nxyz/nyh
      do 840 m = 1, indy1
      ns = 2**(m - 1)
      ns2 = ns + ns
      km = nyhh/ns
      kmr = km*nry
      do 850 l = 1, kblok
      do 860 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 870 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = sct(1+kmr*(j-1))
      do 880 ix = 1,nxh
      do 890 i = 1, kzp
      t = s*f(ix,j2,i,l)
      f(ix,j2,i,l) = f(ix,j1,i,l) - t
      f(ix,j1,i,l) = f(ix,j1,i,l) + t
  890 continue
  880 continue  
  870 continue
  860 continue
  850 continue
  840 continue
c---------------------------------------
c unscramble coefficients and normalize
c ---------------------------------------
      kmr = nxyz/ny
      ani = 1./float(2*nx*ny*nz)
      do 900 l = 1, kblok
      do 910 j = 2, nyhh
      t1 = cmplx(aimag(sct(1+kmr*(j-1))),-real(sct(1+kmr*(j-1))))
      do 920 ix = 1,nxh
      do 930 k = 1, kzp
      t = conjg(f(ix,nyh2-j,k,l))
      s = f(ix,j,k,l) + t
      t = (f(ix,j,k,l) - t)*t1
      f(ix,j,k,l) = ani*(s + t)
      f(ix,nyh2-j,k,l) = ani*conjg(s - t)
  930 continue
  920 continue 
  910 continue
      do 940 k = 1, kzp
      do 950 ix = 1,nxh
      f(ix,nyhh+1,k,l) = 2.*ani*conjg(f(ix,nyhh+1,k,l))
      f(ix,1,k,l) = 2.*ani*cmplx(real(f(ix,1,k,l)) + aimag(f(ix,1,k,l)),
     1 real(f(ix,1,k,l)) - aimag(f(ix,1,k,l)))
  950 continue 
  940 continue
  900 continue

c --------------------------------------
c transpose f array to g along x-z plane
c --------------------------------------
      
C      print *,'geldik'
C      print *,nxvh,nyvh,kzp,kblok
C      do i=1,nxvh
C       do j=1,nyvh
C         do k=1,kzp
C            do l=1,kblok
C              tf(i,j,k,l) = f(i,j,k,l) 
C              print *,'before',i,j,k,l,f(i,j,k,l)
C           enddo
C           enddo
C           enddo
C           enddo
    

C!       print *,f
    
      call ptpose(f,g,bs,br,nxh,nyh,nz,kstrt,nxvh,nyvh,nzv,
     1 kxp,kyp,kzp,nyvh,nxvh,jblok,kblok)

C      print *,'after ptpose'
      
C      call ptpose(g,f,bs,br,nz,nyh,nxh,kstrt,nzv,nyvh,nxvh,
C     1 kzp,kyp,kxp,nyvh,nxvh,jblok,kblok)
CC      print *,'after ptpose' 
C      do i=1,nxvh
C       do j=1,nyvh
C         do k=1,kzp
C            do l=1,kblok
C              print *,tf(i,j,k,l)-f(i,j,k,l)
C              print *,tf(i,j,k,l)
C              print *,'after',i,j,k,l,f(i,j,k,l)
C           enddo
C           enddo
C           enddo
C           enddo
    

C      stop         


      nrz = nxyhz/nz
      do 210 l = 1, jblok
c ---------------------------------
c bit-reverse array elements in z
c ---------------------------------
      do 200 k = 1, nz
      k1 = (mixup(k) - 1)/nrz + 1
      if (k.ge.k1) go to 200
      do 190 j = 1, kxp
      do 199 iy = 1,nyh
      t = g(k1,iy,j,l)
      g(k1,iy,j,l) = g(k,iy,j,l)
      g(k,iy,j,l) = t
  199 continue
  190 continue
  200 continue
  210 continue
c ---------------------
c then transform in z
c ---------------------
      nrz = nxyz/nz
      do 260 m = 1, indz
      ns = 2**(m - 1)
      ns2 = ns + ns
      km = nzh/ns
      kmr = km*nrz
      do 250 l = 1, jblok
      do 240 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 230 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = sct(1+kmr*(j-1))
      do 222 iy = 1,nyh
      do 220 i = 1, kxp
      t = s*g(j2,iy,i,l)
      g(j2,iy,i,l) = g(j1,iy,i,l) - t
      g(j1,iy,i,l) = g(j1,iy,i,l) + t
  220 continue
  222 continue 
  230 continue
  240 continue
  250 continue
  260 continue
c -------------------------------
c unscramble modes kx = 0, nx/2
c -------------------------------
      do 280 l = 1, jblok
      if ((l+ks).gt.0) go to 280
      do 277 iy = 1,nyh
      do 270 k = 2, nzh
      s = g(nz2-k,iy,1,l)
      g(nz2-k,iy,1,l) = .5*cmplx(aimag(g(k,iy,1,l) + s),
     1 real(g(k,iy,1,l) - s))
      g(k,iy,1,l) = .5*cmplx(real(g(k,iy,1,l) + s),
     1 aimag(g(k,iy,1,l) - s))
  270 continue
  277 continue
  280 continue

c -------------------------------
c unscramble modes ky = 0, ny/2
c -------------------------------
      do 2801 l = 1, jblok
      do 2771 ix = 1, kxp
      do 2701 k = 2, nzh
      s = g(nz2-k,1,ix,l)
      g(nz2-k,1,ix,l) = .5*cmplx(aimag(g(k,1,ix,l) + s),
     1 real(g(k,1,ix,l) - s))
      g(k,1,ix,l) = .5*cmplx(real(g(k,1,ix,l) + s),
     1 aimag(g(k,1,ix,l) - s))
 2701 continue
 2771 continue
 2801 continue


C these two calls of transpose are not made in this fft .So skip them.
c ------------------------
c transpose g array to f
c ------------------------
      return


c ---------------------------
c forward fourier transform
c transpose f array to g
c ---------------------------
      
  300 nrz = nxyhz/nz
      do 350 l = 1, jblok

c ---------------------------
c scramble modes ky = 0, ny/2
c ---------------------------

      do 3111 ix = 1,kxp
      do 3101 k = 2, nzh
      s = cmplx(aimag(g(nz2-k,1,ix,l)),real(g(nz2-k,1,ix,l)))
      g(nz2-k,1,ix,l) = conjg(g(k,1,ix,l) - s)
      g(k,1,ix,l) = g(k,1,ix,l) + s
 3101 continue
 3111 continue 

c ---------------------------
c scramble modes kx = 0, nx/2
c ---------------------------
      if ((l+ks).gt.0) go to 320
      do 311 iy = 1,nyh
      do 310 k = 2, nzh
      s = cmplx(aimag(g(nz2-k,iy,1,l)),real(g(nz2-k,iy,1,l)))
      g(nz2-k,iy,1,l) = conjg(g(k,iy,1,l) - s)
      g(k,iy,1,l) = g(k,iy,1,l) + s
  310 continue
  311 continue 
c ----------------------------------
c bit-reverse array elements in z
c ----------------------------------
  320 do 340 k = 1, nz
      k1 = (mixup(k) - 1)/nrz + 1
      if (k.ge.k1) go to 340
      do 330 j = 1, kxp
      do 333 iy = 1,nyh
      t = g(k1,iy,j,l)
      g(k1,iy,j,l) = g(k,iy,j,l)
      g(k,iy,j,l) = t
  333 continue
  330 continue
  340 continue
  350 continue
c ----------------------
c first transform in z
c ----------------------
      nrz = nxyz/nz
      do 400 m = 1, indz
      ns = 2**(m - 1)
      ns2 = ns + ns
      km = nzh/ns
      kmr = km*nrz
      do 390 l = 1, jblok
      do 380 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 370 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = conjg(sct(1+kmr*(j-1)))
      do 360 i = 1, kxp
      do 366 iy = 1,nyh
      t = s*g(j2,iy,i,l)
      g(j2,iy,i,l) = g(j1,iy,i,l) - t
      g(j1,iy,i,l) = g(j1,iy,i,l) + t
  366 continue
  360 continue
  370 continue
  380 continue
  390 continue
  400 continue
c ------------------------
c transpose g array to f
c ------------------------

  410 call ptpose(g,f,bs,br,nz,nyh,nxh,kstrt,nzv,nyvh,nxvh,
     1 kzp,kyp,kxp,nyvh,nxvh,jblok,kblok)

      nrx = nxyhz/nxh
      kmr = nxyz/nx
      do 470 l = 1, kblok
c ----------------------
c scramble coefficients
c ----------------------
      do 430 j = 2, nxhh
      t1 = cmplx(aimag(sct(1+kmr*(j-1))),real(sct(1+kmr*(j-1))))
      do 420 k = 1, kzp
      do 422 iy = 1,nyh
      t = conjg(f(nxh2-j,iy,k,l))
      s = f(j,iy,k,l) + t
      t = (f(j,iy,k,l) - t)*t1
      f(j,iy,k,l) = s + t
      f(nxh2-j,iy,k,l) = conjg(s - t)
  422 continue 
  420 continue
  430 continue
      do 440 k = 1, kzp
      do 444 iy = 1,nyh
      f(nxhh+1,iy,k,l) = 2.*conjg(f(nxhh+1,iy,k,l))
      f(1,iy,k,l) = cmplx(real(f(1,iy,k,l)) + aimag(f(1,iy,k,l)),
     1 real(f(1,iy,k,l)) - aimag(f(1,iy,k,l))) 
  444 continue 
  440 continue
c ---------------------------------
c bit-reverse array elements in x
c ---------------------------------
      do 460 j = 1, nxh
      j1 = (mixup(j) - 1)/nrx + 1
      if (j.ge.j1) go to 460
      do 450 k = 1, kzp
      do 455 iy = 1,nyh
      t = f(j1,iy,k,l)
      f(j1,iy,k,l) = f(j,iy,k,l)
      f(j,iy,k,l) = t
  455 continue 
  450 continue
  460 continue
  470 continue
c ---------------------
c then transform in x
c ---------------------
      nrx = nxyz/nxh
      do 520 m = 1, indx1
      ns = 2**(m - 1)
      ns2 = ns + ns
      km = nxhh/ns
      kmr = km*nrx
      do 510 l = 1, kblok
      do 500 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 490 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = conjg(sct(1+kmr*(j-1)))
      do 480 i = 1, kzp
      do 488 iy = 1,nyh
      t = s*f(j2,iy,i,l)
      f(j2,iy,i,l) = f(j1,iy,i,l) - t
      f(j1,iy,i,l) = f(j1,iy,i,l) + t
  488 continue 
  480 continue
  490 continue
  500 continue
  510 continue
  520 continue

C take fft along y direction 

      nry = nxyhz/nyh
      kmr = nxyz/ny
      do 600 l = 1, kblok
c ----------------------
c scramble coefficients
c ----------------------
      do 610 j = 2, nyhh
      t1 = cmplx(aimag(sct(1+kmr*(j-1))),real(sct(1+kmr*(j-1))))
      do 620 k = 1, kzp
      do 630 ix = 1,nxh
      t = conjg(f(ix,nyh2-j,k,l))
      s = f(ix,j,k,l) + t
      t = (f(ix,j,k,l) - t)*t1
      f(ix,j,k,l) = s + t
      f(ix,nyh2-j,k,l) = conjg(s - t)
  630 continue 
  620 continue
  610 continue
      do 640 k = 1, kzp
      do 650 ix = 1,nxh
      f(ix,nyhh+1,k,l) = 2.*conjg(f(ix,nyhh+1,k,l))
      f(ix,1,k,l) = cmplx(real(f(ix,1,k,l)) + aimag(f(ix,1,k,l)),
     1 real(f(ix,1,k,l)) - aimag(f(ix,1,k,l))) 
  650 continue 
  640 continue
c ---------------------------------
c bit-reverse array elements in y
c ---------------------------------
      do 660 j = 1, nyh
      j1 = (mixup(j) - 1)/nry + 1
      if (j.ge.j1) go to 660
      do 670 k = 1, kzp
      do 680 ix = 1,nxh
      t = f(ix,j1,k,l)
      f(ix,j1,k,l) = f(ix,j,k,l)
      f(ix,j,k,l) = t
  680 continue 
  670 continue
  660 continue
  600 continue
c ---------------------
c then transform in y
c ---------------------
      nry = nxyz/nyh
      do 690 m = 1, indy1
      ns = 2**(m - 1)
      ns2 = ns + ns
      km = nyhh/ns
      kmr = km*nry
      do 700 l = 1, kblok
      do 710 k = 1, km
      k1 = ns2*(k - 1)
      k2 = k1 + ns
      do 720 j = 1, ns
      j1 = j + k1
      j2 = j + k2
      s = conjg(sct(1+kmr*(j-1)))
      do 730 i = 1, kzp
      do 740 ix = 1,nxh
      t = s*f(ix,j2,i,l)
      f(ix,j2,i,l) = f(ix,j1,i,l) - t
      f(ix,j1,i,l) = f(ix,j1,i,l) + t
  740 continue 
  730 continue
  720 continue
  710 continue
  700 continue
  690 continue


      return
      end
C ========================================================================
c
c  This subroutine performs a transpose of a matrix f, distributed in y,
c  to a matrix g, distributed in x, that is,
c    g(k+kyp*(m-1),j,l) = f(j+kxp*(l-1),k,m), where
c    1 <= j <= kxp, 1 <= k <= kyp, 1 <= l <= nx/kxp, 1 <= m <= ny/kyp
c  and where indices l and m can be distributed across processors.
c    f = complex input array
c    g = complex output array
c    s, t = complex scratch arrays
C ========================================================================

  
      subroutine ptpose(f,g,s,t,nx,ny,nz,kstrt,nxv,nyv,nzv,
     1 kxp,kyp,kzp,nyvh,nxvh,jblok,kblok)
      complex f, g, s, t
      dimension f(nxv,nyvh,kzp,kblok), g(nzv,nyvh,kxp,jblok)
      dimension s(kxp,kyp,kzp,kblok), t(kxp,kyp,kzp,jblok)


      ks = kstrt - 2
      kxb = nx/kxp
      kzb = nz/kzp

      do 11 iy = 1,nyvh
      do 40 l = 1, jblok
      joff = kxp*(l + ks)
c      do 11 iy = 1,nyvh 
      do 30 i = 1, kzb
      koff = kzp*(i - 1)
      do 20 k = 1, kzp
      do 10 j = 1, kxp
      g(k+koff,iy,j,l) = f(j+joff,iy,k,i)
   10 continue
   20 continue
   30 continue
   40 continue
   11 continue  

      return
      end

      function ranorm(d)
c this program calculates a random number y from a gaussian distribution
c with zero mean and unit variance, according to the method of
c mueller and box:
c    y(k) = (-2*ln(x(k)))**1/2*sin(2*pi*x(k+1))
c    y(k+1) = (-2*ln(x(k)))**1/2*cos(2*pi*x(k+1)),
c where x is a random number uniformly distributed on (0,1).
c written for the ibm by viktor k. decyk, ucla
      integer r1,r2,r4,r5
      double precision ranorm,h1l,h1u,h2l,r0,r3,asc,bsc,temp
      save iflg,r1,r2,r4,r5,h1l,h1u,h2l,r0
      data r1,r2,r4,r5 /885098780,1824280461,1396483093,55318673/
      data h1l,h1u,h2l /65531.0d0,32767.0d0,65525.0d0/
      data iflg,r0 /0,0.0d0/
      if (iflg.eq.0) go to 10
      ranorm = r0
      r0 = 0.0d0
      iflg = 0
      return
   10 isc = 65536
      asc = dfloat(isc)
      bsc = asc*asc
      i1 = r1 - (r1/isc)*isc
      r3 = h1l*dfloat(r1) + asc*h1u*dfloat(i1)
      i1 = r3/bsc
      r3 = r3 - dfloat(i1)*bsc
      bsc = 0.5d0*bsc
      i1 = r2/isc
      isc = r2 - i1*isc
      r0 = h1l*dfloat(r2) + asc*h1u*dfloat(isc)
      asc = 1.0d0/bsc
      isc = r0*asc
      r2 = r0 - dfloat(isc)*bsc
      r3 = r3 + (dfloat(isc) + 2.0d0*h1u*dfloat(i1))
      isc = r3*asc
      r1 = r3 - dfloat(isc)*bsc
      temp = dsqrt(-2.0d0*dlog((dfloat(r1) + dfloat(r2)*asc)*asc))
      isc = 65536
      asc = dfloat(isc)
      bsc = asc*asc
      i1 = r4 - (r4/isc)*isc
      r3 = h2l*dfloat(r4) + asc*h1u*dfloat(i1)
      i1 = r3/bsc
      r3 = r3 - dfloat(i1)*bsc
      bsc = 0.5d0*bsc
      i1 = r5/isc
      isc = r5 - i1*isc
      r0 = h2l*dfloat(r5) + asc*h1u*dfloat(isc)
      asc = 1.0d0/bsc
      isc = r0*asc
      r5 = r0 - dfloat(isc)*bsc
      r3 = r3 + (dfloat(isc) + 2.0d0*h1u*dfloat(i1))
      isc = r3*asc
      r4 = r3 - dfloat(isc)*bsc
      r0 = 6.28318530717959d0*((dfloat(r4) + dfloat(r5)*asc)*asc)
      ranorm = temp*dsin(r0)
      r0 = temp*dcos(r0)
      iflg = 1
      return
      end

      SUBROUTINE TIMERA(T)
C   ==================================================================
C   ===                                                            ===
C   ===    This subroutine returns the elapsed time in seconds     ===
C   ===                    since its first CALL                    ===
C   ===                                                            ===
C   ==================================================================
 
      real T
      real secs, tarray(2)
c      secs = etime(tarray)
      t = secs
      return

      END
